# generated with: ./merge-translators.sh

translators = (
[
  ('Asturian', 'ast', [
    ('Xuacu Saturio', 'https://launchpad.net/~xuacusk8'),
    ('ivarela', 'https://launchpad.net/~ivarela'),
  ]),
  ('Bengali', 'bn', [
    ('Iftekhar Mohammad', 'https://launchpad.net/~iftekhar'),
  ]),
  ('Bosnian', 'bs', [
    ('Kenan Dervišević', 'https://launchpad.net/~kenan3008'),
  ]),
  ('Brazilian Portuguese', 'pt_BR', [
    ('Fábio Nogueira', 'https://launchpad.net/~fnogueira'),
    ('Rafael Neri', 'https://launchpad.net/~rafepel'),
  ]),
  ('Chinese (Simplified)', 'zh_CN', [
    ('Xiaoxing Ye', 'https://launchpad.net/~xiaoxing'),
  ]),
  ('Chinese (Traditional)', 'zh_TW', [
    ('Po-Chun Huang', 'https://launchpad.net/~aphroteus'),
  ]),
  ('English (United Kingdom)', 'en_GB', [
    ('Andi Chandler', 'https://launchpad.net/~bing'),
    ('Anthony Harrington', 'https://launchpad.net/~untaintableangel'),
    ('B Clausius', 'https://launchpad.net/~barcc'),
  ]),
  ('French', 'fr', [
    ('Aurélien Ribeiro', 'https://launchpad.net/~aurel-koala'),
    ('Baptiste Fontaine', 'https://launchpad.net/~bfontaine'),
    ('Célestin Taramarcaz', 'https://launchpad.net/~celestin'),
    ('Havok Novak', 'https://launchpad.net/~havok-novak'),
    ('Nicolas Delvaux', 'https://launchpad.net/~malizor'),
    ('Sylvie Gallet', 'https://launchpad.net/~sylvie-gallet'),
    ('lann', 'https://launchpad.net/~lann'),
  ]),
  ('Galician', 'gl', [
    ('Miguel Anxo Bouzada', 'https://launchpad.net/~mbouzada'),
  ]),
  ('German', 'de', [
    ('B Clausius', 'https://launchpad.net/~barcc'),
  ]),
  ('Greek', 'el', [
    ('George Christofis', 'https://launchpad.net/~geochr'),
    ('mara sdr', 'https://launchpad.net/~paren8esis'),
  ]),
  ('Hebrew', 'he', [
    ('Yaron', 'https://launchpad.net/~sh-yaron'),
  ]),
  ('Indonesian', 'id', [
    ('Trisno Pamuji', 'https://launchpad.net/~tri.snowman'),
  ]),
  ('Italian', 'it', [
    ('Alfio Missaglia', 'https://launchpad.net/~missaglialfio'),
    ('Claudio Arseni', 'https://launchpad.net/~claudio.arseni'),
    ('Francesco Muriana', 'https://launchpad.net/~f-muriana'),
  ]),
  ('Japanese', 'ja', [
    ('epii', 'https://launchpad.net/~epii'),
  ]),
  ('Kirghiz', 'ky', [
    ('ballpen', 'https://launchpad.net/~ballpen'),
  ]),
  ('Malay', 'ms', [
    ('abuyop', 'https://launchpad.net/~abuyop'),
  ]),
  ('Polish', 'pl', [
    ('Mike08', 'https://launchpad.net/~micou8'),
    ('Szymon Nieznański', 'https://launchpad.net/~isamu715'),
  ]),
  ('Russian', 'ru', [
    ('Oleg Koptev', 'https://launchpad.net/~koptev-oleg'),
    ('scientistnik', 'https://launchpad.net/~nozdrin-plotnitsky'),
  ]),
  ('Serbian', 'sr', [
    ('Мирослав Николић', 'https://launchpad.net/~lipek'),
  ]),
  ('Spanish', 'es', [
    ('Adolfo Jayme Barrientos', 'https://launchpad.net/~fitoschido'),
    ('Dante Díaz', 'https://launchpad.net/~dante'),
    ('Eduardo Alberto Calvo', 'https://launchpad.net/~edu5800'),
  ]),
  ('Ukrainian', 'uk', [
    ('Yuri Chornoivan', 'https://launchpad.net/~yurchor-gmail'),
  ]),
  ('Uzbek', 'uz', [
    ('Akmal Xushvaqov', 'https://launchpad.net/~uzadmin'),
  ]),
]
)
