# pybiklib/gl.pxd
# generated with ./tools/create-gl-pxd.py

from libc.stddef cimport ptrdiff_t
from libc.stdint cimport int32_t, intptr_t, int8_t, uint8_t


# from /usr/include/GL/gl.h:

cdef extern from 'GL/gl.h':
    enum: GL_VERSION_1_1
    enum: GL_VERSION_1_2
    enum: GL_VERSION_1_3
    enum: GL_ARB_imaging
    enum: GL_FALSE
    enum: GL_TRUE
    enum: GL_BYTE
    enum: GL_UNSIGNED_BYTE
    enum: GL_SHORT
    enum: GL_UNSIGNED_SHORT
    enum: GL_INT
    enum: GL_UNSIGNED_INT
    enum: GL_FLOAT
    enum: GL_2_BYTES
    enum: GL_3_BYTES
    enum: GL_4_BYTES
    enum: GL_DOUBLE
    enum: GL_POINTS
    enum: GL_LINES
    enum: GL_LINE_LOOP
    enum: GL_LINE_STRIP
    enum: GL_TRIANGLES
    enum: GL_TRIANGLE_STRIP
    enum: GL_TRIANGLE_FAN
    enum: GL_QUADS
    enum: GL_QUAD_STRIP
    enum: GL_POLYGON
    enum: GL_VERTEX_ARRAY
    enum: GL_NORMAL_ARRAY
    enum: GL_COLOR_ARRAY
    enum: GL_INDEX_ARRAY
    enum: GL_TEXTURE_COORD_ARRAY
    enum: GL_EDGE_FLAG_ARRAY
    enum: GL_VERTEX_ARRAY_SIZE
    enum: GL_VERTEX_ARRAY_TYPE
    enum: GL_VERTEX_ARRAY_STRIDE
    enum: GL_NORMAL_ARRAY_TYPE
    enum: GL_NORMAL_ARRAY_STRIDE
    enum: GL_COLOR_ARRAY_SIZE
    enum: GL_COLOR_ARRAY_TYPE
    enum: GL_COLOR_ARRAY_STRIDE
    enum: GL_INDEX_ARRAY_TYPE
    enum: GL_INDEX_ARRAY_STRIDE
    enum: GL_TEXTURE_COORD_ARRAY_SIZE
    enum: GL_TEXTURE_COORD_ARRAY_TYPE
    enum: GL_TEXTURE_COORD_ARRAY_STRIDE
    enum: GL_EDGE_FLAG_ARRAY_STRIDE
    enum: GL_VERTEX_ARRAY_POINTER
    enum: GL_NORMAL_ARRAY_POINTER
    enum: GL_COLOR_ARRAY_POINTER
    enum: GL_INDEX_ARRAY_POINTER
    enum: GL_TEXTURE_COORD_ARRAY_POINTER
    enum: GL_EDGE_FLAG_ARRAY_POINTER
    enum: GL_V2F
    enum: GL_V3F
    enum: GL_C4UB_V2F
    enum: GL_C4UB_V3F
    enum: GL_C3F_V3F
    enum: GL_N3F_V3F
    enum: GL_C4F_N3F_V3F
    enum: GL_T2F_V3F
    enum: GL_T4F_V4F
    enum: GL_T2F_C4UB_V3F
    enum: GL_T2F_C3F_V3F
    enum: GL_T2F_N3F_V3F
    enum: GL_T2F_C4F_N3F_V3F
    enum: GL_T4F_C4F_N3F_V4F
    enum: GL_MATRIX_MODE
    enum: GL_MODELVIEW
    enum: GL_PROJECTION
    enum: GL_TEXTURE
    enum: GL_POINT_SMOOTH
    enum: GL_POINT_SIZE
    enum: GL_POINT_SIZE_GRANULARITY
    enum: GL_POINT_SIZE_RANGE
    enum: GL_LINE_SMOOTH
    enum: GL_LINE_STIPPLE
    enum: GL_LINE_STIPPLE_PATTERN
    enum: GL_LINE_STIPPLE_REPEAT
    enum: GL_LINE_WIDTH
    enum: GL_LINE_WIDTH_GRANULARITY
    enum: GL_LINE_WIDTH_RANGE
    enum: GL_POINT
    enum: GL_LINE
    enum: GL_FILL
    enum: GL_CW
    enum: GL_CCW
    enum: GL_FRONT
    enum: GL_BACK
    enum: GL_POLYGON_MODE
    enum: GL_POLYGON_SMOOTH
    enum: GL_POLYGON_STIPPLE
    enum: GL_EDGE_FLAG
    enum: GL_CULL_FACE
    enum: GL_CULL_FACE_MODE
    enum: GL_FRONT_FACE
    enum: GL_POLYGON_OFFSET_FACTOR
    enum: GL_POLYGON_OFFSET_UNITS
    enum: GL_POLYGON_OFFSET_POINT
    enum: GL_POLYGON_OFFSET_LINE
    enum: GL_POLYGON_OFFSET_FILL
    enum: GL_COMPILE
    enum: GL_COMPILE_AND_EXECUTE
    enum: GL_LIST_BASE
    enum: GL_LIST_INDEX
    enum: GL_LIST_MODE
    enum: GL_NEVER
    enum: GL_LESS
    enum: GL_EQUAL
    enum: GL_LEQUAL
    enum: GL_GREATER
    enum: GL_NOTEQUAL
    enum: GL_GEQUAL
    enum: GL_ALWAYS
    enum: GL_DEPTH_TEST
    enum: GL_DEPTH_BITS
    enum: GL_DEPTH_CLEAR_VALUE
    enum: GL_DEPTH_FUNC
    enum: GL_DEPTH_RANGE
    enum: GL_DEPTH_WRITEMASK
    enum: GL_DEPTH_COMPONENT
    enum: GL_LIGHTING
    enum: GL_LIGHT0
    enum: GL_LIGHT1
    enum: GL_LIGHT2
    enum: GL_LIGHT3
    enum: GL_LIGHT4
    enum: GL_LIGHT5
    enum: GL_LIGHT6
    enum: GL_LIGHT7
    enum: GL_SPOT_EXPONENT
    enum: GL_SPOT_CUTOFF
    enum: GL_CONSTANT_ATTENUATION
    enum: GL_LINEAR_ATTENUATION
    enum: GL_QUADRATIC_ATTENUATION
    enum: GL_AMBIENT
    enum: GL_DIFFUSE
    enum: GL_SPECULAR
    enum: GL_SHININESS
    enum: GL_EMISSION
    enum: GL_POSITION
    enum: GL_SPOT_DIRECTION
    enum: GL_AMBIENT_AND_DIFFUSE
    enum: GL_COLOR_INDEXES
    enum: GL_LIGHT_MODEL_TWO_SIDE
    enum: GL_LIGHT_MODEL_LOCAL_VIEWER
    enum: GL_LIGHT_MODEL_AMBIENT
    enum: GL_FRONT_AND_BACK
    enum: GL_SHADE_MODEL
    enum: GL_FLAT
    enum: GL_SMOOTH
    enum: GL_COLOR_MATERIAL
    enum: GL_COLOR_MATERIAL_FACE
    enum: GL_COLOR_MATERIAL_PARAMETER
    enum: GL_NORMALIZE
    enum: GL_CLIP_PLANE0
    enum: GL_CLIP_PLANE1
    enum: GL_CLIP_PLANE2
    enum: GL_CLIP_PLANE3
    enum: GL_CLIP_PLANE4
    enum: GL_CLIP_PLANE5
    enum: GL_ACCUM_RED_BITS
    enum: GL_ACCUM_GREEN_BITS
    enum: GL_ACCUM_BLUE_BITS
    enum: GL_ACCUM_ALPHA_BITS
    enum: GL_ACCUM_CLEAR_VALUE
    enum: GL_ACCUM
    enum: GL_ADD
    enum: GL_LOAD
    enum: GL_MULT
    enum: GL_RETURN
    enum: GL_ALPHA_TEST
    enum: GL_ALPHA_TEST_REF
    enum: GL_ALPHA_TEST_FUNC
    enum: GL_BLEND
    enum: GL_BLEND_SRC
    enum: GL_BLEND_DST
    enum: GL_ZERO
    enum: GL_ONE
    enum: GL_SRC_COLOR
    enum: GL_ONE_MINUS_SRC_COLOR
    enum: GL_SRC_ALPHA
    enum: GL_ONE_MINUS_SRC_ALPHA
    enum: GL_DST_ALPHA
    enum: GL_ONE_MINUS_DST_ALPHA
    enum: GL_DST_COLOR
    enum: GL_ONE_MINUS_DST_COLOR
    enum: GL_SRC_ALPHA_SATURATE
    enum: GL_FEEDBACK
    enum: GL_RENDER
    enum: GL_SELECT
    enum: GL_2D
    enum: GL_3D
    enum: GL_3D_COLOR
    enum: GL_3D_COLOR_TEXTURE
    enum: GL_4D_COLOR_TEXTURE
    enum: GL_POINT_TOKEN
    enum: GL_LINE_TOKEN
    enum: GL_LINE_RESET_TOKEN
    enum: GL_POLYGON_TOKEN
    enum: GL_BITMAP_TOKEN
    enum: GL_DRAW_PIXEL_TOKEN
    enum: GL_COPY_PIXEL_TOKEN
    enum: GL_PASS_THROUGH_TOKEN
    enum: GL_FEEDBACK_BUFFER_POINTER
    enum: GL_FEEDBACK_BUFFER_SIZE
    enum: GL_FEEDBACK_BUFFER_TYPE
    enum: GL_SELECTION_BUFFER_POINTER
    enum: GL_SELECTION_BUFFER_SIZE
    enum: GL_FOG
    enum: GL_FOG_MODE
    enum: GL_FOG_DENSITY
    enum: GL_FOG_COLOR
    enum: GL_FOG_INDEX
    enum: GL_FOG_START
    enum: GL_FOG_END
    enum: GL_LINEAR
    enum: GL_EXP
    enum: GL_EXP2
    enum: GL_LOGIC_OP
    enum: GL_INDEX_LOGIC_OP
    enum: GL_COLOR_LOGIC_OP
    enum: GL_LOGIC_OP_MODE
    enum: GL_CLEAR
    enum: GL_SET
    enum: GL_COPY
    enum: GL_COPY_INVERTED
    enum: GL_NOOP
    enum: GL_INVERT
    enum: GL_AND
    enum: GL_NAND
    enum: GL_OR
    enum: GL_NOR
    enum: GL_XOR
    enum: GL_EQUIV
    enum: GL_AND_REVERSE
    enum: GL_AND_INVERTED
    enum: GL_OR_REVERSE
    enum: GL_OR_INVERTED
    enum: GL_STENCIL_BITS
    enum: GL_STENCIL_TEST
    enum: GL_STENCIL_CLEAR_VALUE
    enum: GL_STENCIL_FUNC
    enum: GL_STENCIL_VALUE_MASK
    enum: GL_STENCIL_FAIL
    enum: GL_STENCIL_PASS_DEPTH_FAIL
    enum: GL_STENCIL_PASS_DEPTH_PASS
    enum: GL_STENCIL_REF
    enum: GL_STENCIL_WRITEMASK
    enum: GL_STENCIL_INDEX
    enum: GL_KEEP
    enum: GL_REPLACE
    enum: GL_INCR
    enum: GL_DECR
    enum: GL_NONE
    enum: GL_LEFT
    enum: GL_RIGHT
    enum: GL_FRONT_LEFT
    enum: GL_FRONT_RIGHT
    enum: GL_BACK_LEFT
    enum: GL_BACK_RIGHT
    enum: GL_AUX0
    enum: GL_AUX1
    enum: GL_AUX2
    enum: GL_AUX3
    enum: GL_COLOR_INDEX
    enum: GL_RED
    enum: GL_GREEN
    enum: GL_BLUE
    enum: GL_ALPHA
    enum: GL_LUMINANCE
    enum: GL_LUMINANCE_ALPHA
    enum: GL_ALPHA_BITS
    enum: GL_RED_BITS
    enum: GL_GREEN_BITS
    enum: GL_BLUE_BITS
    enum: GL_INDEX_BITS
    enum: GL_SUBPIXEL_BITS
    enum: GL_AUX_BUFFERS
    enum: GL_READ_BUFFER
    enum: GL_DRAW_BUFFER
    enum: GL_DOUBLEBUFFER
    enum: GL_STEREO
    enum: GL_BITMAP
    enum: GL_COLOR
    enum: GL_DEPTH
    enum: GL_STENCIL
    enum: GL_DITHER
    enum: GL_RGB
    enum: GL_RGBA
    enum: GL_MAX_LIST_NESTING
    enum: GL_MAX_EVAL_ORDER
    enum: GL_MAX_LIGHTS
    enum: GL_MAX_CLIP_PLANES
    enum: GL_MAX_TEXTURE_SIZE
    enum: GL_MAX_PIXEL_MAP_TABLE
    enum: GL_MAX_ATTRIB_STACK_DEPTH
    enum: GL_MAX_MODELVIEW_STACK_DEPTH
    enum: GL_MAX_NAME_STACK_DEPTH
    enum: GL_MAX_PROJECTION_STACK_DEPTH
    enum: GL_MAX_TEXTURE_STACK_DEPTH
    enum: GL_MAX_VIEWPORT_DIMS
    enum: GL_MAX_CLIENT_ATTRIB_STACK_DEPTH
    enum: GL_ATTRIB_STACK_DEPTH
    enum: GL_CLIENT_ATTRIB_STACK_DEPTH
    enum: GL_COLOR_CLEAR_VALUE
    enum: GL_COLOR_WRITEMASK
    enum: GL_CURRENT_INDEX
    enum: GL_CURRENT_COLOR
    enum: GL_CURRENT_NORMAL
    enum: GL_CURRENT_RASTER_COLOR
    enum: GL_CURRENT_RASTER_DISTANCE
    enum: GL_CURRENT_RASTER_INDEX
    enum: GL_CURRENT_RASTER_POSITION
    enum: GL_CURRENT_RASTER_TEXTURE_COORDS
    enum: GL_CURRENT_RASTER_POSITION_VALID
    enum: GL_CURRENT_TEXTURE_COORDS
    enum: GL_INDEX_CLEAR_VALUE
    enum: GL_INDEX_MODE
    enum: GL_INDEX_WRITEMASK
    enum: GL_MODELVIEW_MATRIX
    enum: GL_MODELVIEW_STACK_DEPTH
    enum: GL_NAME_STACK_DEPTH
    enum: GL_PROJECTION_MATRIX
    enum: GL_PROJECTION_STACK_DEPTH
    enum: GL_RENDER_MODE
    enum: GL_RGBA_MODE
    enum: GL_TEXTURE_MATRIX
    enum: GL_TEXTURE_STACK_DEPTH
    enum: GL_VIEWPORT
    enum: GL_AUTO_NORMAL
    enum: GL_MAP1_COLOR_4
    enum: GL_MAP1_INDEX
    enum: GL_MAP1_NORMAL
    enum: GL_MAP1_TEXTURE_COORD_1
    enum: GL_MAP1_TEXTURE_COORD_2
    enum: GL_MAP1_TEXTURE_COORD_3
    enum: GL_MAP1_TEXTURE_COORD_4
    enum: GL_MAP1_VERTEX_3
    enum: GL_MAP1_VERTEX_4
    enum: GL_MAP2_COLOR_4
    enum: GL_MAP2_INDEX
    enum: GL_MAP2_NORMAL
    enum: GL_MAP2_TEXTURE_COORD_1
    enum: GL_MAP2_TEXTURE_COORD_2
    enum: GL_MAP2_TEXTURE_COORD_3
    enum: GL_MAP2_TEXTURE_COORD_4
    enum: GL_MAP2_VERTEX_3
    enum: GL_MAP2_VERTEX_4
    enum: GL_MAP1_GRID_DOMAIN
    enum: GL_MAP1_GRID_SEGMENTS
    enum: GL_MAP2_GRID_DOMAIN
    enum: GL_MAP2_GRID_SEGMENTS
    enum: GL_COEFF
    enum: GL_ORDER
    enum: GL_DOMAIN
    enum: GL_PERSPECTIVE_CORRECTION_HINT
    enum: GL_POINT_SMOOTH_HINT
    enum: GL_LINE_SMOOTH_HINT
    enum: GL_POLYGON_SMOOTH_HINT
    enum: GL_FOG_HINT
    enum: GL_DONT_CARE
    enum: GL_FASTEST
    enum: GL_NICEST
    enum: GL_SCISSOR_BOX
    enum: GL_SCISSOR_TEST
    enum: GL_MAP_COLOR
    enum: GL_MAP_STENCIL
    enum: GL_INDEX_SHIFT
    enum: GL_INDEX_OFFSET
    enum: GL_RED_SCALE
    enum: GL_RED_BIAS
    enum: GL_GREEN_SCALE
    enum: GL_GREEN_BIAS
    enum: GL_BLUE_SCALE
    enum: GL_BLUE_BIAS
    enum: GL_ALPHA_SCALE
    enum: GL_ALPHA_BIAS
    enum: GL_DEPTH_SCALE
    enum: GL_DEPTH_BIAS
    enum: GL_PIXEL_MAP_S_TO_S_SIZE
    enum: GL_PIXEL_MAP_I_TO_I_SIZE
    enum: GL_PIXEL_MAP_I_TO_R_SIZE
    enum: GL_PIXEL_MAP_I_TO_G_SIZE
    enum: GL_PIXEL_MAP_I_TO_B_SIZE
    enum: GL_PIXEL_MAP_I_TO_A_SIZE
    enum: GL_PIXEL_MAP_R_TO_R_SIZE
    enum: GL_PIXEL_MAP_G_TO_G_SIZE
    enum: GL_PIXEL_MAP_B_TO_B_SIZE
    enum: GL_PIXEL_MAP_A_TO_A_SIZE
    enum: GL_PIXEL_MAP_S_TO_S
    enum: GL_PIXEL_MAP_I_TO_I
    enum: GL_PIXEL_MAP_I_TO_R
    enum: GL_PIXEL_MAP_I_TO_G
    enum: GL_PIXEL_MAP_I_TO_B
    enum: GL_PIXEL_MAP_I_TO_A
    enum: GL_PIXEL_MAP_R_TO_R
    enum: GL_PIXEL_MAP_G_TO_G
    enum: GL_PIXEL_MAP_B_TO_B
    enum: GL_PIXEL_MAP_A_TO_A
    enum: GL_PACK_ALIGNMENT
    enum: GL_PACK_LSB_FIRST
    enum: GL_PACK_ROW_LENGTH
    enum: GL_PACK_SKIP_PIXELS
    enum: GL_PACK_SKIP_ROWS
    enum: GL_PACK_SWAP_BYTES
    enum: GL_UNPACK_ALIGNMENT
    enum: GL_UNPACK_LSB_FIRST
    enum: GL_UNPACK_ROW_LENGTH
    enum: GL_UNPACK_SKIP_PIXELS
    enum: GL_UNPACK_SKIP_ROWS
    enum: GL_UNPACK_SWAP_BYTES
    enum: GL_ZOOM_X
    enum: GL_ZOOM_Y
    enum: GL_TEXTURE_ENV
    enum: GL_TEXTURE_ENV_MODE
    enum: GL_TEXTURE_1D
    enum: GL_TEXTURE_2D
    enum: GL_TEXTURE_WRAP_S
    enum: GL_TEXTURE_WRAP_T
    enum: GL_TEXTURE_MAG_FILTER
    enum: GL_TEXTURE_MIN_FILTER
    enum: GL_TEXTURE_ENV_COLOR
    enum: GL_TEXTURE_GEN_S
    enum: GL_TEXTURE_GEN_T
    enum: GL_TEXTURE_GEN_R
    enum: GL_TEXTURE_GEN_Q
    enum: GL_TEXTURE_GEN_MODE
    enum: GL_TEXTURE_BORDER_COLOR
    enum: GL_TEXTURE_WIDTH
    enum: GL_TEXTURE_HEIGHT
    enum: GL_TEXTURE_BORDER
    enum: GL_TEXTURE_COMPONENTS
    enum: GL_TEXTURE_RED_SIZE
    enum: GL_TEXTURE_GREEN_SIZE
    enum: GL_TEXTURE_BLUE_SIZE
    enum: GL_TEXTURE_ALPHA_SIZE
    enum: GL_TEXTURE_LUMINANCE_SIZE
    enum: GL_TEXTURE_INTENSITY_SIZE
    enum: GL_NEAREST_MIPMAP_NEAREST
    enum: GL_NEAREST_MIPMAP_LINEAR
    enum: GL_LINEAR_MIPMAP_NEAREST
    enum: GL_LINEAR_MIPMAP_LINEAR
    enum: GL_OBJECT_LINEAR
    enum: GL_OBJECT_PLANE
    enum: GL_EYE_LINEAR
    enum: GL_EYE_PLANE
    enum: GL_SPHERE_MAP
    enum: GL_DECAL
    enum: GL_MODULATE
    enum: GL_NEAREST
    enum: GL_REPEAT
    enum: GL_CLAMP
    enum: GL_S
    enum: GL_T
    enum: GL_R
    enum: GL_Q
    enum: GL_VENDOR
    enum: GL_RENDERER
    enum: GL_VERSION
    enum: GL_EXTENSIONS
    enum: GL_NO_ERROR
    enum: GL_INVALID_ENUM
    enum: GL_INVALID_VALUE
    enum: GL_INVALID_OPERATION
    enum: GL_STACK_OVERFLOW
    enum: GL_STACK_UNDERFLOW
    enum: GL_OUT_OF_MEMORY
    enum: GL_CURRENT_BIT
    enum: GL_POINT_BIT
    enum: GL_LINE_BIT
    enum: GL_POLYGON_BIT
    enum: GL_POLYGON_STIPPLE_BIT
    enum: GL_PIXEL_MODE_BIT
    enum: GL_LIGHTING_BIT
    enum: GL_FOG_BIT
    enum: GL_DEPTH_BUFFER_BIT
    enum: GL_ACCUM_BUFFER_BIT
    enum: GL_STENCIL_BUFFER_BIT
    enum: GL_VIEWPORT_BIT
    enum: GL_TRANSFORM_BIT
    enum: GL_ENABLE_BIT
    enum: GL_COLOR_BUFFER_BIT
    enum: GL_HINT_BIT
    enum: GL_EVAL_BIT
    enum: GL_LIST_BIT
    enum: GL_TEXTURE_BIT
    enum: GL_SCISSOR_BIT
    enum: GL_ALL_ATTRIB_BITS
    enum: GL_PROXY_TEXTURE_1D
    enum: GL_PROXY_TEXTURE_2D
    enum: GL_TEXTURE_PRIORITY
    enum: GL_TEXTURE_RESIDENT
    enum: GL_TEXTURE_BINDING_1D
    enum: GL_TEXTURE_BINDING_2D
    enum: GL_TEXTURE_INTERNAL_FORMAT
    enum: GL_ALPHA4
    enum: GL_ALPHA8
    enum: GL_ALPHA12
    enum: GL_ALPHA16
    enum: GL_LUMINANCE4
    enum: GL_LUMINANCE8
    enum: GL_LUMINANCE12
    enum: GL_LUMINANCE16
    enum: GL_LUMINANCE4_ALPHA4
    enum: GL_LUMINANCE6_ALPHA2
    enum: GL_LUMINANCE8_ALPHA8
    enum: GL_LUMINANCE12_ALPHA4
    enum: GL_LUMINANCE12_ALPHA12
    enum: GL_LUMINANCE16_ALPHA16
    enum: GL_INTENSITY
    enum: GL_INTENSITY4
    enum: GL_INTENSITY8
    enum: GL_INTENSITY12
    enum: GL_INTENSITY16
    enum: GL_R3_G3_B2
    enum: GL_RGB4
    enum: GL_RGB5
    enum: GL_RGB8
    enum: GL_RGB10
    enum: GL_RGB12
    enum: GL_RGB16
    enum: GL_RGBA2
    enum: GL_RGBA4
    enum: GL_RGB5_A1
    enum: GL_RGBA8
    enum: GL_RGB10_A2
    enum: GL_RGBA12
    enum: GL_RGBA16
    enum: GL_CLIENT_PIXEL_STORE_BIT
    enum: GL_CLIENT_VERTEX_ARRAY_BIT
    enum: GL_ALL_CLIENT_ATTRIB_BITS
    enum: GL_CLIENT_ALL_ATTRIB_BITS
    enum: GL_RESCALE_NORMAL
    enum: GL_CLAMP_TO_EDGE
    enum: GL_MAX_ELEMENTS_VERTICES
    enum: GL_MAX_ELEMENTS_INDICES
    enum: GL_BGR
    enum: GL_BGRA
    enum: GL_UNSIGNED_BYTE_3_3_2
    enum: GL_UNSIGNED_BYTE_2_3_3_REV
    enum: GL_UNSIGNED_SHORT_5_6_5
    enum: GL_UNSIGNED_SHORT_5_6_5_REV
    enum: GL_UNSIGNED_SHORT_4_4_4_4
    enum: GL_UNSIGNED_SHORT_4_4_4_4_REV
    enum: GL_UNSIGNED_SHORT_5_5_5_1
    enum: GL_UNSIGNED_SHORT_1_5_5_5_REV
    enum: GL_UNSIGNED_INT_8_8_8_8
    enum: GL_UNSIGNED_INT_8_8_8_8_REV
    enum: GL_UNSIGNED_INT_10_10_10_2
    enum: GL_UNSIGNED_INT_2_10_10_10_REV
    enum: GL_LIGHT_MODEL_COLOR_CONTROL
    enum: GL_SINGLE_COLOR
    enum: GL_SEPARATE_SPECULAR_COLOR
    enum: GL_TEXTURE_MIN_LOD
    enum: GL_TEXTURE_MAX_LOD
    enum: GL_TEXTURE_BASE_LEVEL
    enum: GL_TEXTURE_MAX_LEVEL
    enum: GL_SMOOTH_POINT_SIZE_RANGE
    enum: GL_SMOOTH_POINT_SIZE_GRANULARITY
    enum: GL_SMOOTH_LINE_WIDTH_RANGE
    enum: GL_SMOOTH_LINE_WIDTH_GRANULARITY
    enum: GL_ALIASED_POINT_SIZE_RANGE
    enum: GL_ALIASED_LINE_WIDTH_RANGE
    enum: GL_PACK_SKIP_IMAGES
    enum: GL_PACK_IMAGE_HEIGHT
    enum: GL_UNPACK_SKIP_IMAGES
    enum: GL_UNPACK_IMAGE_HEIGHT
    enum: GL_TEXTURE_3D
    enum: GL_PROXY_TEXTURE_3D
    enum: GL_TEXTURE_DEPTH
    enum: GL_TEXTURE_WRAP_R
    enum: GL_MAX_3D_TEXTURE_SIZE
    enum: GL_TEXTURE_BINDING_3D
    enum: GL_CONSTANT_COLOR
    enum: GL_ONE_MINUS_CONSTANT_COLOR
    enum: GL_CONSTANT_ALPHA
    enum: GL_ONE_MINUS_CONSTANT_ALPHA
    enum: GL_COLOR_TABLE
    enum: GL_POST_CONVOLUTION_COLOR_TABLE
    enum: GL_POST_COLOR_MATRIX_COLOR_TABLE
    enum: GL_PROXY_COLOR_TABLE
    enum: GL_PROXY_POST_CONVOLUTION_COLOR_TABLE
    enum: GL_PROXY_POST_COLOR_MATRIX_COLOR_TABLE
    enum: GL_COLOR_TABLE_SCALE
    enum: GL_COLOR_TABLE_BIAS
    enum: GL_COLOR_TABLE_FORMAT
    enum: GL_COLOR_TABLE_WIDTH
    enum: GL_COLOR_TABLE_RED_SIZE
    enum: GL_COLOR_TABLE_GREEN_SIZE
    enum: GL_COLOR_TABLE_BLUE_SIZE
    enum: GL_COLOR_TABLE_ALPHA_SIZE
    enum: GL_COLOR_TABLE_LUMINANCE_SIZE
    enum: GL_COLOR_TABLE_INTENSITY_SIZE
    enum: GL_CONVOLUTION_1D
    enum: GL_CONVOLUTION_2D
    enum: GL_SEPARABLE_2D
    enum: GL_CONVOLUTION_BORDER_MODE
    enum: GL_CONVOLUTION_FILTER_SCALE
    enum: GL_CONVOLUTION_FILTER_BIAS
    enum: GL_REDUCE
    enum: GL_CONVOLUTION_FORMAT
    enum: GL_CONVOLUTION_WIDTH
    enum: GL_CONVOLUTION_HEIGHT
    enum: GL_MAX_CONVOLUTION_WIDTH
    enum: GL_MAX_CONVOLUTION_HEIGHT
    enum: GL_POST_CONVOLUTION_RED_SCALE
    enum: GL_POST_CONVOLUTION_GREEN_SCALE
    enum: GL_POST_CONVOLUTION_BLUE_SCALE
    enum: GL_POST_CONVOLUTION_ALPHA_SCALE
    enum: GL_POST_CONVOLUTION_RED_BIAS
    enum: GL_POST_CONVOLUTION_GREEN_BIAS
    enum: GL_POST_CONVOLUTION_BLUE_BIAS
    enum: GL_POST_CONVOLUTION_ALPHA_BIAS
    enum: GL_CONSTANT_BORDER
    enum: GL_REPLICATE_BORDER
    enum: GL_CONVOLUTION_BORDER_COLOR
    enum: GL_COLOR_MATRIX
    enum: GL_COLOR_MATRIX_STACK_DEPTH
    enum: GL_MAX_COLOR_MATRIX_STACK_DEPTH
    enum: GL_POST_COLOR_MATRIX_RED_SCALE
    enum: GL_POST_COLOR_MATRIX_GREEN_SCALE
    enum: GL_POST_COLOR_MATRIX_BLUE_SCALE
    enum: GL_POST_COLOR_MATRIX_ALPHA_SCALE
    enum: GL_POST_COLOR_MATRIX_RED_BIAS
    enum: GL_POST_COLOR_MATRIX_GREEN_BIAS
    enum: GL_POST_COLOR_MATRIX_BLUE_BIAS
    enum: GL_POST_COLOR_MATRIX_ALPHA_BIAS
    enum: GL_HISTOGRAM
    enum: GL_PROXY_HISTOGRAM
    enum: GL_HISTOGRAM_WIDTH
    enum: GL_HISTOGRAM_FORMAT
    enum: GL_HISTOGRAM_RED_SIZE
    enum: GL_HISTOGRAM_GREEN_SIZE
    enum: GL_HISTOGRAM_BLUE_SIZE
    enum: GL_HISTOGRAM_ALPHA_SIZE
    enum: GL_HISTOGRAM_LUMINANCE_SIZE
    enum: GL_HISTOGRAM_SINK
    enum: GL_MINMAX
    enum: GL_MINMAX_FORMAT
    enum: GL_MINMAX_SINK
    enum: GL_TABLE_TOO_LARGE
    enum: GL_BLEND_EQUATION
    enum: GL_MIN
    enum: GL_MAX
    enum: GL_FUNC_ADD
    enum: GL_FUNC_SUBTRACT
    enum: GL_FUNC_REVERSE_SUBTRACT
    enum: GL_BLEND_COLOR
    enum: GL_TEXTURE0
    enum: GL_TEXTURE1
    enum: GL_TEXTURE2
    enum: GL_TEXTURE3
    enum: GL_TEXTURE4
    enum: GL_TEXTURE5
    enum: GL_TEXTURE6
    enum: GL_TEXTURE7
    enum: GL_TEXTURE8
    enum: GL_TEXTURE9
    enum: GL_TEXTURE10
    enum: GL_TEXTURE11
    enum: GL_TEXTURE12
    enum: GL_TEXTURE13
    enum: GL_TEXTURE14
    enum: GL_TEXTURE15
    enum: GL_TEXTURE16
    enum: GL_TEXTURE17
    enum: GL_TEXTURE18
    enum: GL_TEXTURE19
    enum: GL_TEXTURE20
    enum: GL_TEXTURE21
    enum: GL_TEXTURE22
    enum: GL_TEXTURE23
    enum: GL_TEXTURE24
    enum: GL_TEXTURE25
    enum: GL_TEXTURE26
    enum: GL_TEXTURE27
    enum: GL_TEXTURE28
    enum: GL_TEXTURE29
    enum: GL_TEXTURE30
    enum: GL_TEXTURE31
    enum: GL_ACTIVE_TEXTURE
    enum: GL_CLIENT_ACTIVE_TEXTURE
    enum: GL_MAX_TEXTURE_UNITS
    enum: GL_NORMAL_MAP
    enum: GL_REFLECTION_MAP
    enum: GL_TEXTURE_CUBE_MAP
    enum: GL_TEXTURE_BINDING_CUBE_MAP
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_X
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_X
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_Y
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_Y
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_Z
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
    enum: GL_PROXY_TEXTURE_CUBE_MAP
    enum: GL_MAX_CUBE_MAP_TEXTURE_SIZE
    enum: GL_COMPRESSED_ALPHA
    enum: GL_COMPRESSED_LUMINANCE
    enum: GL_COMPRESSED_LUMINANCE_ALPHA
    enum: GL_COMPRESSED_INTENSITY
    enum: GL_COMPRESSED_RGB
    enum: GL_COMPRESSED_RGBA
    enum: GL_TEXTURE_COMPRESSION_HINT
    enum: GL_TEXTURE_COMPRESSED_IMAGE_SIZE
    enum: GL_TEXTURE_COMPRESSED
    enum: GL_NUM_COMPRESSED_TEXTURE_FORMATS
    enum: GL_COMPRESSED_TEXTURE_FORMATS
    enum: GL_MULTISAMPLE
    enum: GL_SAMPLE_ALPHA_TO_COVERAGE
    enum: GL_SAMPLE_ALPHA_TO_ONE
    enum: GL_SAMPLE_COVERAGE
    enum: GL_SAMPLE_BUFFERS
    enum: GL_SAMPLES
    enum: GL_SAMPLE_COVERAGE_VALUE
    enum: GL_SAMPLE_COVERAGE_INVERT
    enum: GL_MULTISAMPLE_BIT
    enum: GL_TRANSPOSE_MODELVIEW_MATRIX
    enum: GL_TRANSPOSE_PROJECTION_MATRIX
    enum: GL_TRANSPOSE_TEXTURE_MATRIX
    enum: GL_TRANSPOSE_COLOR_MATRIX
    enum: GL_COMBINE
    enum: GL_COMBINE_RGB
    enum: GL_COMBINE_ALPHA
    enum: GL_SOURCE0_RGB
    enum: GL_SOURCE1_RGB
    enum: GL_SOURCE2_RGB
    enum: GL_SOURCE0_ALPHA
    enum: GL_SOURCE1_ALPHA
    enum: GL_SOURCE2_ALPHA
    enum: GL_OPERAND0_RGB
    enum: GL_OPERAND1_RGB
    enum: GL_OPERAND2_RGB
    enum: GL_OPERAND0_ALPHA
    enum: GL_OPERAND1_ALPHA
    enum: GL_OPERAND2_ALPHA
    enum: GL_RGB_SCALE
    enum: GL_ADD_SIGNED
    enum: GL_INTERPOLATE
    enum: GL_SUBTRACT
    enum: GL_CONSTANT
    enum: GL_PRIMARY_COLOR
    enum: GL_PREVIOUS
    enum: GL_DOT3_RGB
    enum: GL_DOT3_RGBA
    enum: GL_CLAMP_TO_BORDER
    enum: GL_ARB_multitexture
    enum: GL_TEXTURE0_ARB
    enum: GL_TEXTURE1_ARB
    enum: GL_TEXTURE2_ARB
    enum: GL_TEXTURE3_ARB
    enum: GL_TEXTURE4_ARB
    enum: GL_TEXTURE5_ARB
    enum: GL_TEXTURE6_ARB
    enum: GL_TEXTURE7_ARB
    enum: GL_TEXTURE8_ARB
    enum: GL_TEXTURE9_ARB
    enum: GL_TEXTURE10_ARB
    enum: GL_TEXTURE11_ARB
    enum: GL_TEXTURE12_ARB
    enum: GL_TEXTURE13_ARB
    enum: GL_TEXTURE14_ARB
    enum: GL_TEXTURE15_ARB
    enum: GL_TEXTURE16_ARB
    enum: GL_TEXTURE17_ARB
    enum: GL_TEXTURE18_ARB
    enum: GL_TEXTURE19_ARB
    enum: GL_TEXTURE20_ARB
    enum: GL_TEXTURE21_ARB
    enum: GL_TEXTURE22_ARB
    enum: GL_TEXTURE23_ARB
    enum: GL_TEXTURE24_ARB
    enum: GL_TEXTURE25_ARB
    enum: GL_TEXTURE26_ARB
    enum: GL_TEXTURE27_ARB
    enum: GL_TEXTURE28_ARB
    enum: GL_TEXTURE29_ARB
    enum: GL_TEXTURE30_ARB
    enum: GL_TEXTURE31_ARB
    enum: GL_ACTIVE_TEXTURE_ARB
    enum: GL_CLIENT_ACTIVE_TEXTURE_ARB
    enum: GL_MAX_TEXTURE_UNITS_ARB
    enum: GL_MESA_shader_debug
    enum: GL_DEBUG_OBJECT_MESA
    enum: GL_DEBUG_PRINT_MESA
    enum: GL_DEBUG_ASSERT_MESA
    enum: GL_MESA_packed_depth_stencil
    enum: GL_DEPTH_STENCIL_MESA
    enum: GL_UNSIGNED_INT_24_8_MESA
    enum: GL_UNSIGNED_INT_8_24_REV_MESA
    enum: GL_UNSIGNED_SHORT_15_1_MESA
    enum: GL_UNSIGNED_SHORT_1_15_REV_MESA
    enum: GL_MESA_program_debug
    enum: GL_FRAGMENT_PROGRAM_POSITION_MESA
    enum: GL_FRAGMENT_PROGRAM_CALLBACK_MESA
    enum: GL_FRAGMENT_PROGRAM_CALLBACK_FUNC_MESA
    enum: GL_FRAGMENT_PROGRAM_CALLBACK_DATA_MESA
    enum: GL_VERTEX_PROGRAM_POSITION_MESA
    enum: GL_VERTEX_PROGRAM_CALLBACK_MESA
    enum: GL_VERTEX_PROGRAM_CALLBACK_FUNC_MESA
    enum: GL_VERTEX_PROGRAM_CALLBACK_DATA_MESA
    enum: GL_MESA_texture_array
    enum: GL_TEXTURE_1D_ARRAY_EXT
    enum: GL_PROXY_TEXTURE_1D_ARRAY_EXT
    enum: GL_TEXTURE_2D_ARRAY_EXT
    enum: GL_PROXY_TEXTURE_2D_ARRAY_EXT
    enum: GL_TEXTURE_BINDING_1D_ARRAY_EXT
    enum: GL_TEXTURE_BINDING_2D_ARRAY_EXT
    enum: GL_MAX_ARRAY_TEXTURE_LAYERS_EXT
    enum: GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER_EXT
    enum: GL_ATI_blend_equation_separate
    enum: GL_ALPHA_BLEND_EQUATION_ATI
    enum: GL_OES_EGL_image


# from /usr/include/GL/glext.h:

cdef extern from 'GL/glext.h':
    enum: GL_GLEXT_VERSION
    enum: GL_UNSIGNED_BYTE_3_3_2
    enum: GL_UNSIGNED_SHORT_4_4_4_4
    enum: GL_UNSIGNED_SHORT_5_5_5_1
    enum: GL_UNSIGNED_INT_8_8_8_8
    enum: GL_UNSIGNED_INT_10_10_10_2
    enum: GL_TEXTURE_BINDING_3D
    enum: GL_PACK_SKIP_IMAGES
    enum: GL_PACK_IMAGE_HEIGHT
    enum: GL_UNPACK_SKIP_IMAGES
    enum: GL_UNPACK_IMAGE_HEIGHT
    enum: GL_TEXTURE_3D
    enum: GL_PROXY_TEXTURE_3D
    enum: GL_TEXTURE_DEPTH
    enum: GL_TEXTURE_WRAP_R
    enum: GL_MAX_3D_TEXTURE_SIZE
    enum: GL_UNSIGNED_BYTE_2_3_3_REV
    enum: GL_UNSIGNED_SHORT_5_6_5
    enum: GL_UNSIGNED_SHORT_5_6_5_REV
    enum: GL_UNSIGNED_SHORT_4_4_4_4_REV
    enum: GL_UNSIGNED_SHORT_1_5_5_5_REV
    enum: GL_UNSIGNED_INT_8_8_8_8_REV
    enum: GL_UNSIGNED_INT_2_10_10_10_REV
    enum: GL_BGR
    enum: GL_BGRA
    enum: GL_MAX_ELEMENTS_VERTICES
    enum: GL_MAX_ELEMENTS_INDICES
    enum: GL_CLAMP_TO_EDGE
    enum: GL_TEXTURE_MIN_LOD
    enum: GL_TEXTURE_MAX_LOD
    enum: GL_TEXTURE_BASE_LEVEL
    enum: GL_TEXTURE_MAX_LEVEL
    enum: GL_SMOOTH_POINT_SIZE_RANGE
    enum: GL_SMOOTH_POINT_SIZE_GRANULARITY
    enum: GL_SMOOTH_LINE_WIDTH_RANGE
    enum: GL_SMOOTH_LINE_WIDTH_GRANULARITY
    enum: GL_ALIASED_LINE_WIDTH_RANGE
    enum: GL_RESCALE_NORMAL
    enum: GL_LIGHT_MODEL_COLOR_CONTROL
    enum: GL_SINGLE_COLOR
    enum: GL_SEPARATE_SPECULAR_COLOR
    enum: GL_ALIASED_POINT_SIZE_RANGE
    enum: GL_CONSTANT_COLOR
    enum: GL_ONE_MINUS_CONSTANT_COLOR
    enum: GL_CONSTANT_ALPHA
    enum: GL_ONE_MINUS_CONSTANT_ALPHA
    enum: GL_BLEND_COLOR
    enum: GL_FUNC_ADD
    enum: GL_MIN
    enum: GL_MAX
    enum: GL_BLEND_EQUATION
    enum: GL_FUNC_SUBTRACT
    enum: GL_FUNC_REVERSE_SUBTRACT
    enum: GL_CONVOLUTION_1D
    enum: GL_CONVOLUTION_2D
    enum: GL_SEPARABLE_2D
    enum: GL_CONVOLUTION_BORDER_MODE
    enum: GL_CONVOLUTION_FILTER_SCALE
    enum: GL_CONVOLUTION_FILTER_BIAS
    enum: GL_REDUCE
    enum: GL_CONVOLUTION_FORMAT
    enum: GL_CONVOLUTION_WIDTH
    enum: GL_CONVOLUTION_HEIGHT
    enum: GL_MAX_CONVOLUTION_WIDTH
    enum: GL_MAX_CONVOLUTION_HEIGHT
    enum: GL_POST_CONVOLUTION_RED_SCALE
    enum: GL_POST_CONVOLUTION_GREEN_SCALE
    enum: GL_POST_CONVOLUTION_BLUE_SCALE
    enum: GL_POST_CONVOLUTION_ALPHA_SCALE
    enum: GL_POST_CONVOLUTION_RED_BIAS
    enum: GL_POST_CONVOLUTION_GREEN_BIAS
    enum: GL_POST_CONVOLUTION_BLUE_BIAS
    enum: GL_POST_CONVOLUTION_ALPHA_BIAS
    enum: GL_HISTOGRAM
    enum: GL_PROXY_HISTOGRAM
    enum: GL_HISTOGRAM_WIDTH
    enum: GL_HISTOGRAM_FORMAT
    enum: GL_HISTOGRAM_RED_SIZE
    enum: GL_HISTOGRAM_GREEN_SIZE
    enum: GL_HISTOGRAM_BLUE_SIZE
    enum: GL_HISTOGRAM_ALPHA_SIZE
    enum: GL_HISTOGRAM_LUMINANCE_SIZE
    enum: GL_HISTOGRAM_SINK
    enum: GL_MINMAX
    enum: GL_MINMAX_FORMAT
    enum: GL_MINMAX_SINK
    enum: GL_TABLE_TOO_LARGE
    enum: GL_COLOR_MATRIX
    enum: GL_COLOR_MATRIX_STACK_DEPTH
    enum: GL_MAX_COLOR_MATRIX_STACK_DEPTH
    enum: GL_POST_COLOR_MATRIX_RED_SCALE
    enum: GL_POST_COLOR_MATRIX_GREEN_SCALE
    enum: GL_POST_COLOR_MATRIX_BLUE_SCALE
    enum: GL_POST_COLOR_MATRIX_ALPHA_SCALE
    enum: GL_POST_COLOR_MATRIX_RED_BIAS
    enum: GL_POST_COLOR_MATRIX_GREEN_BIAS
    enum: GL_POST_COLOR_MATRIX_BLUE_BIAS
    enum: GL_POST_COLOR_MATRIX_ALPHA_BIAS
    enum: GL_COLOR_TABLE
    enum: GL_POST_CONVOLUTION_COLOR_TABLE
    enum: GL_POST_COLOR_MATRIX_COLOR_TABLE
    enum: GL_PROXY_COLOR_TABLE
    enum: GL_PROXY_POST_CONVOLUTION_COLOR_TABLE
    enum: GL_PROXY_POST_COLOR_MATRIX_COLOR_TABLE
    enum: GL_COLOR_TABLE_SCALE
    enum: GL_COLOR_TABLE_BIAS
    enum: GL_COLOR_TABLE_FORMAT
    enum: GL_COLOR_TABLE_WIDTH
    enum: GL_COLOR_TABLE_RED_SIZE
    enum: GL_COLOR_TABLE_GREEN_SIZE
    enum: GL_COLOR_TABLE_BLUE_SIZE
    enum: GL_COLOR_TABLE_ALPHA_SIZE
    enum: GL_COLOR_TABLE_LUMINANCE_SIZE
    enum: GL_COLOR_TABLE_INTENSITY_SIZE
    enum: GL_CONSTANT_BORDER
    enum: GL_REPLICATE_BORDER
    enum: GL_CONVOLUTION_BORDER_COLOR
    enum: GL_TEXTURE0
    enum: GL_TEXTURE1
    enum: GL_TEXTURE2
    enum: GL_TEXTURE3
    enum: GL_TEXTURE4
    enum: GL_TEXTURE5
    enum: GL_TEXTURE6
    enum: GL_TEXTURE7
    enum: GL_TEXTURE8
    enum: GL_TEXTURE9
    enum: GL_TEXTURE10
    enum: GL_TEXTURE11
    enum: GL_TEXTURE12
    enum: GL_TEXTURE13
    enum: GL_TEXTURE14
    enum: GL_TEXTURE15
    enum: GL_TEXTURE16
    enum: GL_TEXTURE17
    enum: GL_TEXTURE18
    enum: GL_TEXTURE19
    enum: GL_TEXTURE20
    enum: GL_TEXTURE21
    enum: GL_TEXTURE22
    enum: GL_TEXTURE23
    enum: GL_TEXTURE24
    enum: GL_TEXTURE25
    enum: GL_TEXTURE26
    enum: GL_TEXTURE27
    enum: GL_TEXTURE28
    enum: GL_TEXTURE29
    enum: GL_TEXTURE30
    enum: GL_TEXTURE31
    enum: GL_ACTIVE_TEXTURE
    enum: GL_MULTISAMPLE
    enum: GL_SAMPLE_ALPHA_TO_COVERAGE
    enum: GL_SAMPLE_ALPHA_TO_ONE
    enum: GL_SAMPLE_COVERAGE
    enum: GL_SAMPLE_BUFFERS
    enum: GL_SAMPLES
    enum: GL_SAMPLE_COVERAGE_VALUE
    enum: GL_SAMPLE_COVERAGE_INVERT
    enum: GL_TEXTURE_CUBE_MAP
    enum: GL_TEXTURE_BINDING_CUBE_MAP
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_X
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_X
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_Y
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_Y
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_Z
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
    enum: GL_PROXY_TEXTURE_CUBE_MAP
    enum: GL_MAX_CUBE_MAP_TEXTURE_SIZE
    enum: GL_COMPRESSED_RGB
    enum: GL_COMPRESSED_RGBA
    enum: GL_TEXTURE_COMPRESSION_HINT
    enum: GL_TEXTURE_COMPRESSED_IMAGE_SIZE
    enum: GL_TEXTURE_COMPRESSED
    enum: GL_NUM_COMPRESSED_TEXTURE_FORMATS
    enum: GL_COMPRESSED_TEXTURE_FORMATS
    enum: GL_CLAMP_TO_BORDER
    enum: GL_CLIENT_ACTIVE_TEXTURE
    enum: GL_MAX_TEXTURE_UNITS
    enum: GL_TRANSPOSE_MODELVIEW_MATRIX
    enum: GL_TRANSPOSE_PROJECTION_MATRIX
    enum: GL_TRANSPOSE_TEXTURE_MATRIX
    enum: GL_TRANSPOSE_COLOR_MATRIX
    enum: GL_MULTISAMPLE_BIT
    enum: GL_NORMAL_MAP
    enum: GL_REFLECTION_MAP
    enum: GL_COMPRESSED_ALPHA
    enum: GL_COMPRESSED_LUMINANCE
    enum: GL_COMPRESSED_LUMINANCE_ALPHA
    enum: GL_COMPRESSED_INTENSITY
    enum: GL_COMBINE
    enum: GL_COMBINE_RGB
    enum: GL_COMBINE_ALPHA
    enum: GL_SOURCE0_RGB
    enum: GL_SOURCE1_RGB
    enum: GL_SOURCE2_RGB
    enum: GL_SOURCE0_ALPHA
    enum: GL_SOURCE1_ALPHA
    enum: GL_SOURCE2_ALPHA
    enum: GL_OPERAND0_RGB
    enum: GL_OPERAND1_RGB
    enum: GL_OPERAND2_RGB
    enum: GL_OPERAND0_ALPHA
    enum: GL_OPERAND1_ALPHA
    enum: GL_OPERAND2_ALPHA
    enum: GL_RGB_SCALE
    enum: GL_ADD_SIGNED
    enum: GL_INTERPOLATE
    enum: GL_SUBTRACT
    enum: GL_CONSTANT
    enum: GL_PRIMARY_COLOR
    enum: GL_PREVIOUS
    enum: GL_DOT3_RGB
    enum: GL_DOT3_RGBA
    enum: GL_BLEND_DST_RGB
    enum: GL_BLEND_SRC_RGB
    enum: GL_BLEND_DST_ALPHA
    enum: GL_BLEND_SRC_ALPHA
    enum: GL_POINT_FADE_THRESHOLD_SIZE
    enum: GL_DEPTH_COMPONENT16
    enum: GL_DEPTH_COMPONENT24
    enum: GL_DEPTH_COMPONENT32
    enum: GL_MIRRORED_REPEAT
    enum: GL_MAX_TEXTURE_LOD_BIAS
    enum: GL_TEXTURE_LOD_BIAS
    enum: GL_INCR_WRAP
    enum: GL_DECR_WRAP
    enum: GL_TEXTURE_DEPTH_SIZE
    enum: GL_TEXTURE_COMPARE_MODE
    enum: GL_TEXTURE_COMPARE_FUNC
    enum: GL_POINT_SIZE_MIN
    enum: GL_POINT_SIZE_MAX
    enum: GL_POINT_DISTANCE_ATTENUATION
    enum: GL_GENERATE_MIPMAP
    enum: GL_GENERATE_MIPMAP_HINT
    enum: GL_FOG_COORDINATE_SOURCE
    enum: GL_FOG_COORDINATE
    enum: GL_FRAGMENT_DEPTH
    enum: GL_CURRENT_FOG_COORDINATE
    enum: GL_FOG_COORDINATE_ARRAY_TYPE
    enum: GL_FOG_COORDINATE_ARRAY_STRIDE
    enum: GL_FOG_COORDINATE_ARRAY_POINTER
    enum: GL_FOG_COORDINATE_ARRAY
    enum: GL_COLOR_SUM
    enum: GL_CURRENT_SECONDARY_COLOR
    enum: GL_SECONDARY_COLOR_ARRAY_SIZE
    enum: GL_SECONDARY_COLOR_ARRAY_TYPE
    enum: GL_SECONDARY_COLOR_ARRAY_STRIDE
    enum: GL_SECONDARY_COLOR_ARRAY_POINTER
    enum: GL_SECONDARY_COLOR_ARRAY
    enum: GL_TEXTURE_FILTER_CONTROL
    enum: GL_DEPTH_TEXTURE_MODE
    enum: GL_COMPARE_R_TO_TEXTURE
    enum: GL_BUFFER_SIZE
    enum: GL_BUFFER_USAGE
    enum: GL_QUERY_COUNTER_BITS
    enum: GL_CURRENT_QUERY
    enum: GL_QUERY_RESULT
    enum: GL_QUERY_RESULT_AVAILABLE
    enum: GL_ARRAY_BUFFER
    enum: GL_ELEMENT_ARRAY_BUFFER
    enum: GL_ARRAY_BUFFER_BINDING
    enum: GL_ELEMENT_ARRAY_BUFFER_BINDING
    enum: GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING
    enum: GL_READ_ONLY
    enum: GL_WRITE_ONLY
    enum: GL_READ_WRITE
    enum: GL_BUFFER_ACCESS
    enum: GL_BUFFER_MAPPED
    enum: GL_BUFFER_MAP_POINTER
    enum: GL_STREAM_DRAW
    enum: GL_STREAM_READ
    enum: GL_STREAM_COPY
    enum: GL_STATIC_DRAW
    enum: GL_STATIC_READ
    enum: GL_STATIC_COPY
    enum: GL_DYNAMIC_DRAW
    enum: GL_DYNAMIC_READ
    enum: GL_DYNAMIC_COPY
    enum: GL_SAMPLES_PASSED
    enum: GL_SRC1_ALPHA
    enum: GL_VERTEX_ARRAY_BUFFER_BINDING
    enum: GL_NORMAL_ARRAY_BUFFER_BINDING
    enum: GL_COLOR_ARRAY_BUFFER_BINDING
    enum: GL_INDEX_ARRAY_BUFFER_BINDING
    enum: GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING
    enum: GL_EDGE_FLAG_ARRAY_BUFFER_BINDING
    enum: GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING
    enum: GL_FOG_COORDINATE_ARRAY_BUFFER_BINDING
    enum: GL_WEIGHT_ARRAY_BUFFER_BINDING
    enum: GL_FOG_COORD_SRC
    enum: GL_FOG_COORD
    enum: GL_CURRENT_FOG_COORD
    enum: GL_FOG_COORD_ARRAY_TYPE
    enum: GL_FOG_COORD_ARRAY_STRIDE
    enum: GL_FOG_COORD_ARRAY_POINTER
    enum: GL_FOG_COORD_ARRAY
    enum: GL_FOG_COORD_ARRAY_BUFFER_BINDING
    enum: GL_SRC0_RGB
    enum: GL_SRC1_RGB
    enum: GL_SRC2_RGB
    enum: GL_SRC0_ALPHA
    enum: GL_SRC2_ALPHA
    enum: GL_BLEND_EQUATION_RGB
    enum: GL_VERTEX_ATTRIB_ARRAY_ENABLED
    enum: GL_VERTEX_ATTRIB_ARRAY_SIZE
    enum: GL_VERTEX_ATTRIB_ARRAY_STRIDE
    enum: GL_VERTEX_ATTRIB_ARRAY_TYPE
    enum: GL_CURRENT_VERTEX_ATTRIB
    enum: GL_VERTEX_PROGRAM_POINT_SIZE
    enum: GL_VERTEX_ATTRIB_ARRAY_POINTER
    enum: GL_STENCIL_BACK_FUNC
    enum: GL_STENCIL_BACK_FAIL
    enum: GL_STENCIL_BACK_PASS_DEPTH_FAIL
    enum: GL_STENCIL_BACK_PASS_DEPTH_PASS
    enum: GL_MAX_DRAW_BUFFERS
    enum: GL_DRAW_BUFFER0
    enum: GL_DRAW_BUFFER1
    enum: GL_DRAW_BUFFER2
    enum: GL_DRAW_BUFFER3
    enum: GL_DRAW_BUFFER4
    enum: GL_DRAW_BUFFER5
    enum: GL_DRAW_BUFFER6
    enum: GL_DRAW_BUFFER7
    enum: GL_DRAW_BUFFER8
    enum: GL_DRAW_BUFFER9
    enum: GL_DRAW_BUFFER10
    enum: GL_DRAW_BUFFER11
    enum: GL_DRAW_BUFFER12
    enum: GL_DRAW_BUFFER13
    enum: GL_DRAW_BUFFER14
    enum: GL_DRAW_BUFFER15
    enum: GL_BLEND_EQUATION_ALPHA
    enum: GL_MAX_VERTEX_ATTRIBS
    enum: GL_VERTEX_ATTRIB_ARRAY_NORMALIZED
    enum: GL_MAX_TEXTURE_IMAGE_UNITS
    enum: GL_FRAGMENT_SHADER
    enum: GL_VERTEX_SHADER
    enum: GL_MAX_FRAGMENT_UNIFORM_COMPONENTS
    enum: GL_MAX_VERTEX_UNIFORM_COMPONENTS
    enum: GL_MAX_VARYING_FLOATS
    enum: GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS
    enum: GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS
    enum: GL_SHADER_TYPE
    enum: GL_FLOAT_VEC2
    enum: GL_FLOAT_VEC3
    enum: GL_FLOAT_VEC4
    enum: GL_INT_VEC2
    enum: GL_INT_VEC3
    enum: GL_INT_VEC4
    enum: GL_BOOL
    enum: GL_BOOL_VEC2
    enum: GL_BOOL_VEC3
    enum: GL_BOOL_VEC4
    enum: GL_FLOAT_MAT2
    enum: GL_FLOAT_MAT3
    enum: GL_FLOAT_MAT4
    enum: GL_SAMPLER_1D
    enum: GL_SAMPLER_2D
    enum: GL_SAMPLER_3D
    enum: GL_SAMPLER_CUBE
    enum: GL_SAMPLER_1D_SHADOW
    enum: GL_SAMPLER_2D_SHADOW
    enum: GL_DELETE_STATUS
    enum: GL_COMPILE_STATUS
    enum: GL_LINK_STATUS
    enum: GL_VALIDATE_STATUS
    enum: GL_INFO_LOG_LENGTH
    enum: GL_ATTACHED_SHADERS
    enum: GL_ACTIVE_UNIFORMS
    enum: GL_ACTIVE_UNIFORM_MAX_LENGTH
    enum: GL_SHADER_SOURCE_LENGTH
    enum: GL_ACTIVE_ATTRIBUTES
    enum: GL_ACTIVE_ATTRIBUTE_MAX_LENGTH
    enum: GL_FRAGMENT_SHADER_DERIVATIVE_HINT
    enum: GL_SHADING_LANGUAGE_VERSION
    enum: GL_CURRENT_PROGRAM
    enum: GL_POINT_SPRITE_COORD_ORIGIN
    enum: GL_LOWER_LEFT
    enum: GL_UPPER_LEFT
    enum: GL_STENCIL_BACK_REF
    enum: GL_STENCIL_BACK_VALUE_MASK
    enum: GL_STENCIL_BACK_WRITEMASK
    enum: GL_VERTEX_PROGRAM_TWO_SIDE
    enum: GL_POINT_SPRITE
    enum: GL_COORD_REPLACE
    enum: GL_MAX_TEXTURE_COORDS
    enum: GL_PIXEL_PACK_BUFFER
    enum: GL_PIXEL_UNPACK_BUFFER
    enum: GL_PIXEL_PACK_BUFFER_BINDING
    enum: GL_PIXEL_UNPACK_BUFFER_BINDING
    enum: GL_FLOAT_MAT2x3
    enum: GL_FLOAT_MAT2x4
    enum: GL_FLOAT_MAT3x2
    enum: GL_FLOAT_MAT3x4
    enum: GL_FLOAT_MAT4x2
    enum: GL_FLOAT_MAT4x3
    enum: GL_SRGB
    enum: GL_SRGB8
    enum: GL_SRGB_ALPHA
    enum: GL_SRGB8_ALPHA8
    enum: GL_COMPRESSED_SRGB
    enum: GL_COMPRESSED_SRGB_ALPHA
    enum: GL_CURRENT_RASTER_SECONDARY_COLOR
    enum: GL_SLUMINANCE_ALPHA
    enum: GL_SLUMINANCE8_ALPHA8
    enum: GL_SLUMINANCE
    enum: GL_SLUMINANCE8
    enum: GL_COMPRESSED_SLUMINANCE
    enum: GL_COMPRESSED_SLUMINANCE_ALPHA
    enum: GL_COMPARE_REF_TO_TEXTURE
    enum: GL_CLIP_DISTANCE0
    enum: GL_CLIP_DISTANCE1
    enum: GL_CLIP_DISTANCE2
    enum: GL_CLIP_DISTANCE3
    enum: GL_CLIP_DISTANCE4
    enum: GL_CLIP_DISTANCE5
    enum: GL_CLIP_DISTANCE6
    enum: GL_CLIP_DISTANCE7
    enum: GL_MAX_CLIP_DISTANCES
    enum: GL_MAJOR_VERSION
    enum: GL_MINOR_VERSION
    enum: GL_NUM_EXTENSIONS
    enum: GL_CONTEXT_FLAGS
    enum: GL_COMPRESSED_RED
    enum: GL_COMPRESSED_RG
    enum: GL_CONTEXT_FLAG_FORWARD_COMPATIBLE_BIT
    enum: GL_RGBA32F
    enum: GL_RGB32F
    enum: GL_RGBA16F
    enum: GL_RGB16F
    enum: GL_VERTEX_ATTRIB_ARRAY_INTEGER
    enum: GL_MAX_ARRAY_TEXTURE_LAYERS
    enum: GL_MIN_PROGRAM_TEXEL_OFFSET
    enum: GL_MAX_PROGRAM_TEXEL_OFFSET
    enum: GL_CLAMP_READ_COLOR
    enum: GL_FIXED_ONLY
    enum: GL_MAX_VARYING_COMPONENTS
    enum: GL_TEXTURE_1D_ARRAY
    enum: GL_PROXY_TEXTURE_1D_ARRAY
    enum: GL_TEXTURE_2D_ARRAY
    enum: GL_PROXY_TEXTURE_2D_ARRAY
    enum: GL_TEXTURE_BINDING_1D_ARRAY
    enum: GL_TEXTURE_BINDING_2D_ARRAY
    enum: GL_R11F_G11F_B10F
    enum: GL_UNSIGNED_INT_10F_11F_11F_REV
    enum: GL_RGB9_E5
    enum: GL_UNSIGNED_INT_5_9_9_9_REV
    enum: GL_TEXTURE_SHARED_SIZE
    enum: GL_TRANSFORM_FEEDBACK_VARYING_MAX_LENGTH
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_MODE
    enum: GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS
    enum: GL_TRANSFORM_FEEDBACK_VARYINGS
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_START
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_SIZE
    enum: GL_PRIMITIVES_GENERATED
    enum: GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN
    enum: GL_RASTERIZER_DISCARD
    enum: GL_MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS
    enum: GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS
    enum: GL_INTERLEAVED_ATTRIBS
    enum: GL_SEPARATE_ATTRIBS
    enum: GL_TRANSFORM_FEEDBACK_BUFFER
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_BINDING
    enum: GL_RGBA32UI
    enum: GL_RGB32UI
    enum: GL_RGBA16UI
    enum: GL_RGB16UI
    enum: GL_RGBA8UI
    enum: GL_RGB8UI
    enum: GL_RGBA32I
    enum: GL_RGB32I
    enum: GL_RGBA16I
    enum: GL_RGB16I
    enum: GL_RGBA8I
    enum: GL_RGB8I
    enum: GL_RED_INTEGER
    enum: GL_GREEN_INTEGER
    enum: GL_BLUE_INTEGER
    enum: GL_RGB_INTEGER
    enum: GL_RGBA_INTEGER
    enum: GL_BGR_INTEGER
    enum: GL_BGRA_INTEGER
    enum: GL_SAMPLER_1D_ARRAY
    enum: GL_SAMPLER_2D_ARRAY
    enum: GL_SAMPLER_1D_ARRAY_SHADOW
    enum: GL_SAMPLER_2D_ARRAY_SHADOW
    enum: GL_SAMPLER_CUBE_SHADOW
    enum: GL_UNSIGNED_INT_VEC2
    enum: GL_UNSIGNED_INT_VEC3
    enum: GL_UNSIGNED_INT_VEC4
    enum: GL_INT_SAMPLER_1D
    enum: GL_INT_SAMPLER_2D
    enum: GL_INT_SAMPLER_3D
    enum: GL_INT_SAMPLER_CUBE
    enum: GL_INT_SAMPLER_1D_ARRAY
    enum: GL_INT_SAMPLER_2D_ARRAY
    enum: GL_UNSIGNED_INT_SAMPLER_1D
    enum: GL_UNSIGNED_INT_SAMPLER_2D
    enum: GL_UNSIGNED_INT_SAMPLER_3D
    enum: GL_UNSIGNED_INT_SAMPLER_CUBE
    enum: GL_UNSIGNED_INT_SAMPLER_1D_ARRAY
    enum: GL_UNSIGNED_INT_SAMPLER_2D_ARRAY
    enum: GL_QUERY_WAIT
    enum: GL_QUERY_NO_WAIT
    enum: GL_QUERY_BY_REGION_WAIT
    enum: GL_QUERY_BY_REGION_NO_WAIT
    enum: GL_BUFFER_ACCESS_FLAGS
    enum: GL_BUFFER_MAP_LENGTH
    enum: GL_BUFFER_MAP_OFFSET
    enum: GL_CLAMP_VERTEX_COLOR
    enum: GL_CLAMP_FRAGMENT_COLOR
    enum: GL_ALPHA_INTEGER
    enum: GL_SAMPLER_2D_RECT
    enum: GL_SAMPLER_2D_RECT_SHADOW
    enum: GL_SAMPLER_BUFFER
    enum: GL_INT_SAMPLER_2D_RECT
    enum: GL_INT_SAMPLER_BUFFER
    enum: GL_UNSIGNED_INT_SAMPLER_2D_RECT
    enum: GL_UNSIGNED_INT_SAMPLER_BUFFER
    enum: GL_TEXTURE_BUFFER
    enum: GL_MAX_TEXTURE_BUFFER_SIZE
    enum: GL_TEXTURE_BINDING_BUFFER
    enum: GL_TEXTURE_BUFFER_DATA_STORE_BINDING
    enum: GL_TEXTURE_RECTANGLE
    enum: GL_TEXTURE_BINDING_RECTANGLE
    enum: GL_PROXY_TEXTURE_RECTANGLE
    enum: GL_MAX_RECTANGLE_TEXTURE_SIZE
    enum: GL_RED_SNORM
    enum: GL_RG_SNORM
    enum: GL_RGB_SNORM
    enum: GL_RGBA_SNORM
    enum: GL_R8_SNORM
    enum: GL_RG8_SNORM
    enum: GL_RGB8_SNORM
    enum: GL_RGBA8_SNORM
    enum: GL_R16_SNORM
    enum: GL_RG16_SNORM
    enum: GL_RGB16_SNORM
    enum: GL_RGBA16_SNORM
    enum: GL_SIGNED_NORMALIZED
    enum: GL_PRIMITIVE_RESTART
    enum: GL_PRIMITIVE_RESTART_INDEX
    enum: GL_CONTEXT_CORE_PROFILE_BIT
    enum: GL_CONTEXT_COMPATIBILITY_PROFILE_BIT
    enum: GL_LINES_ADJACENCY
    enum: GL_LINE_STRIP_ADJACENCY
    enum: GL_TRIANGLES_ADJACENCY
    enum: GL_TRIANGLE_STRIP_ADJACENCY
    enum: GL_PROGRAM_POINT_SIZE
    enum: GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS
    enum: GL_FRAMEBUFFER_ATTACHMENT_LAYERED
    enum: GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS
    enum: GL_GEOMETRY_SHADER
    enum: GL_GEOMETRY_VERTICES_OUT
    enum: GL_GEOMETRY_INPUT_TYPE
    enum: GL_GEOMETRY_OUTPUT_TYPE
    enum: GL_MAX_GEOMETRY_UNIFORM_COMPONENTS
    enum: GL_MAX_GEOMETRY_OUTPUT_VERTICES
    enum: GL_MAX_GEOMETRY_TOTAL_OUTPUT_COMPONENTS
    enum: GL_MAX_VERTEX_OUTPUT_COMPONENTS
    enum: GL_MAX_GEOMETRY_INPUT_COMPONENTS
    enum: GL_MAX_GEOMETRY_OUTPUT_COMPONENTS
    enum: GL_MAX_FRAGMENT_INPUT_COMPONENTS
    enum: GL_CONTEXT_PROFILE_MASK
    enum: GL_VERTEX_ATTRIB_ARRAY_DIVISOR
    enum: GL_SAMPLE_SHADING
    enum: GL_MIN_SAMPLE_SHADING_VALUE
    enum: GL_MIN_PROGRAM_TEXTURE_GATHER_OFFSET
    enum: GL_MAX_PROGRAM_TEXTURE_GATHER_OFFSET
    enum: GL_TEXTURE_CUBE_MAP_ARRAY
    enum: GL_TEXTURE_BINDING_CUBE_MAP_ARRAY
    enum: GL_PROXY_TEXTURE_CUBE_MAP_ARRAY
    enum: GL_SAMPLER_CUBE_MAP_ARRAY
    enum: GL_SAMPLER_CUBE_MAP_ARRAY_SHADOW
    enum: GL_INT_SAMPLER_CUBE_MAP_ARRAY
    enum: GL_UNSIGNED_INT_SAMPLER_CUBE_MAP_ARRAY
    enum: GL_NUM_SHADING_LANGUAGE_VERSIONS
    enum: GL_VERTEX_ATTRIB_ARRAY_LONG
    enum: GL_TEXTURE0_ARB
    enum: GL_TEXTURE1_ARB
    enum: GL_TEXTURE2_ARB
    enum: GL_TEXTURE3_ARB
    enum: GL_TEXTURE4_ARB
    enum: GL_TEXTURE5_ARB
    enum: GL_TEXTURE6_ARB
    enum: GL_TEXTURE7_ARB
    enum: GL_TEXTURE8_ARB
    enum: GL_TEXTURE9_ARB
    enum: GL_TEXTURE10_ARB
    enum: GL_TEXTURE11_ARB
    enum: GL_TEXTURE12_ARB
    enum: GL_TEXTURE13_ARB
    enum: GL_TEXTURE14_ARB
    enum: GL_TEXTURE15_ARB
    enum: GL_TEXTURE16_ARB
    enum: GL_TEXTURE17_ARB
    enum: GL_TEXTURE18_ARB
    enum: GL_TEXTURE19_ARB
    enum: GL_TEXTURE20_ARB
    enum: GL_TEXTURE21_ARB
    enum: GL_TEXTURE22_ARB
    enum: GL_TEXTURE23_ARB
    enum: GL_TEXTURE24_ARB
    enum: GL_TEXTURE25_ARB
    enum: GL_TEXTURE26_ARB
    enum: GL_TEXTURE27_ARB
    enum: GL_TEXTURE28_ARB
    enum: GL_TEXTURE29_ARB
    enum: GL_TEXTURE30_ARB
    enum: GL_TEXTURE31_ARB
    enum: GL_ACTIVE_TEXTURE_ARB
    enum: GL_CLIENT_ACTIVE_TEXTURE_ARB
    enum: GL_MAX_TEXTURE_UNITS_ARB
    enum: GL_TRANSPOSE_MODELVIEW_MATRIX_ARB
    enum: GL_TRANSPOSE_PROJECTION_MATRIX_ARB
    enum: GL_TRANSPOSE_TEXTURE_MATRIX_ARB
    enum: GL_TRANSPOSE_COLOR_MATRIX_ARB
    enum: GL_MULTISAMPLE_ARB
    enum: GL_SAMPLE_ALPHA_TO_COVERAGE_ARB
    enum: GL_SAMPLE_ALPHA_TO_ONE_ARB
    enum: GL_SAMPLE_COVERAGE_ARB
    enum: GL_SAMPLE_BUFFERS_ARB
    enum: GL_SAMPLES_ARB
    enum: GL_SAMPLE_COVERAGE_VALUE_ARB
    enum: GL_SAMPLE_COVERAGE_INVERT_ARB
    enum: GL_MULTISAMPLE_BIT_ARB
    enum: GL_NORMAL_MAP_ARB
    enum: GL_REFLECTION_MAP_ARB
    enum: GL_TEXTURE_CUBE_MAP_ARB
    enum: GL_TEXTURE_BINDING_CUBE_MAP_ARB
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB
    enum: GL_PROXY_TEXTURE_CUBE_MAP_ARB
    enum: GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB
    enum: GL_COMPRESSED_ALPHA_ARB
    enum: GL_COMPRESSED_LUMINANCE_ARB
    enum: GL_COMPRESSED_LUMINANCE_ALPHA_ARB
    enum: GL_COMPRESSED_INTENSITY_ARB
    enum: GL_COMPRESSED_RGB_ARB
    enum: GL_COMPRESSED_RGBA_ARB
    enum: GL_TEXTURE_COMPRESSION_HINT_ARB
    enum: GL_TEXTURE_COMPRESSED_IMAGE_SIZE_ARB
    enum: GL_TEXTURE_COMPRESSED_ARB
    enum: GL_NUM_COMPRESSED_TEXTURE_FORMATS_ARB
    enum: GL_COMPRESSED_TEXTURE_FORMATS_ARB
    enum: GL_CLAMP_TO_BORDER_ARB
    enum: GL_POINT_SIZE_MIN_ARB
    enum: GL_POINT_SIZE_MAX_ARB
    enum: GL_POINT_FADE_THRESHOLD_SIZE_ARB
    enum: GL_POINT_DISTANCE_ATTENUATION_ARB
    enum: GL_MAX_VERTEX_UNITS_ARB
    enum: GL_ACTIVE_VERTEX_UNITS_ARB
    enum: GL_WEIGHT_SUM_UNITY_ARB
    enum: GL_VERTEX_BLEND_ARB
    enum: GL_CURRENT_WEIGHT_ARB
    enum: GL_WEIGHT_ARRAY_TYPE_ARB
    enum: GL_WEIGHT_ARRAY_STRIDE_ARB
    enum: GL_WEIGHT_ARRAY_SIZE_ARB
    enum: GL_WEIGHT_ARRAY_POINTER_ARB
    enum: GL_WEIGHT_ARRAY_ARB
    enum: GL_MODELVIEW0_ARB
    enum: GL_MODELVIEW1_ARB
    enum: GL_MODELVIEW2_ARB
    enum: GL_MODELVIEW3_ARB
    enum: GL_MODELVIEW4_ARB
    enum: GL_MODELVIEW5_ARB
    enum: GL_MODELVIEW6_ARB
    enum: GL_MODELVIEW7_ARB
    enum: GL_MODELVIEW8_ARB
    enum: GL_MODELVIEW9_ARB
    enum: GL_MODELVIEW10_ARB
    enum: GL_MODELVIEW11_ARB
    enum: GL_MODELVIEW12_ARB
    enum: GL_MODELVIEW13_ARB
    enum: GL_MODELVIEW14_ARB
    enum: GL_MODELVIEW15_ARB
    enum: GL_MODELVIEW16_ARB
    enum: GL_MODELVIEW17_ARB
    enum: GL_MODELVIEW18_ARB
    enum: GL_MODELVIEW19_ARB
    enum: GL_MODELVIEW20_ARB
    enum: GL_MODELVIEW21_ARB
    enum: GL_MODELVIEW22_ARB
    enum: GL_MODELVIEW23_ARB
    enum: GL_MODELVIEW24_ARB
    enum: GL_MODELVIEW25_ARB
    enum: GL_MODELVIEW26_ARB
    enum: GL_MODELVIEW27_ARB
    enum: GL_MODELVIEW28_ARB
    enum: GL_MODELVIEW29_ARB
    enum: GL_MODELVIEW30_ARB
    enum: GL_MODELVIEW31_ARB
    enum: GL_MATRIX_PALETTE_ARB
    enum: GL_MAX_MATRIX_PALETTE_STACK_DEPTH_ARB
    enum: GL_MAX_PALETTE_MATRICES_ARB
    enum: GL_CURRENT_PALETTE_MATRIX_ARB
    enum: GL_MATRIX_INDEX_ARRAY_ARB
    enum: GL_CURRENT_MATRIX_INDEX_ARB
    enum: GL_MATRIX_INDEX_ARRAY_SIZE_ARB
    enum: GL_MATRIX_INDEX_ARRAY_TYPE_ARB
    enum: GL_MATRIX_INDEX_ARRAY_STRIDE_ARB
    enum: GL_MATRIX_INDEX_ARRAY_POINTER_ARB
    enum: GL_COMBINE_ARB
    enum: GL_COMBINE_RGB_ARB
    enum: GL_COMBINE_ALPHA_ARB
    enum: GL_SOURCE0_RGB_ARB
    enum: GL_SOURCE1_RGB_ARB
    enum: GL_SOURCE2_RGB_ARB
    enum: GL_SOURCE0_ALPHA_ARB
    enum: GL_SOURCE1_ALPHA_ARB
    enum: GL_SOURCE2_ALPHA_ARB
    enum: GL_OPERAND0_RGB_ARB
    enum: GL_OPERAND1_RGB_ARB
    enum: GL_OPERAND2_RGB_ARB
    enum: GL_OPERAND0_ALPHA_ARB
    enum: GL_OPERAND1_ALPHA_ARB
    enum: GL_OPERAND2_ALPHA_ARB
    enum: GL_RGB_SCALE_ARB
    enum: GL_ADD_SIGNED_ARB
    enum: GL_INTERPOLATE_ARB
    enum: GL_SUBTRACT_ARB
    enum: GL_CONSTANT_ARB
    enum: GL_PRIMARY_COLOR_ARB
    enum: GL_PREVIOUS_ARB
    enum: GL_DOT3_RGB_ARB
    enum: GL_DOT3_RGBA_ARB
    enum: GL_MIRRORED_REPEAT_ARB
    enum: GL_DEPTH_COMPONENT16_ARB
    enum: GL_DEPTH_COMPONENT24_ARB
    enum: GL_DEPTH_COMPONENT32_ARB
    enum: GL_TEXTURE_DEPTH_SIZE_ARB
    enum: GL_DEPTH_TEXTURE_MODE_ARB
    enum: GL_TEXTURE_COMPARE_MODE_ARB
    enum: GL_TEXTURE_COMPARE_FUNC_ARB
    enum: GL_COMPARE_R_TO_TEXTURE_ARB
    enum: GL_TEXTURE_COMPARE_FAIL_VALUE_ARB
    enum: GL_COLOR_SUM_ARB
    enum: GL_VERTEX_PROGRAM_ARB
    enum: GL_VERTEX_ATTRIB_ARRAY_ENABLED_ARB
    enum: GL_VERTEX_ATTRIB_ARRAY_SIZE_ARB
    enum: GL_VERTEX_ATTRIB_ARRAY_STRIDE_ARB
    enum: GL_VERTEX_ATTRIB_ARRAY_TYPE_ARB
    enum: GL_CURRENT_VERTEX_ATTRIB_ARB
    enum: GL_PROGRAM_LENGTH_ARB
    enum: GL_PROGRAM_STRING_ARB
    enum: GL_MAX_PROGRAM_MATRIX_STACK_DEPTH_ARB
    enum: GL_MAX_PROGRAM_MATRICES_ARB
    enum: GL_CURRENT_MATRIX_STACK_DEPTH_ARB
    enum: GL_CURRENT_MATRIX_ARB
    enum: GL_VERTEX_PROGRAM_POINT_SIZE_ARB
    enum: GL_VERTEX_PROGRAM_TWO_SIDE_ARB
    enum: GL_VERTEX_ATTRIB_ARRAY_POINTER_ARB
    enum: GL_PROGRAM_ERROR_POSITION_ARB
    enum: GL_PROGRAM_BINDING_ARB
    enum: GL_MAX_VERTEX_ATTRIBS_ARB
    enum: GL_VERTEX_ATTRIB_ARRAY_NORMALIZED_ARB
    enum: GL_PROGRAM_ERROR_STRING_ARB
    enum: GL_PROGRAM_FORMAT_ASCII_ARB
    enum: GL_PROGRAM_FORMAT_ARB
    enum: GL_PROGRAM_INSTRUCTIONS_ARB
    enum: GL_MAX_PROGRAM_INSTRUCTIONS_ARB
    enum: GL_PROGRAM_NATIVE_INSTRUCTIONS_ARB
    enum: GL_MAX_PROGRAM_NATIVE_INSTRUCTIONS_ARB
    enum: GL_PROGRAM_TEMPORARIES_ARB
    enum: GL_MAX_PROGRAM_TEMPORARIES_ARB
    enum: GL_PROGRAM_NATIVE_TEMPORARIES_ARB
    enum: GL_MAX_PROGRAM_NATIVE_TEMPORARIES_ARB
    enum: GL_PROGRAM_PARAMETERS_ARB
    enum: GL_MAX_PROGRAM_PARAMETERS_ARB
    enum: GL_PROGRAM_NATIVE_PARAMETERS_ARB
    enum: GL_MAX_PROGRAM_NATIVE_PARAMETERS_ARB
    enum: GL_PROGRAM_ATTRIBS_ARB
    enum: GL_MAX_PROGRAM_ATTRIBS_ARB
    enum: GL_PROGRAM_NATIVE_ATTRIBS_ARB
    enum: GL_MAX_PROGRAM_NATIVE_ATTRIBS_ARB
    enum: GL_PROGRAM_ADDRESS_REGISTERS_ARB
    enum: GL_MAX_PROGRAM_ADDRESS_REGISTERS_ARB
    enum: GL_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB
    enum: GL_MAX_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB
    enum: GL_MAX_PROGRAM_LOCAL_PARAMETERS_ARB
    enum: GL_MAX_PROGRAM_ENV_PARAMETERS_ARB
    enum: GL_PROGRAM_UNDER_NATIVE_LIMITS_ARB
    enum: GL_TRANSPOSE_CURRENT_MATRIX_ARB
    enum: GL_MATRIX0_ARB
    enum: GL_MATRIX1_ARB
    enum: GL_MATRIX2_ARB
    enum: GL_MATRIX3_ARB
    enum: GL_MATRIX4_ARB
    enum: GL_MATRIX5_ARB
    enum: GL_MATRIX6_ARB
    enum: GL_MATRIX7_ARB
    enum: GL_MATRIX8_ARB
    enum: GL_MATRIX9_ARB
    enum: GL_MATRIX10_ARB
    enum: GL_MATRIX11_ARB
    enum: GL_MATRIX12_ARB
    enum: GL_MATRIX13_ARB
    enum: GL_MATRIX14_ARB
    enum: GL_MATRIX15_ARB
    enum: GL_MATRIX16_ARB
    enum: GL_MATRIX17_ARB
    enum: GL_MATRIX18_ARB
    enum: GL_MATRIX19_ARB
    enum: GL_MATRIX20_ARB
    enum: GL_MATRIX21_ARB
    enum: GL_MATRIX22_ARB
    enum: GL_MATRIX23_ARB
    enum: GL_MATRIX24_ARB
    enum: GL_MATRIX25_ARB
    enum: GL_MATRIX26_ARB
    enum: GL_MATRIX27_ARB
    enum: GL_MATRIX28_ARB
    enum: GL_MATRIX29_ARB
    enum: GL_MATRIX30_ARB
    enum: GL_MATRIX31_ARB
    enum: GL_FRAGMENT_PROGRAM_ARB
    enum: GL_PROGRAM_ALU_INSTRUCTIONS_ARB
    enum: GL_PROGRAM_TEX_INSTRUCTIONS_ARB
    enum: GL_PROGRAM_TEX_INDIRECTIONS_ARB
    enum: GL_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB
    enum: GL_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB
    enum: GL_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB
    enum: GL_MAX_PROGRAM_ALU_INSTRUCTIONS_ARB
    enum: GL_MAX_PROGRAM_TEX_INSTRUCTIONS_ARB
    enum: GL_MAX_PROGRAM_TEX_INDIRECTIONS_ARB
    enum: GL_MAX_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB
    enum: GL_MAX_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB
    enum: GL_MAX_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB
    enum: GL_MAX_TEXTURE_COORDS_ARB
    enum: GL_MAX_TEXTURE_IMAGE_UNITS_ARB
    enum: GL_BUFFER_SIZE_ARB
    enum: GL_BUFFER_USAGE_ARB
    enum: GL_ARRAY_BUFFER_ARB
    enum: GL_ELEMENT_ARRAY_BUFFER_ARB
    enum: GL_ARRAY_BUFFER_BINDING_ARB
    enum: GL_ELEMENT_ARRAY_BUFFER_BINDING_ARB
    enum: GL_VERTEX_ARRAY_BUFFER_BINDING_ARB
    enum: GL_NORMAL_ARRAY_BUFFER_BINDING_ARB
    enum: GL_COLOR_ARRAY_BUFFER_BINDING_ARB
    enum: GL_INDEX_ARRAY_BUFFER_BINDING_ARB
    enum: GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING_ARB
    enum: GL_EDGE_FLAG_ARRAY_BUFFER_BINDING_ARB
    enum: GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING_ARB
    enum: GL_FOG_COORDINATE_ARRAY_BUFFER_BINDING_ARB
    enum: GL_WEIGHT_ARRAY_BUFFER_BINDING_ARB
    enum: GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING_ARB
    enum: GL_READ_ONLY_ARB
    enum: GL_WRITE_ONLY_ARB
    enum: GL_READ_WRITE_ARB
    enum: GL_BUFFER_ACCESS_ARB
    enum: GL_BUFFER_MAPPED_ARB
    enum: GL_BUFFER_MAP_POINTER_ARB
    enum: GL_STREAM_DRAW_ARB
    enum: GL_STREAM_READ_ARB
    enum: GL_STREAM_COPY_ARB
    enum: GL_STATIC_DRAW_ARB
    enum: GL_STATIC_READ_ARB
    enum: GL_STATIC_COPY_ARB
    enum: GL_DYNAMIC_DRAW_ARB
    enum: GL_DYNAMIC_READ_ARB
    enum: GL_DYNAMIC_COPY_ARB
    enum: GL_QUERY_COUNTER_BITS_ARB
    enum: GL_CURRENT_QUERY_ARB
    enum: GL_QUERY_RESULT_ARB
    enum: GL_QUERY_RESULT_AVAILABLE_ARB
    enum: GL_SAMPLES_PASSED_ARB
    enum: GL_PROGRAM_OBJECT_ARB
    enum: GL_SHADER_OBJECT_ARB
    enum: GL_OBJECT_TYPE_ARB
    enum: GL_OBJECT_SUBTYPE_ARB
    enum: GL_FLOAT_VEC2_ARB
    enum: GL_FLOAT_VEC3_ARB
    enum: GL_FLOAT_VEC4_ARB
    enum: GL_INT_VEC2_ARB
    enum: GL_INT_VEC3_ARB
    enum: GL_INT_VEC4_ARB
    enum: GL_BOOL_ARB
    enum: GL_BOOL_VEC2_ARB
    enum: GL_BOOL_VEC3_ARB
    enum: GL_BOOL_VEC4_ARB
    enum: GL_FLOAT_MAT2_ARB
    enum: GL_FLOAT_MAT3_ARB
    enum: GL_FLOAT_MAT4_ARB
    enum: GL_SAMPLER_1D_ARB
    enum: GL_SAMPLER_2D_ARB
    enum: GL_SAMPLER_3D_ARB
    enum: GL_SAMPLER_CUBE_ARB
    enum: GL_SAMPLER_1D_SHADOW_ARB
    enum: GL_SAMPLER_2D_SHADOW_ARB
    enum: GL_SAMPLER_2D_RECT_ARB
    enum: GL_SAMPLER_2D_RECT_SHADOW_ARB
    enum: GL_OBJECT_DELETE_STATUS_ARB
    enum: GL_OBJECT_COMPILE_STATUS_ARB
    enum: GL_OBJECT_LINK_STATUS_ARB
    enum: GL_OBJECT_VALIDATE_STATUS_ARB
    enum: GL_OBJECT_INFO_LOG_LENGTH_ARB
    enum: GL_OBJECT_ATTACHED_OBJECTS_ARB
    enum: GL_OBJECT_ACTIVE_UNIFORMS_ARB
    enum: GL_OBJECT_ACTIVE_UNIFORM_MAX_LENGTH_ARB
    enum: GL_OBJECT_SHADER_SOURCE_LENGTH_ARB
    enum: GL_VERTEX_SHADER_ARB
    enum: GL_MAX_VERTEX_UNIFORM_COMPONENTS_ARB
    enum: GL_MAX_VARYING_FLOATS_ARB
    enum: GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS_ARB
    enum: GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS_ARB
    enum: GL_OBJECT_ACTIVE_ATTRIBUTES_ARB
    enum: GL_OBJECT_ACTIVE_ATTRIBUTE_MAX_LENGTH_ARB
    enum: GL_FRAGMENT_SHADER_ARB
    enum: GL_MAX_FRAGMENT_UNIFORM_COMPONENTS_ARB
    enum: GL_FRAGMENT_SHADER_DERIVATIVE_HINT_ARB
    enum: GL_SHADING_LANGUAGE_VERSION_ARB
    enum: GL_POINT_SPRITE_ARB
    enum: GL_COORD_REPLACE_ARB
    enum: GL_MAX_DRAW_BUFFERS_ARB
    enum: GL_DRAW_BUFFER0_ARB
    enum: GL_DRAW_BUFFER1_ARB
    enum: GL_DRAW_BUFFER2_ARB
    enum: GL_DRAW_BUFFER3_ARB
    enum: GL_DRAW_BUFFER4_ARB
    enum: GL_DRAW_BUFFER5_ARB
    enum: GL_DRAW_BUFFER6_ARB
    enum: GL_DRAW_BUFFER7_ARB
    enum: GL_DRAW_BUFFER8_ARB
    enum: GL_DRAW_BUFFER9_ARB
    enum: GL_DRAW_BUFFER10_ARB
    enum: GL_DRAW_BUFFER11_ARB
    enum: GL_DRAW_BUFFER12_ARB
    enum: GL_DRAW_BUFFER13_ARB
    enum: GL_DRAW_BUFFER14_ARB
    enum: GL_DRAW_BUFFER15_ARB
    enum: GL_TEXTURE_RECTANGLE_ARB
    enum: GL_TEXTURE_BINDING_RECTANGLE_ARB
    enum: GL_PROXY_TEXTURE_RECTANGLE_ARB
    enum: GL_MAX_RECTANGLE_TEXTURE_SIZE_ARB
    enum: GL_RGBA_FLOAT_MODE_ARB
    enum: GL_CLAMP_VERTEX_COLOR_ARB
    enum: GL_CLAMP_FRAGMENT_COLOR_ARB
    enum: GL_CLAMP_READ_COLOR_ARB
    enum: GL_FIXED_ONLY_ARB
    enum: GL_HALF_FLOAT_ARB
    enum: GL_TEXTURE_RED_TYPE_ARB
    enum: GL_TEXTURE_GREEN_TYPE_ARB
    enum: GL_TEXTURE_BLUE_TYPE_ARB
    enum: GL_TEXTURE_ALPHA_TYPE_ARB
    enum: GL_TEXTURE_LUMINANCE_TYPE_ARB
    enum: GL_TEXTURE_INTENSITY_TYPE_ARB
    enum: GL_TEXTURE_DEPTH_TYPE_ARB
    enum: GL_UNSIGNED_NORMALIZED_ARB
    enum: GL_RGBA32F_ARB
    enum: GL_RGB32F_ARB
    enum: GL_ALPHA32F_ARB
    enum: GL_INTENSITY32F_ARB
    enum: GL_LUMINANCE32F_ARB
    enum: GL_LUMINANCE_ALPHA32F_ARB
    enum: GL_RGBA16F_ARB
    enum: GL_RGB16F_ARB
    enum: GL_ALPHA16F_ARB
    enum: GL_INTENSITY16F_ARB
    enum: GL_LUMINANCE16F_ARB
    enum: GL_LUMINANCE_ALPHA16F_ARB
    enum: GL_PIXEL_PACK_BUFFER_ARB
    enum: GL_PIXEL_UNPACK_BUFFER_ARB
    enum: GL_PIXEL_PACK_BUFFER_BINDING_ARB
    enum: GL_PIXEL_UNPACK_BUFFER_BINDING_ARB
    enum: GL_DEPTH_COMPONENT32F
    enum: GL_DEPTH32F_STENCIL8
    enum: GL_FLOAT_32_UNSIGNED_INT_24_8_REV
    enum: GL_INVALID_FRAMEBUFFER_OPERATION
    enum: GL_FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING
    enum: GL_FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE
    enum: GL_FRAMEBUFFER_ATTACHMENT_RED_SIZE
    enum: GL_FRAMEBUFFER_ATTACHMENT_GREEN_SIZE
    enum: GL_FRAMEBUFFER_ATTACHMENT_BLUE_SIZE
    enum: GL_FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE
    enum: GL_FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE
    enum: GL_FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE
    enum: GL_FRAMEBUFFER_DEFAULT
    enum: GL_FRAMEBUFFER_UNDEFINED
    enum: GL_DEPTH_STENCIL_ATTACHMENT
    enum: GL_MAX_RENDERBUFFER_SIZE
    enum: GL_DEPTH_STENCIL
    enum: GL_UNSIGNED_INT_24_8
    enum: GL_DEPTH24_STENCIL8
    enum: GL_TEXTURE_STENCIL_SIZE
    enum: GL_TEXTURE_RED_TYPE
    enum: GL_TEXTURE_GREEN_TYPE
    enum: GL_TEXTURE_BLUE_TYPE
    enum: GL_TEXTURE_ALPHA_TYPE
    enum: GL_TEXTURE_DEPTH_TYPE
    enum: GL_UNSIGNED_NORMALIZED
    enum: GL_FRAMEBUFFER_BINDING
    enum: GL_DRAW_FRAMEBUFFER_BINDING
    enum: GL_RENDERBUFFER_BINDING
    enum: GL_READ_FRAMEBUFFER
    enum: GL_DRAW_FRAMEBUFFER
    enum: GL_READ_FRAMEBUFFER_BINDING
    enum: GL_RENDERBUFFER_SAMPLES
    enum: GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE
    enum: GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME
    enum: GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL
    enum: GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE
    enum: GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER
    enum: GL_FRAMEBUFFER_COMPLETE
    enum: GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT
    enum: GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT
    enum: GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER
    enum: GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER
    enum: GL_FRAMEBUFFER_UNSUPPORTED
    enum: GL_MAX_COLOR_ATTACHMENTS
    enum: GL_COLOR_ATTACHMENT0
    enum: GL_COLOR_ATTACHMENT1
    enum: GL_COLOR_ATTACHMENT2
    enum: GL_COLOR_ATTACHMENT3
    enum: GL_COLOR_ATTACHMENT4
    enum: GL_COLOR_ATTACHMENT5
    enum: GL_COLOR_ATTACHMENT6
    enum: GL_COLOR_ATTACHMENT7
    enum: GL_COLOR_ATTACHMENT8
    enum: GL_COLOR_ATTACHMENT9
    enum: GL_COLOR_ATTACHMENT10
    enum: GL_COLOR_ATTACHMENT11
    enum: GL_COLOR_ATTACHMENT12
    enum: GL_COLOR_ATTACHMENT13
    enum: GL_COLOR_ATTACHMENT14
    enum: GL_COLOR_ATTACHMENT15
    enum: GL_DEPTH_ATTACHMENT
    enum: GL_STENCIL_ATTACHMENT
    enum: GL_FRAMEBUFFER
    enum: GL_RENDERBUFFER
    enum: GL_RENDERBUFFER_WIDTH
    enum: GL_RENDERBUFFER_HEIGHT
    enum: GL_RENDERBUFFER_INTERNAL_FORMAT
    enum: GL_STENCIL_INDEX1
    enum: GL_STENCIL_INDEX4
    enum: GL_STENCIL_INDEX8
    enum: GL_STENCIL_INDEX16
    enum: GL_RENDERBUFFER_RED_SIZE
    enum: GL_RENDERBUFFER_GREEN_SIZE
    enum: GL_RENDERBUFFER_BLUE_SIZE
    enum: GL_RENDERBUFFER_ALPHA_SIZE
    enum: GL_RENDERBUFFER_DEPTH_SIZE
    enum: GL_RENDERBUFFER_STENCIL_SIZE
    enum: GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE
    enum: GL_MAX_SAMPLES
    enum: GL_INDEX
    enum: GL_TEXTURE_LUMINANCE_TYPE
    enum: GL_TEXTURE_INTENSITY_TYPE
    enum: GL_FRAMEBUFFER_SRGB
    enum: GL_LINES_ADJACENCY_ARB
    enum: GL_LINE_STRIP_ADJACENCY_ARB
    enum: GL_TRIANGLES_ADJACENCY_ARB
    enum: GL_TRIANGLE_STRIP_ADJACENCY_ARB
    enum: GL_PROGRAM_POINT_SIZE_ARB
    enum: GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS_ARB
    enum: GL_FRAMEBUFFER_ATTACHMENT_LAYERED_ARB
    enum: GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS_ARB
    enum: GL_FRAMEBUFFER_INCOMPLETE_LAYER_COUNT_ARB
    enum: GL_GEOMETRY_SHADER_ARB
    enum: GL_GEOMETRY_VERTICES_OUT_ARB
    enum: GL_GEOMETRY_INPUT_TYPE_ARB
    enum: GL_GEOMETRY_OUTPUT_TYPE_ARB
    enum: GL_MAX_GEOMETRY_VARYING_COMPONENTS_ARB
    enum: GL_MAX_VERTEX_VARYING_COMPONENTS_ARB
    enum: GL_MAX_GEOMETRY_UNIFORM_COMPONENTS_ARB
    enum: GL_MAX_GEOMETRY_OUTPUT_VERTICES_ARB
    enum: GL_MAX_GEOMETRY_TOTAL_OUTPUT_COMPONENTS_ARB
    enum: GL_HALF_FLOAT
    enum: GL_VERTEX_ATTRIB_ARRAY_DIVISOR_ARB
    enum: GL_MAP_READ_BIT
    enum: GL_MAP_WRITE_BIT
    enum: GL_MAP_INVALIDATE_RANGE_BIT
    enum: GL_MAP_INVALIDATE_BUFFER_BIT
    enum: GL_MAP_FLUSH_EXPLICIT_BIT
    enum: GL_MAP_UNSYNCHRONIZED_BIT
    enum: GL_TEXTURE_BUFFER_ARB
    enum: GL_MAX_TEXTURE_BUFFER_SIZE_ARB
    enum: GL_TEXTURE_BINDING_BUFFER_ARB
    enum: GL_TEXTURE_BUFFER_DATA_STORE_BINDING_ARB
    enum: GL_TEXTURE_BUFFER_FORMAT_ARB
    enum: GL_COMPRESSED_RED_RGTC1
    enum: GL_COMPRESSED_SIGNED_RED_RGTC1
    enum: GL_COMPRESSED_RG_RGTC2
    enum: GL_COMPRESSED_SIGNED_RG_RGTC2
    enum: GL_RG
    enum: GL_RG_INTEGER
    enum: GL_R8
    enum: GL_R16
    enum: GL_RG8
    enum: GL_RG16
    enum: GL_R16F
    enum: GL_R32F
    enum: GL_RG16F
    enum: GL_RG32F
    enum: GL_R8I
    enum: GL_R8UI
    enum: GL_R16I
    enum: GL_R16UI
    enum: GL_R32I
    enum: GL_R32UI
    enum: GL_RG8I
    enum: GL_RG8UI
    enum: GL_RG16I
    enum: GL_RG16UI
    enum: GL_RG32I
    enum: GL_RG32UI
    enum: GL_VERTEX_ARRAY_BINDING
    enum: GL_UNIFORM_BUFFER
    enum: GL_UNIFORM_BUFFER_BINDING
    enum: GL_UNIFORM_BUFFER_START
    enum: GL_UNIFORM_BUFFER_SIZE
    enum: GL_MAX_VERTEX_UNIFORM_BLOCKS
    enum: GL_MAX_GEOMETRY_UNIFORM_BLOCKS
    enum: GL_MAX_FRAGMENT_UNIFORM_BLOCKS
    enum: GL_MAX_COMBINED_UNIFORM_BLOCKS
    enum: GL_MAX_UNIFORM_BUFFER_BINDINGS
    enum: GL_MAX_UNIFORM_BLOCK_SIZE
    enum: GL_MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS
    enum: GL_MAX_COMBINED_GEOMETRY_UNIFORM_COMPONENTS
    enum: GL_MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS
    enum: GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT
    enum: GL_ACTIVE_UNIFORM_BLOCK_MAX_NAME_LENGTH
    enum: GL_ACTIVE_UNIFORM_BLOCKS
    enum: GL_UNIFORM_TYPE
    enum: GL_UNIFORM_SIZE
    enum: GL_UNIFORM_NAME_LENGTH
    enum: GL_UNIFORM_BLOCK_INDEX
    enum: GL_UNIFORM_OFFSET
    enum: GL_UNIFORM_ARRAY_STRIDE
    enum: GL_UNIFORM_MATRIX_STRIDE
    enum: GL_UNIFORM_IS_ROW_MAJOR
    enum: GL_UNIFORM_BLOCK_BINDING
    enum: GL_UNIFORM_BLOCK_DATA_SIZE
    enum: GL_UNIFORM_BLOCK_NAME_LENGTH
    enum: GL_UNIFORM_BLOCK_ACTIVE_UNIFORMS
    enum: GL_UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES
    enum: GL_UNIFORM_BLOCK_REFERENCED_BY_VERTEX_SHADER
    enum: GL_UNIFORM_BLOCK_REFERENCED_BY_GEOMETRY_SHADER
    enum: GL_UNIFORM_BLOCK_REFERENCED_BY_FRAGMENT_SHADER
    enum: GL_INVALID_INDEX
    enum: GL_COPY_READ_BUFFER_BINDING
    enum: GL_COPY_READ_BUFFER
    enum: GL_COPY_WRITE_BUFFER_BINDING
    enum: GL_COPY_WRITE_BUFFER
    enum: GL_DEPTH_CLAMP
    enum: GL_QUADS_FOLLOW_PROVOKING_VERTEX_CONVENTION
    enum: GL_FIRST_VERTEX_CONVENTION
    enum: GL_LAST_VERTEX_CONVENTION
    enum: GL_PROVOKING_VERTEX
    enum: GL_TEXTURE_CUBE_MAP_SEAMLESS
    enum: GL_MAX_SERVER_WAIT_TIMEOUT
    enum: GL_OBJECT_TYPE
    enum: GL_SYNC_CONDITION
    enum: GL_SYNC_STATUS
    enum: GL_SYNC_FLAGS
    enum: GL_SYNC_FENCE
    enum: GL_SYNC_GPU_COMMANDS_COMPLETE
    enum: GL_UNSIGNALED
    enum: GL_SIGNALED
    enum: GL_ALREADY_SIGNALED
    enum: GL_TIMEOUT_EXPIRED
    enum: GL_CONDITION_SATISFIED
    enum: GL_WAIT_FAILED
    enum: GL_SYNC_FLUSH_COMMANDS_BIT
    enum: GL_TIMEOUT_IGNORED
    enum: GL_SAMPLE_POSITION
    enum: GL_SAMPLE_MASK
    enum: GL_SAMPLE_MASK_VALUE
    enum: GL_MAX_SAMPLE_MASK_WORDS
    enum: GL_TEXTURE_2D_MULTISAMPLE
    enum: GL_PROXY_TEXTURE_2D_MULTISAMPLE
    enum: GL_TEXTURE_2D_MULTISAMPLE_ARRAY
    enum: GL_PROXY_TEXTURE_2D_MULTISAMPLE_ARRAY
    enum: GL_TEXTURE_BINDING_2D_MULTISAMPLE
    enum: GL_TEXTURE_BINDING_2D_MULTISAMPLE_ARRAY
    enum: GL_TEXTURE_SAMPLES
    enum: GL_TEXTURE_FIXED_SAMPLE_LOCATIONS
    enum: GL_SAMPLER_2D_MULTISAMPLE
    enum: GL_INT_SAMPLER_2D_MULTISAMPLE
    enum: GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE
    enum: GL_SAMPLER_2D_MULTISAMPLE_ARRAY
    enum: GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY
    enum: GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY
    enum: GL_MAX_COLOR_TEXTURE_SAMPLES
    enum: GL_MAX_DEPTH_TEXTURE_SAMPLES
    enum: GL_MAX_INTEGER_SAMPLES
    enum: GL_SAMPLE_SHADING_ARB
    enum: GL_MIN_SAMPLE_SHADING_VALUE_ARB
    enum: GL_TEXTURE_CUBE_MAP_ARRAY_ARB
    enum: GL_TEXTURE_BINDING_CUBE_MAP_ARRAY_ARB
    enum: GL_PROXY_TEXTURE_CUBE_MAP_ARRAY_ARB
    enum: GL_SAMPLER_CUBE_MAP_ARRAY_ARB
    enum: GL_SAMPLER_CUBE_MAP_ARRAY_SHADOW_ARB
    enum: GL_INT_SAMPLER_CUBE_MAP_ARRAY_ARB
    enum: GL_UNSIGNED_INT_SAMPLER_CUBE_MAP_ARRAY_ARB
    enum: GL_MIN_PROGRAM_TEXTURE_GATHER_OFFSET_ARB
    enum: GL_MAX_PROGRAM_TEXTURE_GATHER_OFFSET_ARB
    enum: GL_MAX_PROGRAM_TEXTURE_GATHER_COMPONENTS_ARB
    enum: GL_SHADER_INCLUDE_ARB
    enum: GL_NAMED_STRING_LENGTH_ARB
    enum: GL_NAMED_STRING_TYPE_ARB
    enum: GL_COMPRESSED_RGBA_BPTC_UNORM_ARB
    enum: GL_COMPRESSED_SRGB_ALPHA_BPTC_UNORM_ARB
    enum: GL_COMPRESSED_RGB_BPTC_SIGNED_FLOAT_ARB
    enum: GL_COMPRESSED_RGB_BPTC_UNSIGNED_FLOAT_ARB
    enum: GL_SRC1_COLOR
    enum: GL_ONE_MINUS_SRC1_COLOR
    enum: GL_ONE_MINUS_SRC1_ALPHA
    enum: GL_MAX_DUAL_SOURCE_DRAW_BUFFERS
    enum: GL_ANY_SAMPLES_PASSED
    enum: GL_SAMPLER_BINDING
    enum: GL_RGB10_A2UI
    enum: GL_TEXTURE_SWIZZLE_R
    enum: GL_TEXTURE_SWIZZLE_G
    enum: GL_TEXTURE_SWIZZLE_B
    enum: GL_TEXTURE_SWIZZLE_A
    enum: GL_TEXTURE_SWIZZLE_RGBA
    enum: GL_TIME_ELAPSED
    enum: GL_TIMESTAMP
    enum: GL_INT_2_10_10_10_REV
    enum: GL_DRAW_INDIRECT_BUFFER
    enum: GL_DRAW_INDIRECT_BUFFER_BINDING
    enum: GL_GEOMETRY_SHADER_INVOCATIONS
    enum: GL_MAX_GEOMETRY_SHADER_INVOCATIONS
    enum: GL_MIN_FRAGMENT_INTERPOLATION_OFFSET
    enum: GL_MAX_FRAGMENT_INTERPOLATION_OFFSET
    enum: GL_FRAGMENT_INTERPOLATION_OFFSET_BITS
    enum: GL_DOUBLE_VEC2
    enum: GL_DOUBLE_VEC3
    enum: GL_DOUBLE_VEC4
    enum: GL_DOUBLE_MAT2
    enum: GL_DOUBLE_MAT3
    enum: GL_DOUBLE_MAT4
    enum: GL_DOUBLE_MAT2x3
    enum: GL_DOUBLE_MAT2x4
    enum: GL_DOUBLE_MAT3x2
    enum: GL_DOUBLE_MAT3x4
    enum: GL_DOUBLE_MAT4x2
    enum: GL_DOUBLE_MAT4x3
    enum: GL_ACTIVE_SUBROUTINES
    enum: GL_ACTIVE_SUBROUTINE_UNIFORMS
    enum: GL_ACTIVE_SUBROUTINE_UNIFORM_LOCATIONS
    enum: GL_ACTIVE_SUBROUTINE_MAX_LENGTH
    enum: GL_ACTIVE_SUBROUTINE_UNIFORM_MAX_LENGTH
    enum: GL_MAX_SUBROUTINES
    enum: GL_MAX_SUBROUTINE_UNIFORM_LOCATIONS
    enum: GL_NUM_COMPATIBLE_SUBROUTINES
    enum: GL_COMPATIBLE_SUBROUTINES
    enum: GL_PATCHES
    enum: GL_PATCH_VERTICES
    enum: GL_PATCH_DEFAULT_INNER_LEVEL
    enum: GL_PATCH_DEFAULT_OUTER_LEVEL
    enum: GL_TESS_CONTROL_OUTPUT_VERTICES
    enum: GL_TESS_GEN_MODE
    enum: GL_TESS_GEN_SPACING
    enum: GL_TESS_GEN_VERTEX_ORDER
    enum: GL_TESS_GEN_POINT_MODE
    enum: GL_ISOLINES
    enum: GL_FRACTIONAL_ODD
    enum: GL_FRACTIONAL_EVEN
    enum: GL_MAX_PATCH_VERTICES
    enum: GL_MAX_TESS_GEN_LEVEL
    enum: GL_MAX_TESS_CONTROL_UNIFORM_COMPONENTS
    enum: GL_MAX_TESS_EVALUATION_UNIFORM_COMPONENTS
    enum: GL_MAX_TESS_CONTROL_TEXTURE_IMAGE_UNITS
    enum: GL_MAX_TESS_EVALUATION_TEXTURE_IMAGE_UNITS
    enum: GL_MAX_TESS_CONTROL_OUTPUT_COMPONENTS
    enum: GL_MAX_TESS_PATCH_COMPONENTS
    enum: GL_MAX_TESS_CONTROL_TOTAL_OUTPUT_COMPONENTS
    enum: GL_MAX_TESS_EVALUATION_OUTPUT_COMPONENTS
    enum: GL_MAX_TESS_CONTROL_UNIFORM_BLOCKS
    enum: GL_MAX_TESS_EVALUATION_UNIFORM_BLOCKS
    enum: GL_MAX_TESS_CONTROL_INPUT_COMPONENTS
    enum: GL_MAX_TESS_EVALUATION_INPUT_COMPONENTS
    enum: GL_MAX_COMBINED_TESS_CONTROL_UNIFORM_COMPONENTS
    enum: GL_MAX_COMBINED_TESS_EVALUATION_UNIFORM_COMPONENTS
    enum: GL_UNIFORM_BLOCK_REFERENCED_BY_TESS_CONTROL_SHADER
    enum: GL_UNIFORM_BLOCK_REFERENCED_BY_TESS_EVALUATION_SHADER
    enum: GL_TESS_EVALUATION_SHADER
    enum: GL_TESS_CONTROL_SHADER
    enum: GL_TRANSFORM_FEEDBACK
    enum: GL_TRANSFORM_FEEDBACK_PAUSED
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_PAUSED
    enum: GL_TRANSFORM_FEEDBACK_ACTIVE
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_ACTIVE
    enum: GL_TRANSFORM_FEEDBACK_BINDING
    enum: GL_MAX_TRANSFORM_FEEDBACK_BUFFERS
    enum: GL_MAX_VERTEX_STREAMS
    enum: GL_FIXED
    enum: GL_IMPLEMENTATION_COLOR_READ_TYPE
    enum: GL_IMPLEMENTATION_COLOR_READ_FORMAT
    enum: GL_LOW_FLOAT
    enum: GL_MEDIUM_FLOAT
    enum: GL_HIGH_FLOAT
    enum: GL_LOW_INT
    enum: GL_MEDIUM_INT
    enum: GL_HIGH_INT
    enum: GL_SHADER_COMPILER
    enum: GL_SHADER_BINARY_FORMATS
    enum: GL_NUM_SHADER_BINARY_FORMATS
    enum: GL_MAX_VERTEX_UNIFORM_VECTORS
    enum: GL_MAX_VARYING_VECTORS
    enum: GL_MAX_FRAGMENT_UNIFORM_VECTORS
    enum: GL_RGB565
    enum: GL_PROGRAM_BINARY_RETRIEVABLE_HINT
    enum: GL_PROGRAM_BINARY_LENGTH
    enum: GL_NUM_PROGRAM_BINARY_FORMATS
    enum: GL_PROGRAM_BINARY_FORMATS
    enum: GL_VERTEX_SHADER_BIT
    enum: GL_FRAGMENT_SHADER_BIT
    enum: GL_GEOMETRY_SHADER_BIT
    enum: GL_TESS_CONTROL_SHADER_BIT
    enum: GL_TESS_EVALUATION_SHADER_BIT
    enum: GL_ALL_SHADER_BITS
    enum: GL_PROGRAM_SEPARABLE
    enum: GL_ACTIVE_PROGRAM
    enum: GL_PROGRAM_PIPELINE_BINDING
    enum: GL_MAX_VIEWPORTS
    enum: GL_VIEWPORT_SUBPIXEL_BITS
    enum: GL_VIEWPORT_BOUNDS_RANGE
    enum: GL_LAYER_PROVOKING_VERTEX
    enum: GL_VIEWPORT_INDEX_PROVOKING_VERTEX
    enum: GL_UNDEFINED_VERTEX
    enum: GL_SYNC_CL_EVENT_ARB
    enum: GL_SYNC_CL_EVENT_COMPLETE_ARB
    enum: GL_DEBUG_OUTPUT_SYNCHRONOUS_ARB
    enum: GL_DEBUG_NEXT_LOGGED_MESSAGE_LENGTH_ARB
    enum: GL_DEBUG_CALLBACK_FUNCTION_ARB
    enum: GL_DEBUG_CALLBACK_USER_PARAM_ARB
    enum: GL_DEBUG_SOURCE_API_ARB
    enum: GL_DEBUG_SOURCE_WINDOW_SYSTEM_ARB
    enum: GL_DEBUG_SOURCE_SHADER_COMPILER_ARB
    enum: GL_DEBUG_SOURCE_THIRD_PARTY_ARB
    enum: GL_DEBUG_SOURCE_APPLICATION_ARB
    enum: GL_DEBUG_SOURCE_OTHER_ARB
    enum: GL_DEBUG_TYPE_ERROR_ARB
    enum: GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR_ARB
    enum: GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR_ARB
    enum: GL_DEBUG_TYPE_PORTABILITY_ARB
    enum: GL_DEBUG_TYPE_PERFORMANCE_ARB
    enum: GL_DEBUG_TYPE_OTHER_ARB
    enum: GL_MAX_DEBUG_MESSAGE_LENGTH_ARB
    enum: GL_MAX_DEBUG_LOGGED_MESSAGES_ARB
    enum: GL_DEBUG_LOGGED_MESSAGES_ARB
    enum: GL_DEBUG_SEVERITY_HIGH_ARB
    enum: GL_DEBUG_SEVERITY_MEDIUM_ARB
    enum: GL_DEBUG_SEVERITY_LOW_ARB
    enum: GL_CONTEXT_FLAG_ROBUST_ACCESS_BIT_ARB
    enum: GL_LOSE_CONTEXT_ON_RESET_ARB
    enum: GL_GUILTY_CONTEXT_RESET_ARB
    enum: GL_INNOCENT_CONTEXT_RESET_ARB
    enum: GL_UNKNOWN_CONTEXT_RESET_ARB
    enum: GL_RESET_NOTIFICATION_STRATEGY_ARB
    enum: GL_NO_RESET_NOTIFICATION_ARB
    enum: GL_UNPACK_COMPRESSED_BLOCK_WIDTH
    enum: GL_UNPACK_COMPRESSED_BLOCK_HEIGHT
    enum: GL_UNPACK_COMPRESSED_BLOCK_DEPTH
    enum: GL_UNPACK_COMPRESSED_BLOCK_SIZE
    enum: GL_PACK_COMPRESSED_BLOCK_WIDTH
    enum: GL_PACK_COMPRESSED_BLOCK_HEIGHT
    enum: GL_PACK_COMPRESSED_BLOCK_DEPTH
    enum: GL_PACK_COMPRESSED_BLOCK_SIZE
    enum: GL_NUM_SAMPLE_COUNTS
    enum: GL_MIN_MAP_BUFFER_ALIGNMENT
    enum: GL_ATOMIC_COUNTER_BUFFER
    enum: GL_ATOMIC_COUNTER_BUFFER_BINDING
    enum: GL_ATOMIC_COUNTER_BUFFER_START
    enum: GL_ATOMIC_COUNTER_BUFFER_SIZE
    enum: GL_ATOMIC_COUNTER_BUFFER_DATA_SIZE
    enum: GL_ATOMIC_COUNTER_BUFFER_ACTIVE_ATOMIC_COUNTERS
    enum: GL_ATOMIC_COUNTER_BUFFER_ACTIVE_ATOMIC_COUNTER_INDICES
    enum: GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_VERTEX_SHADER
    enum: GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_TESS_CONTROL_SHADER
    enum: GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_TESS_EVALUATION_SHADER
    enum: GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_GEOMETRY_SHADER
    enum: GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_FRAGMENT_SHADER
    enum: GL_MAX_VERTEX_ATOMIC_COUNTER_BUFFERS
    enum: GL_MAX_TESS_CONTROL_ATOMIC_COUNTER_BUFFERS
    enum: GL_MAX_TESS_EVALUATION_ATOMIC_COUNTER_BUFFERS
    enum: GL_MAX_GEOMETRY_ATOMIC_COUNTER_BUFFERS
    enum: GL_MAX_FRAGMENT_ATOMIC_COUNTER_BUFFERS
    enum: GL_MAX_COMBINED_ATOMIC_COUNTER_BUFFERS
    enum: GL_MAX_VERTEX_ATOMIC_COUNTERS
    enum: GL_MAX_TESS_CONTROL_ATOMIC_COUNTERS
    enum: GL_MAX_TESS_EVALUATION_ATOMIC_COUNTERS
    enum: GL_MAX_GEOMETRY_ATOMIC_COUNTERS
    enum: GL_MAX_FRAGMENT_ATOMIC_COUNTERS
    enum: GL_MAX_COMBINED_ATOMIC_COUNTERS
    enum: GL_MAX_ATOMIC_COUNTER_BUFFER_SIZE
    enum: GL_MAX_ATOMIC_COUNTER_BUFFER_BINDINGS
    enum: GL_ACTIVE_ATOMIC_COUNTER_BUFFERS
    enum: GL_UNIFORM_ATOMIC_COUNTER_BUFFER_INDEX
    enum: GL_UNSIGNED_INT_ATOMIC_COUNTER
    enum: GL_VERTEX_ATTRIB_ARRAY_BARRIER_BIT
    enum: GL_ELEMENT_ARRAY_BARRIER_BIT
    enum: GL_UNIFORM_BARRIER_BIT
    enum: GL_TEXTURE_FETCH_BARRIER_BIT
    enum: GL_SHADER_IMAGE_ACCESS_BARRIER_BIT
    enum: GL_COMMAND_BARRIER_BIT
    enum: GL_PIXEL_BUFFER_BARRIER_BIT
    enum: GL_TEXTURE_UPDATE_BARRIER_BIT
    enum: GL_BUFFER_UPDATE_BARRIER_BIT
    enum: GL_FRAMEBUFFER_BARRIER_BIT
    enum: GL_TRANSFORM_FEEDBACK_BARRIER_BIT
    enum: GL_ATOMIC_COUNTER_BARRIER_BIT
    enum: GL_ALL_BARRIER_BITS
    enum: GL_MAX_IMAGE_UNITS
    enum: GL_MAX_COMBINED_IMAGE_UNITS_AND_FRAGMENT_OUTPUTS
    enum: GL_IMAGE_BINDING_NAME
    enum: GL_IMAGE_BINDING_LEVEL
    enum: GL_IMAGE_BINDING_LAYERED
    enum: GL_IMAGE_BINDING_LAYER
    enum: GL_IMAGE_BINDING_ACCESS
    enum: GL_IMAGE_1D
    enum: GL_IMAGE_2D
    enum: GL_IMAGE_3D
    enum: GL_IMAGE_2D_RECT
    enum: GL_IMAGE_CUBE
    enum: GL_IMAGE_BUFFER
    enum: GL_IMAGE_1D_ARRAY
    enum: GL_IMAGE_2D_ARRAY
    enum: GL_IMAGE_CUBE_MAP_ARRAY
    enum: GL_IMAGE_2D_MULTISAMPLE
    enum: GL_IMAGE_2D_MULTISAMPLE_ARRAY
    enum: GL_INT_IMAGE_1D
    enum: GL_INT_IMAGE_2D
    enum: GL_INT_IMAGE_3D
    enum: GL_INT_IMAGE_2D_RECT
    enum: GL_INT_IMAGE_CUBE
    enum: GL_INT_IMAGE_BUFFER
    enum: GL_INT_IMAGE_1D_ARRAY
    enum: GL_INT_IMAGE_2D_ARRAY
    enum: GL_INT_IMAGE_CUBE_MAP_ARRAY
    enum: GL_INT_IMAGE_2D_MULTISAMPLE
    enum: GL_INT_IMAGE_2D_MULTISAMPLE_ARRAY
    enum: GL_UNSIGNED_INT_IMAGE_1D
    enum: GL_UNSIGNED_INT_IMAGE_2D
    enum: GL_UNSIGNED_INT_IMAGE_3D
    enum: GL_UNSIGNED_INT_IMAGE_2D_RECT
    enum: GL_UNSIGNED_INT_IMAGE_CUBE
    enum: GL_UNSIGNED_INT_IMAGE_BUFFER
    enum: GL_UNSIGNED_INT_IMAGE_1D_ARRAY
    enum: GL_UNSIGNED_INT_IMAGE_2D_ARRAY
    enum: GL_UNSIGNED_INT_IMAGE_CUBE_MAP_ARRAY
    enum: GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE
    enum: GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE_ARRAY
    enum: GL_MAX_IMAGE_SAMPLES
    enum: GL_IMAGE_BINDING_FORMAT
    enum: GL_IMAGE_FORMAT_COMPATIBILITY_TYPE
    enum: GL_IMAGE_FORMAT_COMPATIBILITY_BY_SIZE
    enum: GL_IMAGE_FORMAT_COMPATIBILITY_BY_CLASS
    enum: GL_MAX_VERTEX_IMAGE_UNIFORMS
    enum: GL_MAX_TESS_CONTROL_IMAGE_UNIFORMS
    enum: GL_MAX_TESS_EVALUATION_IMAGE_UNIFORMS
    enum: GL_MAX_GEOMETRY_IMAGE_UNIFORMS
    enum: GL_MAX_FRAGMENT_IMAGE_UNIFORMS
    enum: GL_MAX_COMBINED_IMAGE_UNIFORMS
    enum: GL_TEXTURE_IMMUTABLE_FORMAT
    enum: GL_COMPRESSED_RGBA_ASTC_4x4_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_5x4_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_5x5_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_6x5_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_6x6_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_8x5_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_8x6_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_8x8_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_10x5_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_10x6_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_10x8_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_10x10_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_12x10_KHR
    enum: GL_COMPRESSED_RGBA_ASTC_12x12_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_4x4_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_5x4_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_5x5_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_6x5_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_6x6_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_8x5_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_8x6_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_8x8_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_10x5_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_10x6_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_10x8_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_10x10_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_12x10_KHR
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ASTC_12x12_KHR
    enum: GL_DEBUG_OUTPUT_SYNCHRONOUS
    enum: GL_DEBUG_NEXT_LOGGED_MESSAGE_LENGTH
    enum: GL_DEBUG_CALLBACK_FUNCTION
    enum: GL_DEBUG_CALLBACK_USER_PARAM
    enum: GL_DEBUG_SOURCE_API
    enum: GL_DEBUG_SOURCE_WINDOW_SYSTEM
    enum: GL_DEBUG_SOURCE_SHADER_COMPILER
    enum: GL_DEBUG_SOURCE_THIRD_PARTY
    enum: GL_DEBUG_SOURCE_APPLICATION
    enum: GL_DEBUG_SOURCE_OTHER
    enum: GL_DEBUG_TYPE_ERROR
    enum: GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR
    enum: GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR
    enum: GL_DEBUG_TYPE_PORTABILITY
    enum: GL_DEBUG_TYPE_PERFORMANCE
    enum: GL_DEBUG_TYPE_OTHER
    enum: GL_DEBUG_TYPE_MARKER
    enum: GL_DEBUG_TYPE_PUSH_GROUP
    enum: GL_DEBUG_TYPE_POP_GROUP
    enum: GL_DEBUG_SEVERITY_NOTIFICATION
    enum: GL_MAX_DEBUG_GROUP_STACK_DEPTH
    enum: GL_DEBUG_GROUP_STACK_DEPTH
    enum: GL_BUFFER
    enum: GL_SHADER
    enum: GL_PROGRAM
    enum: GL_QUERY
    enum: GL_PROGRAM_PIPELINE
    enum: GL_SAMPLER
    enum: GL_DISPLAY_LIST
    enum: GL_MAX_LABEL_LENGTH
    enum: GL_MAX_DEBUG_MESSAGE_LENGTH
    enum: GL_MAX_DEBUG_LOGGED_MESSAGES
    enum: GL_DEBUG_LOGGED_MESSAGES
    enum: GL_DEBUG_SEVERITY_HIGH
    enum: GL_DEBUG_SEVERITY_MEDIUM
    enum: GL_DEBUG_SEVERITY_LOW
    enum: GL_DEBUG_OUTPUT
    enum: GL_CONTEXT_FLAG_DEBUG_BIT
    enum: GL_COMPUTE_SHADER
    enum: GL_MAX_COMPUTE_UNIFORM_BLOCKS
    enum: GL_MAX_COMPUTE_TEXTURE_IMAGE_UNITS
    enum: GL_MAX_COMPUTE_IMAGE_UNIFORMS
    enum: GL_MAX_COMPUTE_SHARED_MEMORY_SIZE
    enum: GL_MAX_COMPUTE_UNIFORM_COMPONENTS
    enum: GL_MAX_COMPUTE_ATOMIC_COUNTER_BUFFERS
    enum: GL_MAX_COMPUTE_ATOMIC_COUNTERS
    enum: GL_MAX_COMBINED_COMPUTE_UNIFORM_COMPONENTS
    enum: GL_MAX_COMPUTE_LOCAL_INVOCATIONS
    enum: GL_MAX_COMPUTE_WORK_GROUP_COUNT
    enum: GL_MAX_COMPUTE_WORK_GROUP_SIZE
    enum: GL_COMPUTE_LOCAL_WORK_SIZE
    enum: GL_UNIFORM_BLOCK_REFERENCED_BY_COMPUTE_SHADER
    enum: GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_COMPUTE_SHADER
    enum: GL_DISPATCH_INDIRECT_BUFFER
    enum: GL_DISPATCH_INDIRECT_BUFFER_BINDING
    enum: GL_COMPUTE_SHADER_BIT
    enum: GL_TEXTURE_VIEW_MIN_LEVEL
    enum: GL_TEXTURE_VIEW_NUM_LEVELS
    enum: GL_TEXTURE_VIEW_MIN_LAYER
    enum: GL_TEXTURE_VIEW_NUM_LAYERS
    enum: GL_TEXTURE_IMMUTABLE_LEVELS
    enum: GL_VERTEX_ATTRIB_BINDING
    enum: GL_VERTEX_ATTRIB_RELATIVE_OFFSET
    enum: GL_VERTEX_BINDING_DIVISOR
    enum: GL_VERTEX_BINDING_OFFSET
    enum: GL_VERTEX_BINDING_STRIDE
    enum: GL_MAX_VERTEX_ATTRIB_RELATIVE_OFFSET
    enum: GL_MAX_VERTEX_ATTRIB_BINDINGS
    enum: GL_COMPRESSED_RGB8_ETC2
    enum: GL_COMPRESSED_SRGB8_ETC2
    enum: GL_COMPRESSED_RGB8_PUNCHTHROUGH_ALPHA1_ETC2
    enum: GL_COMPRESSED_SRGB8_PUNCHTHROUGH_ALPHA1_ETC2
    enum: GL_COMPRESSED_RGBA8_ETC2_EAC
    enum: GL_COMPRESSED_SRGB8_ALPHA8_ETC2_EAC
    enum: GL_COMPRESSED_R11_EAC
    enum: GL_COMPRESSED_SIGNED_R11_EAC
    enum: GL_COMPRESSED_RG11_EAC
    enum: GL_COMPRESSED_SIGNED_RG11_EAC
    enum: GL_PRIMITIVE_RESTART_FIXED_INDEX
    enum: GL_ANY_SAMPLES_PASSED_CONSERVATIVE
    enum: GL_MAX_ELEMENT_INDEX
    enum: GL_MAX_UNIFORM_LOCATIONS
    enum: GL_FRAMEBUFFER_DEFAULT_WIDTH
    enum: GL_FRAMEBUFFER_DEFAULT_HEIGHT
    enum: GL_FRAMEBUFFER_DEFAULT_LAYERS
    enum: GL_FRAMEBUFFER_DEFAULT_SAMPLES
    enum: GL_FRAMEBUFFER_DEFAULT_FIXED_SAMPLE_LOCATIONS
    enum: GL_MAX_FRAMEBUFFER_WIDTH
    enum: GL_MAX_FRAMEBUFFER_HEIGHT
    enum: GL_MAX_FRAMEBUFFER_LAYERS
    enum: GL_MAX_FRAMEBUFFER_SAMPLES
    enum: GL_INTERNALFORMAT_SUPPORTED
    enum: GL_INTERNALFORMAT_PREFERRED
    enum: GL_INTERNALFORMAT_RED_SIZE
    enum: GL_INTERNALFORMAT_GREEN_SIZE
    enum: GL_INTERNALFORMAT_BLUE_SIZE
    enum: GL_INTERNALFORMAT_ALPHA_SIZE
    enum: GL_INTERNALFORMAT_DEPTH_SIZE
    enum: GL_INTERNALFORMAT_STENCIL_SIZE
    enum: GL_INTERNALFORMAT_SHARED_SIZE
    enum: GL_INTERNALFORMAT_RED_TYPE
    enum: GL_INTERNALFORMAT_GREEN_TYPE
    enum: GL_INTERNALFORMAT_BLUE_TYPE
    enum: GL_INTERNALFORMAT_ALPHA_TYPE
    enum: GL_INTERNALFORMAT_DEPTH_TYPE
    enum: GL_INTERNALFORMAT_STENCIL_TYPE
    enum: GL_MAX_WIDTH
    enum: GL_MAX_HEIGHT
    enum: GL_MAX_DEPTH
    enum: GL_MAX_LAYERS
    enum: GL_MAX_COMBINED_DIMENSIONS
    enum: GL_COLOR_COMPONENTS
    enum: GL_DEPTH_COMPONENTS
    enum: GL_STENCIL_COMPONENTS
    enum: GL_COLOR_RENDERABLE
    enum: GL_DEPTH_RENDERABLE
    enum: GL_STENCIL_RENDERABLE
    enum: GL_FRAMEBUFFER_RENDERABLE
    enum: GL_FRAMEBUFFER_RENDERABLE_LAYERED
    enum: GL_FRAMEBUFFER_BLEND
    enum: GL_READ_PIXELS
    enum: GL_READ_PIXELS_FORMAT
    enum: GL_READ_PIXELS_TYPE
    enum: GL_TEXTURE_IMAGE_FORMAT
    enum: GL_TEXTURE_IMAGE_TYPE
    enum: GL_GET_TEXTURE_IMAGE_FORMAT
    enum: GL_GET_TEXTURE_IMAGE_TYPE
    enum: GL_MIPMAP
    enum: GL_MANUAL_GENERATE_MIPMAP
    enum: GL_AUTO_GENERATE_MIPMAP
    enum: GL_COLOR_ENCODING
    enum: GL_SRGB_READ
    enum: GL_SRGB_WRITE
    enum: GL_SRGB_DECODE_ARB
    enum: GL_FILTER
    enum: GL_VERTEX_TEXTURE
    enum: GL_TESS_CONTROL_TEXTURE
    enum: GL_TESS_EVALUATION_TEXTURE
    enum: GL_GEOMETRY_TEXTURE
    enum: GL_FRAGMENT_TEXTURE
    enum: GL_COMPUTE_TEXTURE
    enum: GL_TEXTURE_SHADOW
    enum: GL_TEXTURE_GATHER
    enum: GL_TEXTURE_GATHER_SHADOW
    enum: GL_SHADER_IMAGE_LOAD
    enum: GL_SHADER_IMAGE_STORE
    enum: GL_SHADER_IMAGE_ATOMIC
    enum: GL_IMAGE_TEXEL_SIZE
    enum: GL_IMAGE_COMPATIBILITY_CLASS
    enum: GL_IMAGE_PIXEL_FORMAT
    enum: GL_IMAGE_PIXEL_TYPE
    enum: GL_SIMULTANEOUS_TEXTURE_AND_DEPTH_TEST
    enum: GL_SIMULTANEOUS_TEXTURE_AND_STENCIL_TEST
    enum: GL_SIMULTANEOUS_TEXTURE_AND_DEPTH_WRITE
    enum: GL_SIMULTANEOUS_TEXTURE_AND_STENCIL_WRITE
    enum: GL_TEXTURE_COMPRESSED_BLOCK_WIDTH
    enum: GL_TEXTURE_COMPRESSED_BLOCK_HEIGHT
    enum: GL_TEXTURE_COMPRESSED_BLOCK_SIZE
    enum: GL_CLEAR_BUFFER
    enum: GL_TEXTURE_VIEW
    enum: GL_VIEW_COMPATIBILITY_CLASS
    enum: GL_FULL_SUPPORT
    enum: GL_CAVEAT_SUPPORT
    enum: GL_IMAGE_CLASS_4_X_32
    enum: GL_IMAGE_CLASS_2_X_32
    enum: GL_IMAGE_CLASS_1_X_32
    enum: GL_IMAGE_CLASS_4_X_16
    enum: GL_IMAGE_CLASS_2_X_16
    enum: GL_IMAGE_CLASS_1_X_16
    enum: GL_IMAGE_CLASS_4_X_8
    enum: GL_IMAGE_CLASS_2_X_8
    enum: GL_IMAGE_CLASS_1_X_8
    enum: GL_IMAGE_CLASS_11_11_10
    enum: GL_IMAGE_CLASS_10_10_10_2
    enum: GL_VIEW_CLASS_128_BITS
    enum: GL_VIEW_CLASS_96_BITS
    enum: GL_VIEW_CLASS_64_BITS
    enum: GL_VIEW_CLASS_48_BITS
    enum: GL_VIEW_CLASS_32_BITS
    enum: GL_VIEW_CLASS_24_BITS
    enum: GL_VIEW_CLASS_16_BITS
    enum: GL_VIEW_CLASS_8_BITS
    enum: GL_VIEW_CLASS_S3TC_DXT1_RGB
    enum: GL_VIEW_CLASS_S3TC_DXT1_RGBA
    enum: GL_VIEW_CLASS_S3TC_DXT3_RGBA
    enum: GL_VIEW_CLASS_S3TC_DXT5_RGBA
    enum: GL_VIEW_CLASS_RGTC1_RED
    enum: GL_VIEW_CLASS_RGTC2_RG
    enum: GL_VIEW_CLASS_BPTC_UNORM
    enum: GL_VIEW_CLASS_BPTC_FLOAT
    enum: GL_UNIFORM
    enum: GL_UNIFORM_BLOCK
    enum: GL_PROGRAM_INPUT
    enum: GL_PROGRAM_OUTPUT
    enum: GL_BUFFER_VARIABLE
    enum: GL_SHADER_STORAGE_BLOCK
    enum: GL_VERTEX_SUBROUTINE
    enum: GL_TESS_CONTROL_SUBROUTINE
    enum: GL_TESS_EVALUATION_SUBROUTINE
    enum: GL_GEOMETRY_SUBROUTINE
    enum: GL_FRAGMENT_SUBROUTINE
    enum: GL_COMPUTE_SUBROUTINE
    enum: GL_VERTEX_SUBROUTINE_UNIFORM
    enum: GL_TESS_CONTROL_SUBROUTINE_UNIFORM
    enum: GL_TESS_EVALUATION_SUBROUTINE_UNIFORM
    enum: GL_GEOMETRY_SUBROUTINE_UNIFORM
    enum: GL_FRAGMENT_SUBROUTINE_UNIFORM
    enum: GL_COMPUTE_SUBROUTINE_UNIFORM
    enum: GL_TRANSFORM_FEEDBACK_VARYING
    enum: GL_ACTIVE_RESOURCES
    enum: GL_MAX_NAME_LENGTH
    enum: GL_MAX_NUM_ACTIVE_VARIABLES
    enum: GL_MAX_NUM_COMPATIBLE_SUBROUTINES
    enum: GL_NAME_LENGTH
    enum: GL_TYPE
    enum: GL_ARRAY_SIZE
    enum: GL_OFFSET
    enum: GL_BLOCK_INDEX
    enum: GL_ARRAY_STRIDE
    enum: GL_MATRIX_STRIDE
    enum: GL_IS_ROW_MAJOR
    enum: GL_ATOMIC_COUNTER_BUFFER_INDEX
    enum: GL_BUFFER_BINDING
    enum: GL_BUFFER_DATA_SIZE
    enum: GL_NUM_ACTIVE_VARIABLES
    enum: GL_ACTIVE_VARIABLES
    enum: GL_REFERENCED_BY_VERTEX_SHADER
    enum: GL_REFERENCED_BY_TESS_CONTROL_SHADER
    enum: GL_REFERENCED_BY_TESS_EVALUATION_SHADER
    enum: GL_REFERENCED_BY_GEOMETRY_SHADER
    enum: GL_REFERENCED_BY_FRAGMENT_SHADER
    enum: GL_REFERENCED_BY_COMPUTE_SHADER
    enum: GL_TOP_LEVEL_ARRAY_SIZE
    enum: GL_TOP_LEVEL_ARRAY_STRIDE
    enum: GL_LOCATION
    enum: GL_LOCATION_INDEX
    enum: GL_IS_PER_PATCH
    enum: GL_SHADER_STORAGE_BUFFER
    enum: GL_SHADER_STORAGE_BUFFER_BINDING
    enum: GL_SHADER_STORAGE_BUFFER_START
    enum: GL_SHADER_STORAGE_BUFFER_SIZE
    enum: GL_MAX_VERTEX_SHADER_STORAGE_BLOCKS
    enum: GL_MAX_GEOMETRY_SHADER_STORAGE_BLOCKS
    enum: GL_MAX_TESS_CONTROL_SHADER_STORAGE_BLOCKS
    enum: GL_MAX_TESS_EVALUATION_SHADER_STORAGE_BLOCKS
    enum: GL_MAX_FRAGMENT_SHADER_STORAGE_BLOCKS
    enum: GL_MAX_COMPUTE_SHADER_STORAGE_BLOCKS
    enum: GL_MAX_COMBINED_SHADER_STORAGE_BLOCKS
    enum: GL_MAX_SHADER_STORAGE_BUFFER_BINDINGS
    enum: GL_MAX_SHADER_STORAGE_BLOCK_SIZE
    enum: GL_SHADER_STORAGE_BUFFER_OFFSET_ALIGNMENT
    enum: GL_SHADER_STORAGE_BARRIER_BIT
    enum: GL_MAX_COMBINED_SHADER_OUTPUT_RESOURCES
    enum: GL_DEPTH_STENCIL_TEXTURE_MODE
    enum: GL_TEXTURE_BUFFER_OFFSET
    enum: GL_TEXTURE_BUFFER_SIZE
    enum: GL_TEXTURE_BUFFER_OFFSET_ALIGNMENT
    enum: GL_ABGR_EXT
    enum: GL_CONSTANT_COLOR_EXT
    enum: GL_ONE_MINUS_CONSTANT_COLOR_EXT
    enum: GL_CONSTANT_ALPHA_EXT
    enum: GL_ONE_MINUS_CONSTANT_ALPHA_EXT
    enum: GL_BLEND_COLOR_EXT
    enum: GL_POLYGON_OFFSET_EXT
    enum: GL_POLYGON_OFFSET_FACTOR_EXT
    enum: GL_POLYGON_OFFSET_BIAS_EXT
    enum: GL_ALPHA4_EXT
    enum: GL_ALPHA8_EXT
    enum: GL_ALPHA12_EXT
    enum: GL_ALPHA16_EXT
    enum: GL_LUMINANCE4_EXT
    enum: GL_LUMINANCE8_EXT
    enum: GL_LUMINANCE12_EXT
    enum: GL_LUMINANCE16_EXT
    enum: GL_LUMINANCE4_ALPHA4_EXT
    enum: GL_LUMINANCE6_ALPHA2_EXT
    enum: GL_LUMINANCE8_ALPHA8_EXT
    enum: GL_LUMINANCE12_ALPHA4_EXT
    enum: GL_LUMINANCE12_ALPHA12_EXT
    enum: GL_LUMINANCE16_ALPHA16_EXT
    enum: GL_INTENSITY_EXT
    enum: GL_INTENSITY4_EXT
    enum: GL_INTENSITY8_EXT
    enum: GL_INTENSITY12_EXT
    enum: GL_INTENSITY16_EXT
    enum: GL_RGB2_EXT
    enum: GL_RGB4_EXT
    enum: GL_RGB5_EXT
    enum: GL_RGB8_EXT
    enum: GL_RGB10_EXT
    enum: GL_RGB12_EXT
    enum: GL_RGB16_EXT
    enum: GL_RGBA2_EXT
    enum: GL_RGBA4_EXT
    enum: GL_RGB5_A1_EXT
    enum: GL_RGBA8_EXT
    enum: GL_RGB10_A2_EXT
    enum: GL_RGBA12_EXT
    enum: GL_RGBA16_EXT
    enum: GL_TEXTURE_RED_SIZE_EXT
    enum: GL_TEXTURE_GREEN_SIZE_EXT
    enum: GL_TEXTURE_BLUE_SIZE_EXT
    enum: GL_TEXTURE_ALPHA_SIZE_EXT
    enum: GL_TEXTURE_LUMINANCE_SIZE_EXT
    enum: GL_TEXTURE_INTENSITY_SIZE_EXT
    enum: GL_REPLACE_EXT
    enum: GL_PROXY_TEXTURE_1D_EXT
    enum: GL_PROXY_TEXTURE_2D_EXT
    enum: GL_TEXTURE_TOO_LARGE_EXT
    enum: GL_PACK_SKIP_IMAGES_EXT
    enum: GL_PACK_IMAGE_HEIGHT_EXT
    enum: GL_UNPACK_SKIP_IMAGES_EXT
    enum: GL_UNPACK_IMAGE_HEIGHT_EXT
    enum: GL_TEXTURE_3D_EXT
    enum: GL_PROXY_TEXTURE_3D_EXT
    enum: GL_TEXTURE_DEPTH_EXT
    enum: GL_TEXTURE_WRAP_R_EXT
    enum: GL_MAX_3D_TEXTURE_SIZE_EXT
    enum: GL_FILTER4_SGIS
    enum: GL_TEXTURE_FILTER4_SIZE_SGIS
    enum: GL_HISTOGRAM_EXT
    enum: GL_PROXY_HISTOGRAM_EXT
    enum: GL_HISTOGRAM_WIDTH_EXT
    enum: GL_HISTOGRAM_FORMAT_EXT
    enum: GL_HISTOGRAM_RED_SIZE_EXT
    enum: GL_HISTOGRAM_GREEN_SIZE_EXT
    enum: GL_HISTOGRAM_BLUE_SIZE_EXT
    enum: GL_HISTOGRAM_ALPHA_SIZE_EXT
    enum: GL_HISTOGRAM_LUMINANCE_SIZE_EXT
    enum: GL_HISTOGRAM_SINK_EXT
    enum: GL_MINMAX_EXT
    enum: GL_MINMAX_FORMAT_EXT
    enum: GL_MINMAX_SINK_EXT
    enum: GL_TABLE_TOO_LARGE_EXT
    enum: GL_CONVOLUTION_1D_EXT
    enum: GL_CONVOLUTION_2D_EXT
    enum: GL_SEPARABLE_2D_EXT
    enum: GL_CONVOLUTION_BORDER_MODE_EXT
    enum: GL_CONVOLUTION_FILTER_SCALE_EXT
    enum: GL_CONVOLUTION_FILTER_BIAS_EXT
    enum: GL_REDUCE_EXT
    enum: GL_CONVOLUTION_FORMAT_EXT
    enum: GL_CONVOLUTION_WIDTH_EXT
    enum: GL_CONVOLUTION_HEIGHT_EXT
    enum: GL_MAX_CONVOLUTION_WIDTH_EXT
    enum: GL_MAX_CONVOLUTION_HEIGHT_EXT
    enum: GL_POST_CONVOLUTION_RED_SCALE_EXT
    enum: GL_POST_CONVOLUTION_GREEN_SCALE_EXT
    enum: GL_POST_CONVOLUTION_BLUE_SCALE_EXT
    enum: GL_POST_CONVOLUTION_ALPHA_SCALE_EXT
    enum: GL_POST_CONVOLUTION_RED_BIAS_EXT
    enum: GL_POST_CONVOLUTION_GREEN_BIAS_EXT
    enum: GL_POST_CONVOLUTION_BLUE_BIAS_EXT
    enum: GL_POST_CONVOLUTION_ALPHA_BIAS_EXT
    enum: GL_COLOR_MATRIX_SGI
    enum: GL_COLOR_MATRIX_STACK_DEPTH_SGI
    enum: GL_MAX_COLOR_MATRIX_STACK_DEPTH_SGI
    enum: GL_POST_COLOR_MATRIX_RED_SCALE_SGI
    enum: GL_POST_COLOR_MATRIX_GREEN_SCALE_SGI
    enum: GL_POST_COLOR_MATRIX_BLUE_SCALE_SGI
    enum: GL_POST_COLOR_MATRIX_ALPHA_SCALE_SGI
    enum: GL_POST_COLOR_MATRIX_RED_BIAS_SGI
    enum: GL_POST_COLOR_MATRIX_GREEN_BIAS_SGI
    enum: GL_POST_COLOR_MATRIX_BLUE_BIAS_SGI
    enum: GL_POST_COLOR_MATRIX_ALPHA_BIAS_SGI
    enum: GL_COLOR_TABLE_SGI
    enum: GL_POST_CONVOLUTION_COLOR_TABLE_SGI
    enum: GL_POST_COLOR_MATRIX_COLOR_TABLE_SGI
    enum: GL_PROXY_COLOR_TABLE_SGI
    enum: GL_PROXY_POST_CONVOLUTION_COLOR_TABLE_SGI
    enum: GL_PROXY_POST_COLOR_MATRIX_COLOR_TABLE_SGI
    enum: GL_COLOR_TABLE_SCALE_SGI
    enum: GL_COLOR_TABLE_BIAS_SGI
    enum: GL_COLOR_TABLE_FORMAT_SGI
    enum: GL_COLOR_TABLE_WIDTH_SGI
    enum: GL_COLOR_TABLE_RED_SIZE_SGI
    enum: GL_COLOR_TABLE_GREEN_SIZE_SGI
    enum: GL_COLOR_TABLE_BLUE_SIZE_SGI
    enum: GL_COLOR_TABLE_ALPHA_SIZE_SGI
    enum: GL_COLOR_TABLE_LUMINANCE_SIZE_SGI
    enum: GL_COLOR_TABLE_INTENSITY_SIZE_SGI
    enum: GL_PIXEL_TEXTURE_SGIS
    enum: GL_PIXEL_FRAGMENT_RGB_SOURCE_SGIS
    enum: GL_PIXEL_FRAGMENT_ALPHA_SOURCE_SGIS
    enum: GL_PIXEL_GROUP_COLOR_SGIS
    enum: GL_PIXEL_TEX_GEN_SGIX
    enum: GL_PIXEL_TEX_GEN_MODE_SGIX
    enum: GL_PACK_SKIP_VOLUMES_SGIS
    enum: GL_PACK_IMAGE_DEPTH_SGIS
    enum: GL_UNPACK_SKIP_VOLUMES_SGIS
    enum: GL_UNPACK_IMAGE_DEPTH_SGIS
    enum: GL_TEXTURE_4D_SGIS
    enum: GL_PROXY_TEXTURE_4D_SGIS
    enum: GL_TEXTURE_4DSIZE_SGIS
    enum: GL_TEXTURE_WRAP_Q_SGIS
    enum: GL_MAX_4D_TEXTURE_SIZE_SGIS
    enum: GL_TEXTURE_4D_BINDING_SGIS
    enum: GL_TEXTURE_COLOR_TABLE_SGI
    enum: GL_PROXY_TEXTURE_COLOR_TABLE_SGI
    enum: GL_CMYK_EXT
    enum: GL_CMYKA_EXT
    enum: GL_PACK_CMYK_HINT_EXT
    enum: GL_UNPACK_CMYK_HINT_EXT
    enum: GL_TEXTURE_PRIORITY_EXT
    enum: GL_TEXTURE_RESIDENT_EXT
    enum: GL_TEXTURE_1D_BINDING_EXT
    enum: GL_TEXTURE_2D_BINDING_EXT
    enum: GL_TEXTURE_3D_BINDING_EXT
    enum: GL_DETAIL_TEXTURE_2D_SGIS
    enum: GL_DETAIL_TEXTURE_2D_BINDING_SGIS
    enum: GL_LINEAR_DETAIL_SGIS
    enum: GL_LINEAR_DETAIL_ALPHA_SGIS
    enum: GL_LINEAR_DETAIL_COLOR_SGIS
    enum: GL_DETAIL_TEXTURE_LEVEL_SGIS
    enum: GL_DETAIL_TEXTURE_MODE_SGIS
    enum: GL_DETAIL_TEXTURE_FUNC_POINTS_SGIS
    enum: GL_LINEAR_SHARPEN_SGIS
    enum: GL_LINEAR_SHARPEN_ALPHA_SGIS
    enum: GL_LINEAR_SHARPEN_COLOR_SGIS
    enum: GL_SHARPEN_TEXTURE_FUNC_POINTS_SGIS
    enum: GL_UNSIGNED_BYTE_3_3_2_EXT
    enum: GL_UNSIGNED_SHORT_4_4_4_4_EXT
    enum: GL_UNSIGNED_SHORT_5_5_5_1_EXT
    enum: GL_UNSIGNED_INT_8_8_8_8_EXT
    enum: GL_UNSIGNED_INT_10_10_10_2_EXT
    enum: GL_TEXTURE_MIN_LOD_SGIS
    enum: GL_TEXTURE_MAX_LOD_SGIS
    enum: GL_TEXTURE_BASE_LEVEL_SGIS
    enum: GL_TEXTURE_MAX_LEVEL_SGIS
    enum: GL_MULTISAMPLE_SGIS
    enum: GL_SAMPLE_ALPHA_TO_MASK_SGIS
    enum: GL_SAMPLE_ALPHA_TO_ONE_SGIS
    enum: GL_SAMPLE_MASK_SGIS
    enum: GL_1PASS_SGIS
    enum: GL_2PASS_0_SGIS
    enum: GL_2PASS_1_SGIS
    enum: GL_4PASS_0_SGIS
    enum: GL_4PASS_1_SGIS
    enum: GL_4PASS_2_SGIS
    enum: GL_4PASS_3_SGIS
    enum: GL_SAMPLE_BUFFERS_SGIS
    enum: GL_SAMPLES_SGIS
    enum: GL_SAMPLE_MASK_VALUE_SGIS
    enum: GL_SAMPLE_MASK_INVERT_SGIS
    enum: GL_SAMPLE_PATTERN_SGIS
    enum: GL_RESCALE_NORMAL_EXT
    enum: GL_VERTEX_ARRAY_EXT
    enum: GL_NORMAL_ARRAY_EXT
    enum: GL_COLOR_ARRAY_EXT
    enum: GL_INDEX_ARRAY_EXT
    enum: GL_TEXTURE_COORD_ARRAY_EXT
    enum: GL_EDGE_FLAG_ARRAY_EXT
    enum: GL_VERTEX_ARRAY_SIZE_EXT
    enum: GL_VERTEX_ARRAY_TYPE_EXT
    enum: GL_VERTEX_ARRAY_STRIDE_EXT
    enum: GL_VERTEX_ARRAY_COUNT_EXT
    enum: GL_NORMAL_ARRAY_TYPE_EXT
    enum: GL_NORMAL_ARRAY_STRIDE_EXT
    enum: GL_NORMAL_ARRAY_COUNT_EXT
    enum: GL_COLOR_ARRAY_SIZE_EXT
    enum: GL_COLOR_ARRAY_TYPE_EXT
    enum: GL_COLOR_ARRAY_STRIDE_EXT
    enum: GL_COLOR_ARRAY_COUNT_EXT
    enum: GL_INDEX_ARRAY_TYPE_EXT
    enum: GL_INDEX_ARRAY_STRIDE_EXT
    enum: GL_INDEX_ARRAY_COUNT_EXT
    enum: GL_TEXTURE_COORD_ARRAY_SIZE_EXT
    enum: GL_TEXTURE_COORD_ARRAY_TYPE_EXT
    enum: GL_TEXTURE_COORD_ARRAY_STRIDE_EXT
    enum: GL_TEXTURE_COORD_ARRAY_COUNT_EXT
    enum: GL_EDGE_FLAG_ARRAY_STRIDE_EXT
    enum: GL_EDGE_FLAG_ARRAY_COUNT_EXT
    enum: GL_VERTEX_ARRAY_POINTER_EXT
    enum: GL_NORMAL_ARRAY_POINTER_EXT
    enum: GL_COLOR_ARRAY_POINTER_EXT
    enum: GL_INDEX_ARRAY_POINTER_EXT
    enum: GL_TEXTURE_COORD_ARRAY_POINTER_EXT
    enum: GL_EDGE_FLAG_ARRAY_POINTER_EXT
    enum: GL_GENERATE_MIPMAP_SGIS
    enum: GL_GENERATE_MIPMAP_HINT_SGIS
    enum: GL_LINEAR_CLIPMAP_LINEAR_SGIX
    enum: GL_TEXTURE_CLIPMAP_CENTER_SGIX
    enum: GL_TEXTURE_CLIPMAP_FRAME_SGIX
    enum: GL_TEXTURE_CLIPMAP_OFFSET_SGIX
    enum: GL_TEXTURE_CLIPMAP_VIRTUAL_DEPTH_SGIX
    enum: GL_TEXTURE_CLIPMAP_LOD_OFFSET_SGIX
    enum: GL_TEXTURE_CLIPMAP_DEPTH_SGIX
    enum: GL_MAX_CLIPMAP_DEPTH_SGIX
    enum: GL_MAX_CLIPMAP_VIRTUAL_DEPTH_SGIX
    enum: GL_NEAREST_CLIPMAP_NEAREST_SGIX
    enum: GL_NEAREST_CLIPMAP_LINEAR_SGIX
    enum: GL_LINEAR_CLIPMAP_NEAREST_SGIX
    enum: GL_TEXTURE_COMPARE_SGIX
    enum: GL_TEXTURE_COMPARE_OPERATOR_SGIX
    enum: GL_TEXTURE_LEQUAL_R_SGIX
    enum: GL_TEXTURE_GEQUAL_R_SGIX
    enum: GL_CLAMP_TO_EDGE_SGIS
    enum: GL_CLAMP_TO_BORDER_SGIS
    enum: GL_FUNC_ADD_EXT
    enum: GL_MIN_EXT
    enum: GL_MAX_EXT
    enum: GL_BLEND_EQUATION_EXT
    enum: GL_FUNC_SUBTRACT_EXT
    enum: GL_FUNC_REVERSE_SUBTRACT_EXT
    enum: GL_INTERLACE_SGIX
    enum: GL_PIXEL_TILE_BEST_ALIGNMENT_SGIX
    enum: GL_PIXEL_TILE_CACHE_INCREMENT_SGIX
    enum: GL_PIXEL_TILE_WIDTH_SGIX
    enum: GL_PIXEL_TILE_HEIGHT_SGIX
    enum: GL_PIXEL_TILE_GRID_WIDTH_SGIX
    enum: GL_PIXEL_TILE_GRID_HEIGHT_SGIX
    enum: GL_PIXEL_TILE_GRID_DEPTH_SGIX
    enum: GL_PIXEL_TILE_CACHE_SIZE_SGIX
    enum: GL_DUAL_ALPHA4_SGIS
    enum: GL_DUAL_ALPHA8_SGIS
    enum: GL_DUAL_ALPHA12_SGIS
    enum: GL_DUAL_ALPHA16_SGIS
    enum: GL_DUAL_LUMINANCE4_SGIS
    enum: GL_DUAL_LUMINANCE8_SGIS
    enum: GL_DUAL_LUMINANCE12_SGIS
    enum: GL_DUAL_LUMINANCE16_SGIS
    enum: GL_DUAL_INTENSITY4_SGIS
    enum: GL_DUAL_INTENSITY8_SGIS
    enum: GL_DUAL_INTENSITY12_SGIS
    enum: GL_DUAL_INTENSITY16_SGIS
    enum: GL_DUAL_LUMINANCE_ALPHA4_SGIS
    enum: GL_DUAL_LUMINANCE_ALPHA8_SGIS
    enum: GL_QUAD_ALPHA4_SGIS
    enum: GL_QUAD_ALPHA8_SGIS
    enum: GL_QUAD_LUMINANCE4_SGIS
    enum: GL_QUAD_LUMINANCE8_SGIS
    enum: GL_QUAD_INTENSITY4_SGIS
    enum: GL_QUAD_INTENSITY8_SGIS
    enum: GL_DUAL_TEXTURE_SELECT_SGIS
    enum: GL_QUAD_TEXTURE_SELECT_SGIS
    enum: GL_SPRITE_SGIX
    enum: GL_SPRITE_MODE_SGIX
    enum: GL_SPRITE_AXIS_SGIX
    enum: GL_SPRITE_TRANSLATION_SGIX
    enum: GL_SPRITE_AXIAL_SGIX
    enum: GL_SPRITE_OBJECT_ALIGNED_SGIX
    enum: GL_SPRITE_EYE_ALIGNED_SGIX
    enum: GL_TEXTURE_MULTI_BUFFER_HINT_SGIX
    enum: GL_POINT_SIZE_MIN_EXT
    enum: GL_POINT_SIZE_MAX_EXT
    enum: GL_POINT_FADE_THRESHOLD_SIZE_EXT
    enum: GL_DISTANCE_ATTENUATION_EXT
    enum: GL_POINT_SIZE_MIN_SGIS
    enum: GL_POINT_SIZE_MAX_SGIS
    enum: GL_POINT_FADE_THRESHOLD_SIZE_SGIS
    enum: GL_DISTANCE_ATTENUATION_SGIS
    enum: GL_INSTRUMENT_BUFFER_POINTER_SGIX
    enum: GL_INSTRUMENT_MEASUREMENTS_SGIX
    enum: GL_POST_TEXTURE_FILTER_BIAS_SGIX
    enum: GL_POST_TEXTURE_FILTER_SCALE_SGIX
    enum: GL_POST_TEXTURE_FILTER_BIAS_RANGE_SGIX
    enum: GL_POST_TEXTURE_FILTER_SCALE_RANGE_SGIX
    enum: GL_FRAMEZOOM_SGIX
    enum: GL_FRAMEZOOM_FACTOR_SGIX
    enum: GL_MAX_FRAMEZOOM_FACTOR_SGIX
    enum: GL_TEXTURE_DEFORMATION_BIT_SGIX
    enum: GL_GEOMETRY_DEFORMATION_BIT_SGIX
    enum: GL_GEOMETRY_DEFORMATION_SGIX
    enum: GL_TEXTURE_DEFORMATION_SGIX
    enum: GL_DEFORMATIONS_MASK_SGIX
    enum: GL_MAX_DEFORMATION_ORDER_SGIX
    enum: GL_REFERENCE_PLANE_SGIX
    enum: GL_REFERENCE_PLANE_EQUATION_SGIX
    enum: GL_DEPTH_COMPONENT16_SGIX
    enum: GL_DEPTH_COMPONENT24_SGIX
    enum: GL_DEPTH_COMPONENT32_SGIX
    enum: GL_FOG_FUNC_SGIS
    enum: GL_FOG_FUNC_POINTS_SGIS
    enum: GL_MAX_FOG_FUNC_POINTS_SGIS
    enum: GL_FOG_OFFSET_SGIX
    enum: GL_FOG_OFFSET_VALUE_SGIX
    enum: GL_IMAGE_SCALE_X_HP
    enum: GL_IMAGE_SCALE_Y_HP
    enum: GL_IMAGE_TRANSLATE_X_HP
    enum: GL_IMAGE_TRANSLATE_Y_HP
    enum: GL_IMAGE_ROTATE_ANGLE_HP
    enum: GL_IMAGE_ROTATE_ORIGIN_X_HP
    enum: GL_IMAGE_ROTATE_ORIGIN_Y_HP
    enum: GL_IMAGE_MAG_FILTER_HP
    enum: GL_IMAGE_MIN_FILTER_HP
    enum: GL_IMAGE_CUBIC_WEIGHT_HP
    enum: GL_CUBIC_HP
    enum: GL_AVERAGE_HP
    enum: GL_IMAGE_TRANSFORM_2D_HP
    enum: GL_POST_IMAGE_TRANSFORM_COLOR_TABLE_HP
    enum: GL_PROXY_POST_IMAGE_TRANSFORM_COLOR_TABLE_HP
    enum: GL_IGNORE_BORDER_HP
    enum: GL_CONSTANT_BORDER_HP
    enum: GL_REPLICATE_BORDER_HP
    enum: GL_CONVOLUTION_BORDER_COLOR_HP
    enum: GL_TEXTURE_ENV_BIAS_SGIX
    enum: GL_VERTEX_DATA_HINT_PGI
    enum: GL_VERTEX_CONSISTENT_HINT_PGI
    enum: GL_MATERIAL_SIDE_HINT_PGI
    enum: GL_MAX_VERTEX_HINT_PGI
    enum: GL_COLOR3_BIT_PGI
    enum: GL_COLOR4_BIT_PGI
    enum: GL_EDGEFLAG_BIT_PGI
    enum: GL_INDEX_BIT_PGI
    enum: GL_MAT_AMBIENT_BIT_PGI
    enum: GL_MAT_AMBIENT_AND_DIFFUSE_BIT_PGI
    enum: GL_MAT_DIFFUSE_BIT_PGI
    enum: GL_MAT_EMISSION_BIT_PGI
    enum: GL_MAT_COLOR_INDEXES_BIT_PGI
    enum: GL_MAT_SHININESS_BIT_PGI
    enum: GL_MAT_SPECULAR_BIT_PGI
    enum: GL_NORMAL_BIT_PGI
    enum: GL_TEXCOORD1_BIT_PGI
    enum: GL_TEXCOORD2_BIT_PGI
    enum: GL_TEXCOORD3_BIT_PGI
    enum: GL_TEXCOORD4_BIT_PGI
    enum: GL_VERTEX23_BIT_PGI
    enum: GL_VERTEX4_BIT_PGI
    enum: GL_PREFER_DOUBLEBUFFER_HINT_PGI
    enum: GL_CONSERVE_MEMORY_HINT_PGI
    enum: GL_RECLAIM_MEMORY_HINT_PGI
    enum: GL_NATIVE_GRAPHICS_HANDLE_PGI
    enum: GL_NATIVE_GRAPHICS_BEGIN_HINT_PGI
    enum: GL_NATIVE_GRAPHICS_END_HINT_PGI
    enum: GL_ALWAYS_FAST_HINT_PGI
    enum: GL_ALWAYS_SOFT_HINT_PGI
    enum: GL_ALLOW_DRAW_OBJ_HINT_PGI
    enum: GL_ALLOW_DRAW_WIN_HINT_PGI
    enum: GL_ALLOW_DRAW_FRG_HINT_PGI
    enum: GL_ALLOW_DRAW_MEM_HINT_PGI
    enum: GL_STRICT_DEPTHFUNC_HINT_PGI
    enum: GL_STRICT_LIGHTING_HINT_PGI
    enum: GL_STRICT_SCISSOR_HINT_PGI
    enum: GL_FULL_STIPPLE_HINT_PGI
    enum: GL_CLIP_NEAR_HINT_PGI
    enum: GL_CLIP_FAR_HINT_PGI
    enum: GL_WIDE_LINE_HINT_PGI
    enum: GL_BACK_NORMALS_HINT_PGI
    enum: GL_COLOR_INDEX1_EXT
    enum: GL_COLOR_INDEX2_EXT
    enum: GL_COLOR_INDEX4_EXT
    enum: GL_COLOR_INDEX8_EXT
    enum: GL_COLOR_INDEX12_EXT
    enum: GL_COLOR_INDEX16_EXT
    enum: GL_TEXTURE_INDEX_SIZE_EXT
    enum: GL_CLIP_VOLUME_CLIPPING_HINT_EXT
    enum: GL_LIST_PRIORITY_SGIX
    enum: GL_IR_INSTRUMENT1_SGIX
    enum: GL_CALLIGRAPHIC_FRAGMENT_SGIX
    enum: GL_TEXTURE_LOD_BIAS_S_SGIX
    enum: GL_TEXTURE_LOD_BIAS_T_SGIX
    enum: GL_TEXTURE_LOD_BIAS_R_SGIX
    enum: GL_SHADOW_AMBIENT_SGIX
    enum: GL_INDEX_MATERIAL_EXT
    enum: GL_INDEX_MATERIAL_PARAMETER_EXT
    enum: GL_INDEX_MATERIAL_FACE_EXT
    enum: GL_INDEX_TEST_EXT
    enum: GL_INDEX_TEST_FUNC_EXT
    enum: GL_INDEX_TEST_REF_EXT
    enum: GL_IUI_V2F_EXT
    enum: GL_IUI_V3F_EXT
    enum: GL_IUI_N3F_V2F_EXT
    enum: GL_IUI_N3F_V3F_EXT
    enum: GL_T2F_IUI_V2F_EXT
    enum: GL_T2F_IUI_V3F_EXT
    enum: GL_T2F_IUI_N3F_V2F_EXT
    enum: GL_T2F_IUI_N3F_V3F_EXT
    enum: GL_ARRAY_ELEMENT_LOCK_FIRST_EXT
    enum: GL_ARRAY_ELEMENT_LOCK_COUNT_EXT
    enum: GL_CULL_VERTEX_EXT
    enum: GL_CULL_VERTEX_EYE_POSITION_EXT
    enum: GL_CULL_VERTEX_OBJECT_POSITION_EXT
    enum: GL_YCRCB_422_SGIX
    enum: GL_YCRCB_444_SGIX
    enum: GL_FRAGMENT_LIGHTING_SGIX
    enum: GL_FRAGMENT_COLOR_MATERIAL_SGIX
    enum: GL_FRAGMENT_COLOR_MATERIAL_FACE_SGIX
    enum: GL_FRAGMENT_COLOR_MATERIAL_PARAMETER_SGIX
    enum: GL_MAX_FRAGMENT_LIGHTS_SGIX
    enum: GL_MAX_ACTIVE_LIGHTS_SGIX
    enum: GL_CURRENT_RASTER_NORMAL_SGIX
    enum: GL_LIGHT_ENV_MODE_SGIX
    enum: GL_FRAGMENT_LIGHT_MODEL_LOCAL_VIEWER_SGIX
    enum: GL_FRAGMENT_LIGHT_MODEL_TWO_SIDE_SGIX
    enum: GL_FRAGMENT_LIGHT_MODEL_AMBIENT_SGIX
    enum: GL_FRAGMENT_LIGHT_MODEL_NORMAL_INTERPOLATION_SGIX
    enum: GL_FRAGMENT_LIGHT0_SGIX
    enum: GL_FRAGMENT_LIGHT1_SGIX
    enum: GL_FRAGMENT_LIGHT2_SGIX
    enum: GL_FRAGMENT_LIGHT3_SGIX
    enum: GL_FRAGMENT_LIGHT4_SGIX
    enum: GL_FRAGMENT_LIGHT5_SGIX
    enum: GL_FRAGMENT_LIGHT6_SGIX
    enum: GL_FRAGMENT_LIGHT7_SGIX
    enum: GL_RASTER_POSITION_UNCLIPPED_IBM
    enum: GL_TEXTURE_LIGHTING_MODE_HP
    enum: GL_TEXTURE_POST_SPECULAR_HP
    enum: GL_TEXTURE_PRE_SPECULAR_HP
    enum: GL_MAX_ELEMENTS_VERTICES_EXT
    enum: GL_MAX_ELEMENTS_INDICES_EXT
    enum: GL_PHONG_WIN
    enum: GL_PHONG_HINT_WIN
    enum: GL_FOG_SPECULAR_TEXTURE_WIN
    enum: GL_FRAGMENT_MATERIAL_EXT
    enum: GL_FRAGMENT_NORMAL_EXT
    enum: GL_FRAGMENT_COLOR_EXT
    enum: GL_ATTENUATION_EXT
    enum: GL_SHADOW_ATTENUATION_EXT
    enum: GL_TEXTURE_APPLICATION_MODE_EXT
    enum: GL_TEXTURE_LIGHT_EXT
    enum: GL_TEXTURE_MATERIAL_FACE_EXT
    enum: GL_TEXTURE_MATERIAL_PARAMETER_EXT
    enum: GL_ALPHA_MIN_SGIX
    enum: GL_ALPHA_MAX_SGIX
    enum: GL_PIXEL_TEX_GEN_Q_CEILING_SGIX
    enum: GL_PIXEL_TEX_GEN_Q_ROUND_SGIX
    enum: GL_PIXEL_TEX_GEN_Q_FLOOR_SGIX
    enum: GL_PIXEL_TEX_GEN_ALPHA_REPLACE_SGIX
    enum: GL_PIXEL_TEX_GEN_ALPHA_NO_REPLACE_SGIX
    enum: GL_PIXEL_TEX_GEN_ALPHA_LS_SGIX
    enum: GL_PIXEL_TEX_GEN_ALPHA_MS_SGIX
    enum: GL_BGR_EXT
    enum: GL_BGRA_EXT
    enum: GL_ASYNC_MARKER_SGIX
    enum: GL_ASYNC_TEX_IMAGE_SGIX
    enum: GL_ASYNC_DRAW_PIXELS_SGIX
    enum: GL_ASYNC_READ_PIXELS_SGIX
    enum: GL_MAX_ASYNC_TEX_IMAGE_SGIX
    enum: GL_MAX_ASYNC_DRAW_PIXELS_SGIX
    enum: GL_MAX_ASYNC_READ_PIXELS_SGIX
    enum: GL_ASYNC_HISTOGRAM_SGIX
    enum: GL_MAX_ASYNC_HISTOGRAM_SGIX
    enum: GL_PARALLEL_ARRAYS_INTEL
    enum: GL_VERTEX_ARRAY_PARALLEL_POINTERS_INTEL
    enum: GL_NORMAL_ARRAY_PARALLEL_POINTERS_INTEL
    enum: GL_COLOR_ARRAY_PARALLEL_POINTERS_INTEL
    enum: GL_TEXTURE_COORD_ARRAY_PARALLEL_POINTERS_INTEL
    enum: GL_OCCLUSION_TEST_HP
    enum: GL_OCCLUSION_TEST_RESULT_HP
    enum: GL_PIXEL_TRANSFORM_2D_EXT
    enum: GL_PIXEL_MAG_FILTER_EXT
    enum: GL_PIXEL_MIN_FILTER_EXT
    enum: GL_PIXEL_CUBIC_WEIGHT_EXT
    enum: GL_CUBIC_EXT
    enum: GL_AVERAGE_EXT
    enum: GL_PIXEL_TRANSFORM_2D_STACK_DEPTH_EXT
    enum: GL_MAX_PIXEL_TRANSFORM_2D_STACK_DEPTH_EXT
    enum: GL_PIXEL_TRANSFORM_2D_MATRIX_EXT
    enum: GL_SHARED_TEXTURE_PALETTE_EXT
    enum: GL_LIGHT_MODEL_COLOR_CONTROL_EXT
    enum: GL_SINGLE_COLOR_EXT
    enum: GL_SEPARATE_SPECULAR_COLOR_EXT
    enum: GL_COLOR_SUM_EXT
    enum: GL_CURRENT_SECONDARY_COLOR_EXT
    enum: GL_SECONDARY_COLOR_ARRAY_SIZE_EXT
    enum: GL_SECONDARY_COLOR_ARRAY_TYPE_EXT
    enum: GL_SECONDARY_COLOR_ARRAY_STRIDE_EXT
    enum: GL_SECONDARY_COLOR_ARRAY_POINTER_EXT
    enum: GL_SECONDARY_COLOR_ARRAY_EXT
    enum: GL_PERTURB_EXT
    enum: GL_TEXTURE_NORMAL_EXT
    enum: GL_FOG_COORDINATE_SOURCE_EXT
    enum: GL_FOG_COORDINATE_EXT
    enum: GL_FRAGMENT_DEPTH_EXT
    enum: GL_CURRENT_FOG_COORDINATE_EXT
    enum: GL_FOG_COORDINATE_ARRAY_TYPE_EXT
    enum: GL_FOG_COORDINATE_ARRAY_STRIDE_EXT
    enum: GL_FOG_COORDINATE_ARRAY_POINTER_EXT
    enum: GL_FOG_COORDINATE_ARRAY_EXT
    enum: GL_SCREEN_COORDINATES_REND
    enum: GL_INVERTED_SCREEN_W_REND
    enum: GL_TANGENT_ARRAY_EXT
    enum: GL_BINORMAL_ARRAY_EXT
    enum: GL_CURRENT_TANGENT_EXT
    enum: GL_CURRENT_BINORMAL_EXT
    enum: GL_TANGENT_ARRAY_TYPE_EXT
    enum: GL_TANGENT_ARRAY_STRIDE_EXT
    enum: GL_BINORMAL_ARRAY_TYPE_EXT
    enum: GL_BINORMAL_ARRAY_STRIDE_EXT
    enum: GL_TANGENT_ARRAY_POINTER_EXT
    enum: GL_BINORMAL_ARRAY_POINTER_EXT
    enum: GL_MAP1_TANGENT_EXT
    enum: GL_MAP2_TANGENT_EXT
    enum: GL_MAP1_BINORMAL_EXT
    enum: GL_MAP2_BINORMAL_EXT
    enum: GL_COMBINE_EXT
    enum: GL_COMBINE_RGB_EXT
    enum: GL_COMBINE_ALPHA_EXT
    enum: GL_RGB_SCALE_EXT
    enum: GL_ADD_SIGNED_EXT
    enum: GL_INTERPOLATE_EXT
    enum: GL_CONSTANT_EXT
    enum: GL_PRIMARY_COLOR_EXT
    enum: GL_PREVIOUS_EXT
    enum: GL_SOURCE0_RGB_EXT
    enum: GL_SOURCE1_RGB_EXT
    enum: GL_SOURCE2_RGB_EXT
    enum: GL_SOURCE0_ALPHA_EXT
    enum: GL_SOURCE1_ALPHA_EXT
    enum: GL_SOURCE2_ALPHA_EXT
    enum: GL_OPERAND0_RGB_EXT
    enum: GL_OPERAND1_RGB_EXT
    enum: GL_OPERAND2_RGB_EXT
    enum: GL_OPERAND0_ALPHA_EXT
    enum: GL_OPERAND1_ALPHA_EXT
    enum: GL_OPERAND2_ALPHA_EXT
    enum: GL_LIGHT_MODEL_SPECULAR_VECTOR_APPLE
    enum: GL_TRANSFORM_HINT_APPLE
    enum: GL_FOG_SCALE_SGIX
    enum: GL_FOG_SCALE_VALUE_SGIX
    enum: GL_UNPACK_CONSTANT_DATA_SUNX
    enum: GL_TEXTURE_CONSTANT_DATA_SUNX
    enum: GL_GLOBAL_ALPHA_SUN
    enum: GL_GLOBAL_ALPHA_FACTOR_SUN
    enum: GL_RESTART_SUN
    enum: GL_REPLACE_MIDDLE_SUN
    enum: GL_REPLACE_OLDEST_SUN
    enum: GL_TRIANGLE_LIST_SUN
    enum: GL_REPLACEMENT_CODE_SUN
    enum: GL_REPLACEMENT_CODE_ARRAY_SUN
    enum: GL_REPLACEMENT_CODE_ARRAY_TYPE_SUN
    enum: GL_REPLACEMENT_CODE_ARRAY_STRIDE_SUN
    enum: GL_REPLACEMENT_CODE_ARRAY_POINTER_SUN
    enum: GL_R1UI_V3F_SUN
    enum: GL_R1UI_C4UB_V3F_SUN
    enum: GL_R1UI_C3F_V3F_SUN
    enum: GL_R1UI_N3F_V3F_SUN
    enum: GL_R1UI_C4F_N3F_V3F_SUN
    enum: GL_R1UI_T2F_V3F_SUN
    enum: GL_R1UI_T2F_N3F_V3F_SUN
    enum: GL_R1UI_T2F_C4F_N3F_V3F_SUN
    enum: GL_BLEND_DST_RGB_EXT
    enum: GL_BLEND_SRC_RGB_EXT
    enum: GL_BLEND_DST_ALPHA_EXT
    enum: GL_BLEND_SRC_ALPHA_EXT
    enum: GL_RED_MIN_CLAMP_INGR
    enum: GL_GREEN_MIN_CLAMP_INGR
    enum: GL_BLUE_MIN_CLAMP_INGR
    enum: GL_ALPHA_MIN_CLAMP_INGR
    enum: GL_RED_MAX_CLAMP_INGR
    enum: GL_GREEN_MAX_CLAMP_INGR
    enum: GL_BLUE_MAX_CLAMP_INGR
    enum: GL_ALPHA_MAX_CLAMP_INGR
    enum: GL_INTERLACE_READ_INGR
    enum: GL_INCR_WRAP_EXT
    enum: GL_DECR_WRAP_EXT
    enum: GL_422_EXT
    enum: GL_422_REV_EXT
    enum: GL_422_AVERAGE_EXT
    enum: GL_422_REV_AVERAGE_EXT
    enum: GL_NORMAL_MAP_NV
    enum: GL_REFLECTION_MAP_NV
    enum: GL_NORMAL_MAP_EXT
    enum: GL_REFLECTION_MAP_EXT
    enum: GL_TEXTURE_CUBE_MAP_EXT
    enum: GL_TEXTURE_BINDING_CUBE_MAP_EXT
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_X_EXT
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_Y_EXT
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_EXT
    enum: GL_TEXTURE_CUBE_MAP_POSITIVE_Z_EXT
    enum: GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_EXT
    enum: GL_PROXY_TEXTURE_CUBE_MAP_EXT
    enum: GL_MAX_CUBE_MAP_TEXTURE_SIZE_EXT
    enum: GL_WRAP_BORDER_SUN
    enum: GL_MAX_TEXTURE_LOD_BIAS_EXT
    enum: GL_TEXTURE_FILTER_CONTROL_EXT
    enum: GL_TEXTURE_LOD_BIAS_EXT
    enum: GL_TEXTURE_MAX_ANISOTROPY_EXT
    enum: GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT
    enum: GL_MODELVIEW0_STACK_DEPTH_EXT
    enum: GL_MODELVIEW1_STACK_DEPTH_EXT
    enum: GL_MODELVIEW0_MATRIX_EXT
    enum: GL_MODELVIEW1_MATRIX_EXT
    enum: GL_VERTEX_WEIGHTING_EXT
    enum: GL_MODELVIEW0_EXT
    enum: GL_MODELVIEW1_EXT
    enum: GL_CURRENT_VERTEX_WEIGHT_EXT
    enum: GL_VERTEX_WEIGHT_ARRAY_EXT
    enum: GL_VERTEX_WEIGHT_ARRAY_SIZE_EXT
    enum: GL_VERTEX_WEIGHT_ARRAY_TYPE_EXT
    enum: GL_VERTEX_WEIGHT_ARRAY_STRIDE_EXT
    enum: GL_VERTEX_WEIGHT_ARRAY_POINTER_EXT
    enum: GL_MAX_SHININESS_NV
    enum: GL_MAX_SPOT_EXPONENT_NV
    enum: GL_VERTEX_ARRAY_RANGE_NV
    enum: GL_VERTEX_ARRAY_RANGE_LENGTH_NV
    enum: GL_VERTEX_ARRAY_RANGE_VALID_NV
    enum: GL_MAX_VERTEX_ARRAY_RANGE_ELEMENT_NV
    enum: GL_VERTEX_ARRAY_RANGE_POINTER_NV
    enum: GL_REGISTER_COMBINERS_NV
    enum: GL_VARIABLE_A_NV
    enum: GL_VARIABLE_B_NV
    enum: GL_VARIABLE_C_NV
    enum: GL_VARIABLE_D_NV
    enum: GL_VARIABLE_E_NV
    enum: GL_VARIABLE_F_NV
    enum: GL_VARIABLE_G_NV
    enum: GL_CONSTANT_COLOR0_NV
    enum: GL_CONSTANT_COLOR1_NV
    enum: GL_PRIMARY_COLOR_NV
    enum: GL_SECONDARY_COLOR_NV
    enum: GL_SPARE0_NV
    enum: GL_SPARE1_NV
    enum: GL_DISCARD_NV
    enum: GL_E_TIMES_F_NV
    enum: GL_SPARE0_PLUS_SECONDARY_COLOR_NV
    enum: GL_UNSIGNED_IDENTITY_NV
    enum: GL_UNSIGNED_INVERT_NV
    enum: GL_EXPAND_NORMAL_NV
    enum: GL_EXPAND_NEGATE_NV
    enum: GL_HALF_BIAS_NORMAL_NV
    enum: GL_HALF_BIAS_NEGATE_NV
    enum: GL_SIGNED_IDENTITY_NV
    enum: GL_SIGNED_NEGATE_NV
    enum: GL_SCALE_BY_TWO_NV
    enum: GL_SCALE_BY_FOUR_NV
    enum: GL_SCALE_BY_ONE_HALF_NV
    enum: GL_BIAS_BY_NEGATIVE_ONE_HALF_NV
    enum: GL_COMBINER_INPUT_NV
    enum: GL_COMBINER_MAPPING_NV
    enum: GL_COMBINER_COMPONENT_USAGE_NV
    enum: GL_COMBINER_AB_DOT_PRODUCT_NV
    enum: GL_COMBINER_CD_DOT_PRODUCT_NV
    enum: GL_COMBINER_MUX_SUM_NV
    enum: GL_COMBINER_SCALE_NV
    enum: GL_COMBINER_BIAS_NV
    enum: GL_COMBINER_AB_OUTPUT_NV
    enum: GL_COMBINER_CD_OUTPUT_NV
    enum: GL_COMBINER_SUM_OUTPUT_NV
    enum: GL_MAX_GENERAL_COMBINERS_NV
    enum: GL_NUM_GENERAL_COMBINERS_NV
    enum: GL_COLOR_SUM_CLAMP_NV
    enum: GL_COMBINER0_NV
    enum: GL_COMBINER1_NV
    enum: GL_COMBINER2_NV
    enum: GL_COMBINER3_NV
    enum: GL_COMBINER4_NV
    enum: GL_COMBINER5_NV
    enum: GL_COMBINER6_NV
    enum: GL_COMBINER7_NV
    enum: GL_FOG_DISTANCE_MODE_NV
    enum: GL_EYE_RADIAL_NV
    enum: GL_EYE_PLANE_ABSOLUTE_NV
    enum: GL_EMBOSS_LIGHT_NV
    enum: GL_EMBOSS_CONSTANT_NV
    enum: GL_EMBOSS_MAP_NV
    enum: GL_COMBINE4_NV
    enum: GL_SOURCE3_RGB_NV
    enum: GL_SOURCE3_ALPHA_NV
    enum: GL_OPERAND3_RGB_NV
    enum: GL_OPERAND3_ALPHA_NV
    enum: GL_COMPRESSED_RGB_S3TC_DXT1_EXT
    enum: GL_COMPRESSED_RGBA_S3TC_DXT1_EXT
    enum: GL_COMPRESSED_RGBA_S3TC_DXT3_EXT
    enum: GL_COMPRESSED_RGBA_S3TC_DXT5_EXT
    enum: GL_CULL_VERTEX_IBM
    enum: GL_VERTEX_ARRAY_LIST_IBM
    enum: GL_NORMAL_ARRAY_LIST_IBM
    enum: GL_COLOR_ARRAY_LIST_IBM
    enum: GL_INDEX_ARRAY_LIST_IBM
    enum: GL_TEXTURE_COORD_ARRAY_LIST_IBM
    enum: GL_EDGE_FLAG_ARRAY_LIST_IBM
    enum: GL_FOG_COORDINATE_ARRAY_LIST_IBM
    enum: GL_SECONDARY_COLOR_ARRAY_LIST_IBM
    enum: GL_VERTEX_ARRAY_LIST_STRIDE_IBM
    enum: GL_NORMAL_ARRAY_LIST_STRIDE_IBM
    enum: GL_COLOR_ARRAY_LIST_STRIDE_IBM
    enum: GL_INDEX_ARRAY_LIST_STRIDE_IBM
    enum: GL_TEXTURE_COORD_ARRAY_LIST_STRIDE_IBM
    enum: GL_EDGE_FLAG_ARRAY_LIST_STRIDE_IBM
    enum: GL_FOG_COORDINATE_ARRAY_LIST_STRIDE_IBM
    enum: GL_SECONDARY_COLOR_ARRAY_LIST_STRIDE_IBM
    enum: GL_PACK_SUBSAMPLE_RATE_SGIX
    enum: GL_UNPACK_SUBSAMPLE_RATE_SGIX
    enum: GL_PIXEL_SUBSAMPLE_4444_SGIX
    enum: GL_PIXEL_SUBSAMPLE_2424_SGIX
    enum: GL_PIXEL_SUBSAMPLE_4242_SGIX
    enum: GL_YCRCB_SGIX
    enum: GL_YCRCBA_SGIX
    enum: GL_DEPTH_PASS_INSTRUMENT_SGIX
    enum: GL_DEPTH_PASS_INSTRUMENT_COUNTERS_SGIX
    enum: GL_DEPTH_PASS_INSTRUMENT_MAX_SGIX
    enum: GL_COMPRESSED_RGB_FXT1_3DFX
    enum: GL_COMPRESSED_RGBA_FXT1_3DFX
    enum: GL_MULTISAMPLE_3DFX
    enum: GL_SAMPLE_BUFFERS_3DFX
    enum: GL_SAMPLES_3DFX
    enum: GL_MULTISAMPLE_BIT_3DFX
    enum: GL_MULTISAMPLE_EXT
    enum: GL_SAMPLE_ALPHA_TO_MASK_EXT
    enum: GL_SAMPLE_ALPHA_TO_ONE_EXT
    enum: GL_SAMPLE_MASK_EXT
    enum: GL_1PASS_EXT
    enum: GL_2PASS_0_EXT
    enum: GL_2PASS_1_EXT
    enum: GL_4PASS_0_EXT
    enum: GL_4PASS_1_EXT
    enum: GL_4PASS_2_EXT
    enum: GL_4PASS_3_EXT
    enum: GL_SAMPLE_BUFFERS_EXT
    enum: GL_SAMPLES_EXT
    enum: GL_SAMPLE_MASK_VALUE_EXT
    enum: GL_SAMPLE_MASK_INVERT_EXT
    enum: GL_SAMPLE_PATTERN_EXT
    enum: GL_MULTISAMPLE_BIT_EXT
    enum: GL_VERTEX_PRECLIP_SGIX
    enum: GL_VERTEX_PRECLIP_HINT_SGIX
    enum: GL_CONVOLUTION_HINT_SGIX
    enum: GL_PACK_RESAMPLE_SGIX
    enum: GL_UNPACK_RESAMPLE_SGIX
    enum: GL_RESAMPLE_REPLICATE_SGIX
    enum: GL_RESAMPLE_ZERO_FILL_SGIX
    enum: GL_RESAMPLE_DECIMATE_SGIX
    enum: GL_EYE_DISTANCE_TO_POINT_SGIS
    enum: GL_OBJECT_DISTANCE_TO_POINT_SGIS
    enum: GL_EYE_DISTANCE_TO_LINE_SGIS
    enum: GL_OBJECT_DISTANCE_TO_LINE_SGIS
    enum: GL_EYE_POINT_SGIS
    enum: GL_OBJECT_POINT_SGIS
    enum: GL_EYE_LINE_SGIS
    enum: GL_OBJECT_LINE_SGIS
    enum: GL_TEXTURE_COLOR_WRITEMASK_SGIS
    enum: GL_DOT3_RGB_EXT
    enum: GL_DOT3_RGBA_EXT
    enum: GL_MIRROR_CLAMP_ATI
    enum: GL_MIRROR_CLAMP_TO_EDGE_ATI
    enum: GL_ALL_COMPLETED_NV
    enum: GL_FENCE_STATUS_NV
    enum: GL_FENCE_CONDITION_NV
    enum: GL_MIRRORED_REPEAT_IBM
    enum: GL_EVAL_2D_NV
    enum: GL_EVAL_TRIANGULAR_2D_NV
    enum: GL_MAP_TESSELLATION_NV
    enum: GL_MAP_ATTRIB_U_ORDER_NV
    enum: GL_MAP_ATTRIB_V_ORDER_NV
    enum: GL_EVAL_FRACTIONAL_TESSELLATION_NV
    enum: GL_EVAL_VERTEX_ATTRIB0_NV
    enum: GL_EVAL_VERTEX_ATTRIB1_NV
    enum: GL_EVAL_VERTEX_ATTRIB2_NV
    enum: GL_EVAL_VERTEX_ATTRIB3_NV
    enum: GL_EVAL_VERTEX_ATTRIB4_NV
    enum: GL_EVAL_VERTEX_ATTRIB5_NV
    enum: GL_EVAL_VERTEX_ATTRIB6_NV
    enum: GL_EVAL_VERTEX_ATTRIB7_NV
    enum: GL_EVAL_VERTEX_ATTRIB8_NV
    enum: GL_EVAL_VERTEX_ATTRIB9_NV
    enum: GL_EVAL_VERTEX_ATTRIB10_NV
    enum: GL_EVAL_VERTEX_ATTRIB11_NV
    enum: GL_EVAL_VERTEX_ATTRIB12_NV
    enum: GL_EVAL_VERTEX_ATTRIB13_NV
    enum: GL_EVAL_VERTEX_ATTRIB14_NV
    enum: GL_EVAL_VERTEX_ATTRIB15_NV
    enum: GL_MAX_MAP_TESSELLATION_NV
    enum: GL_MAX_RATIONAL_EVAL_ORDER_NV
    enum: GL_DEPTH_STENCIL_NV
    enum: GL_UNSIGNED_INT_24_8_NV
    enum: GL_PER_STAGE_CONSTANTS_NV
    enum: GL_TEXTURE_RECTANGLE_NV
    enum: GL_TEXTURE_BINDING_RECTANGLE_NV
    enum: GL_PROXY_TEXTURE_RECTANGLE_NV
    enum: GL_MAX_RECTANGLE_TEXTURE_SIZE_NV
    enum: GL_OFFSET_TEXTURE_RECTANGLE_NV
    enum: GL_OFFSET_TEXTURE_RECTANGLE_SCALE_NV
    enum: GL_DOT_PRODUCT_TEXTURE_RECTANGLE_NV
    enum: GL_RGBA_UNSIGNED_DOT_PRODUCT_MAPPING_NV
    enum: GL_UNSIGNED_INT_S8_S8_8_8_NV
    enum: GL_UNSIGNED_INT_8_8_S8_S8_REV_NV
    enum: GL_DSDT_MAG_INTENSITY_NV
    enum: GL_SHADER_CONSISTENT_NV
    enum: GL_TEXTURE_SHADER_NV
    enum: GL_SHADER_OPERATION_NV
    enum: GL_CULL_MODES_NV
    enum: GL_OFFSET_TEXTURE_MATRIX_NV
    enum: GL_OFFSET_TEXTURE_SCALE_NV
    enum: GL_OFFSET_TEXTURE_BIAS_NV
    enum: GL_OFFSET_TEXTURE_2D_MATRIX_NV
    enum: GL_OFFSET_TEXTURE_2D_SCALE_NV
    enum: GL_OFFSET_TEXTURE_2D_BIAS_NV
    enum: GL_PREVIOUS_TEXTURE_INPUT_NV
    enum: GL_CONST_EYE_NV
    enum: GL_PASS_THROUGH_NV
    enum: GL_CULL_FRAGMENT_NV
    enum: GL_OFFSET_TEXTURE_2D_NV
    enum: GL_DEPENDENT_AR_TEXTURE_2D_NV
    enum: GL_DEPENDENT_GB_TEXTURE_2D_NV
    enum: GL_DOT_PRODUCT_NV
    enum: GL_DOT_PRODUCT_DEPTH_REPLACE_NV
    enum: GL_DOT_PRODUCT_TEXTURE_2D_NV
    enum: GL_DOT_PRODUCT_TEXTURE_CUBE_MAP_NV
    enum: GL_DOT_PRODUCT_DIFFUSE_CUBE_MAP_NV
    enum: GL_DOT_PRODUCT_REFLECT_CUBE_MAP_NV
    enum: GL_DOT_PRODUCT_CONST_EYE_REFLECT_CUBE_MAP_NV
    enum: GL_HILO_NV
    enum: GL_DSDT_NV
    enum: GL_DSDT_MAG_NV
    enum: GL_DSDT_MAG_VIB_NV
    enum: GL_HILO16_NV
    enum: GL_SIGNED_HILO_NV
    enum: GL_SIGNED_HILO16_NV
    enum: GL_SIGNED_RGBA_NV
    enum: GL_SIGNED_RGBA8_NV
    enum: GL_SIGNED_RGB_NV
    enum: GL_SIGNED_RGB8_NV
    enum: GL_SIGNED_LUMINANCE_NV
    enum: GL_SIGNED_LUMINANCE8_NV
    enum: GL_SIGNED_LUMINANCE_ALPHA_NV
    enum: GL_SIGNED_LUMINANCE8_ALPHA8_NV
    enum: GL_SIGNED_ALPHA_NV
    enum: GL_SIGNED_ALPHA8_NV
    enum: GL_SIGNED_INTENSITY_NV
    enum: GL_SIGNED_INTENSITY8_NV
    enum: GL_DSDT8_NV
    enum: GL_DSDT8_MAG8_NV
    enum: GL_DSDT8_MAG8_INTENSITY8_NV
    enum: GL_SIGNED_RGB_UNSIGNED_ALPHA_NV
    enum: GL_SIGNED_RGB8_UNSIGNED_ALPHA8_NV
    enum: GL_HI_SCALE_NV
    enum: GL_LO_SCALE_NV
    enum: GL_DS_SCALE_NV
    enum: GL_DT_SCALE_NV
    enum: GL_MAGNITUDE_SCALE_NV
    enum: GL_VIBRANCE_SCALE_NV
    enum: GL_HI_BIAS_NV
    enum: GL_LO_BIAS_NV
    enum: GL_DS_BIAS_NV
    enum: GL_DT_BIAS_NV
    enum: GL_MAGNITUDE_BIAS_NV
    enum: GL_VIBRANCE_BIAS_NV
    enum: GL_TEXTURE_BORDER_VALUES_NV
    enum: GL_TEXTURE_HI_SIZE_NV
    enum: GL_TEXTURE_LO_SIZE_NV
    enum: GL_TEXTURE_DS_SIZE_NV
    enum: GL_TEXTURE_DT_SIZE_NV
    enum: GL_TEXTURE_MAG_SIZE_NV
    enum: GL_DOT_PRODUCT_TEXTURE_3D_NV
    enum: GL_VERTEX_ARRAY_RANGE_WITHOUT_FLUSH_NV
    enum: GL_VERTEX_PROGRAM_NV
    enum: GL_VERTEX_STATE_PROGRAM_NV
    enum: GL_ATTRIB_ARRAY_SIZE_NV
    enum: GL_ATTRIB_ARRAY_STRIDE_NV
    enum: GL_ATTRIB_ARRAY_TYPE_NV
    enum: GL_CURRENT_ATTRIB_NV
    enum: GL_PROGRAM_LENGTH_NV
    enum: GL_PROGRAM_STRING_NV
    enum: GL_MODELVIEW_PROJECTION_NV
    enum: GL_IDENTITY_NV
    enum: GL_INVERSE_NV
    enum: GL_TRANSPOSE_NV
    enum: GL_INVERSE_TRANSPOSE_NV
    enum: GL_MAX_TRACK_MATRIX_STACK_DEPTH_NV
    enum: GL_MAX_TRACK_MATRICES_NV
    enum: GL_MATRIX0_NV
    enum: GL_MATRIX1_NV
    enum: GL_MATRIX2_NV
    enum: GL_MATRIX3_NV
    enum: GL_MATRIX4_NV
    enum: GL_MATRIX5_NV
    enum: GL_MATRIX6_NV
    enum: GL_MATRIX7_NV
    enum: GL_CURRENT_MATRIX_STACK_DEPTH_NV
    enum: GL_CURRENT_MATRIX_NV
    enum: GL_VERTEX_PROGRAM_POINT_SIZE_NV
    enum: GL_VERTEX_PROGRAM_TWO_SIDE_NV
    enum: GL_PROGRAM_PARAMETER_NV
    enum: GL_ATTRIB_ARRAY_POINTER_NV
    enum: GL_PROGRAM_TARGET_NV
    enum: GL_PROGRAM_RESIDENT_NV
    enum: GL_TRACK_MATRIX_NV
    enum: GL_TRACK_MATRIX_TRANSFORM_NV
    enum: GL_VERTEX_PROGRAM_BINDING_NV
    enum: GL_PROGRAM_ERROR_POSITION_NV
    enum: GL_VERTEX_ATTRIB_ARRAY0_NV
    enum: GL_VERTEX_ATTRIB_ARRAY1_NV
    enum: GL_VERTEX_ATTRIB_ARRAY2_NV
    enum: GL_VERTEX_ATTRIB_ARRAY3_NV
    enum: GL_VERTEX_ATTRIB_ARRAY4_NV
    enum: GL_VERTEX_ATTRIB_ARRAY5_NV
    enum: GL_VERTEX_ATTRIB_ARRAY6_NV
    enum: GL_VERTEX_ATTRIB_ARRAY7_NV
    enum: GL_VERTEX_ATTRIB_ARRAY8_NV
    enum: GL_VERTEX_ATTRIB_ARRAY9_NV
    enum: GL_VERTEX_ATTRIB_ARRAY10_NV
    enum: GL_VERTEX_ATTRIB_ARRAY11_NV
    enum: GL_VERTEX_ATTRIB_ARRAY12_NV
    enum: GL_VERTEX_ATTRIB_ARRAY13_NV
    enum: GL_VERTEX_ATTRIB_ARRAY14_NV
    enum: GL_VERTEX_ATTRIB_ARRAY15_NV
    enum: GL_MAP1_VERTEX_ATTRIB0_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB1_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB2_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB3_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB4_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB5_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB6_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB7_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB8_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB9_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB10_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB11_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB12_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB13_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB14_4_NV
    enum: GL_MAP1_VERTEX_ATTRIB15_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB0_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB1_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB2_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB3_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB4_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB5_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB6_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB7_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB8_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB9_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB10_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB11_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB12_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB13_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB14_4_NV
    enum: GL_MAP2_VERTEX_ATTRIB15_4_NV
    enum: GL_TEXTURE_MAX_CLAMP_S_SGIX
    enum: GL_TEXTURE_MAX_CLAMP_T_SGIX
    enum: GL_TEXTURE_MAX_CLAMP_R_SGIX
    enum: GL_SCALEBIAS_HINT_SGIX
    enum: GL_INTERLACE_OML
    enum: GL_INTERLACE_READ_OML
    enum: GL_FORMAT_SUBSAMPLE_24_24_OML
    enum: GL_FORMAT_SUBSAMPLE_244_244_OML
    enum: GL_PACK_RESAMPLE_OML
    enum: GL_UNPACK_RESAMPLE_OML
    enum: GL_RESAMPLE_REPLICATE_OML
    enum: GL_RESAMPLE_ZERO_FILL_OML
    enum: GL_RESAMPLE_AVERAGE_OML
    enum: GL_RESAMPLE_DECIMATE_OML
    enum: GL_DEPTH_STENCIL_TO_RGBA_NV
    enum: GL_DEPTH_STENCIL_TO_BGRA_NV
    enum: GL_BUMP_ROT_MATRIX_ATI
    enum: GL_BUMP_ROT_MATRIX_SIZE_ATI
    enum: GL_BUMP_NUM_TEX_UNITS_ATI
    enum: GL_BUMP_TEX_UNITS_ATI
    enum: GL_DUDV_ATI
    enum: GL_DU8DV8_ATI
    enum: GL_BUMP_ENVMAP_ATI
    enum: GL_BUMP_TARGET_ATI
    enum: GL_FRAGMENT_SHADER_ATI
    enum: GL_REG_0_ATI
    enum: GL_REG_1_ATI
    enum: GL_REG_2_ATI
    enum: GL_REG_3_ATI
    enum: GL_REG_4_ATI
    enum: GL_REG_5_ATI
    enum: GL_REG_6_ATI
    enum: GL_REG_7_ATI
    enum: GL_REG_8_ATI
    enum: GL_REG_9_ATI
    enum: GL_REG_10_ATI
    enum: GL_REG_11_ATI
    enum: GL_REG_12_ATI
    enum: GL_REG_13_ATI
    enum: GL_REG_14_ATI
    enum: GL_REG_15_ATI
    enum: GL_REG_16_ATI
    enum: GL_REG_17_ATI
    enum: GL_REG_18_ATI
    enum: GL_REG_19_ATI
    enum: GL_REG_20_ATI
    enum: GL_REG_21_ATI
    enum: GL_REG_22_ATI
    enum: GL_REG_23_ATI
    enum: GL_REG_24_ATI
    enum: GL_REG_25_ATI
    enum: GL_REG_26_ATI
    enum: GL_REG_27_ATI
    enum: GL_REG_28_ATI
    enum: GL_REG_29_ATI
    enum: GL_REG_30_ATI
    enum: GL_REG_31_ATI
    enum: GL_CON_0_ATI
    enum: GL_CON_1_ATI
    enum: GL_CON_2_ATI
    enum: GL_CON_3_ATI
    enum: GL_CON_4_ATI
    enum: GL_CON_5_ATI
    enum: GL_CON_6_ATI
    enum: GL_CON_7_ATI
    enum: GL_CON_8_ATI
    enum: GL_CON_9_ATI
    enum: GL_CON_10_ATI
    enum: GL_CON_11_ATI
    enum: GL_CON_12_ATI
    enum: GL_CON_13_ATI
    enum: GL_CON_14_ATI
    enum: GL_CON_15_ATI
    enum: GL_CON_16_ATI
    enum: GL_CON_17_ATI
    enum: GL_CON_18_ATI
    enum: GL_CON_19_ATI
    enum: GL_CON_20_ATI
    enum: GL_CON_21_ATI
    enum: GL_CON_22_ATI
    enum: GL_CON_23_ATI
    enum: GL_CON_24_ATI
    enum: GL_CON_25_ATI
    enum: GL_CON_26_ATI
    enum: GL_CON_27_ATI
    enum: GL_CON_28_ATI
    enum: GL_CON_29_ATI
    enum: GL_CON_30_ATI
    enum: GL_CON_31_ATI
    enum: GL_MOV_ATI
    enum: GL_ADD_ATI
    enum: GL_MUL_ATI
    enum: GL_SUB_ATI
    enum: GL_DOT3_ATI
    enum: GL_DOT4_ATI
    enum: GL_MAD_ATI
    enum: GL_LERP_ATI
    enum: GL_CND_ATI
    enum: GL_CND0_ATI
    enum: GL_DOT2_ADD_ATI
    enum: GL_SECONDARY_INTERPOLATOR_ATI
    enum: GL_NUM_FRAGMENT_REGISTERS_ATI
    enum: GL_NUM_FRAGMENT_CONSTANTS_ATI
    enum: GL_NUM_PASSES_ATI
    enum: GL_NUM_INSTRUCTIONS_PER_PASS_ATI
    enum: GL_NUM_INSTRUCTIONS_TOTAL_ATI
    enum: GL_NUM_INPUT_INTERPOLATOR_COMPONENTS_ATI
    enum: GL_NUM_LOOPBACK_COMPONENTS_ATI
    enum: GL_COLOR_ALPHA_PAIRING_ATI
    enum: GL_SWIZZLE_STR_ATI
    enum: GL_SWIZZLE_STQ_ATI
    enum: GL_SWIZZLE_STR_DR_ATI
    enum: GL_SWIZZLE_STQ_DQ_ATI
    enum: GL_SWIZZLE_STRQ_ATI
    enum: GL_SWIZZLE_STRQ_DQ_ATI
    enum: GL_RED_BIT_ATI
    enum: GL_GREEN_BIT_ATI
    enum: GL_BLUE_BIT_ATI
    enum: GL_2X_BIT_ATI
    enum: GL_4X_BIT_ATI
    enum: GL_8X_BIT_ATI
    enum: GL_HALF_BIT_ATI
    enum: GL_QUARTER_BIT_ATI
    enum: GL_EIGHTH_BIT_ATI
    enum: GL_SATURATE_BIT_ATI
    enum: GL_COMP_BIT_ATI
    enum: GL_NEGATE_BIT_ATI
    enum: GL_BIAS_BIT_ATI
    enum: GL_PN_TRIANGLES_ATI
    enum: GL_MAX_PN_TRIANGLES_TESSELATION_LEVEL_ATI
    enum: GL_PN_TRIANGLES_POINT_MODE_ATI
    enum: GL_PN_TRIANGLES_NORMAL_MODE_ATI
    enum: GL_PN_TRIANGLES_TESSELATION_LEVEL_ATI
    enum: GL_PN_TRIANGLES_POINT_MODE_LINEAR_ATI
    enum: GL_PN_TRIANGLES_POINT_MODE_CUBIC_ATI
    enum: GL_PN_TRIANGLES_NORMAL_MODE_LINEAR_ATI
    enum: GL_PN_TRIANGLES_NORMAL_MODE_QUADRATIC_ATI
    enum: GL_STATIC_ATI
    enum: GL_DYNAMIC_ATI
    enum: GL_PRESERVE_ATI
    enum: GL_DISCARD_ATI
    enum: GL_OBJECT_BUFFER_SIZE_ATI
    enum: GL_OBJECT_BUFFER_USAGE_ATI
    enum: GL_ARRAY_OBJECT_BUFFER_ATI
    enum: GL_ARRAY_OBJECT_OFFSET_ATI
    enum: GL_VERTEX_SHADER_EXT
    enum: GL_VERTEX_SHADER_BINDING_EXT
    enum: GL_OP_INDEX_EXT
    enum: GL_OP_NEGATE_EXT
    enum: GL_OP_DOT3_EXT
    enum: GL_OP_DOT4_EXT
    enum: GL_OP_MUL_EXT
    enum: GL_OP_ADD_EXT
    enum: GL_OP_MADD_EXT
    enum: GL_OP_FRAC_EXT
    enum: GL_OP_MAX_EXT
    enum: GL_OP_MIN_EXT
    enum: GL_OP_SET_GE_EXT
    enum: GL_OP_SET_LT_EXT
    enum: GL_OP_CLAMP_EXT
    enum: GL_OP_FLOOR_EXT
    enum: GL_OP_ROUND_EXT
    enum: GL_OP_EXP_BASE_2_EXT
    enum: GL_OP_LOG_BASE_2_EXT
    enum: GL_OP_POWER_EXT
    enum: GL_OP_RECIP_EXT
    enum: GL_OP_RECIP_SQRT_EXT
    enum: GL_OP_SUB_EXT
    enum: GL_OP_CROSS_PRODUCT_EXT
    enum: GL_OP_MULTIPLY_MATRIX_EXT
    enum: GL_OP_MOV_EXT
    enum: GL_OUTPUT_VERTEX_EXT
    enum: GL_OUTPUT_COLOR0_EXT
    enum: GL_OUTPUT_COLOR1_EXT
    enum: GL_OUTPUT_TEXTURE_COORD0_EXT
    enum: GL_OUTPUT_TEXTURE_COORD1_EXT
    enum: GL_OUTPUT_TEXTURE_COORD2_EXT
    enum: GL_OUTPUT_TEXTURE_COORD3_EXT
    enum: GL_OUTPUT_TEXTURE_COORD4_EXT
    enum: GL_OUTPUT_TEXTURE_COORD5_EXT
    enum: GL_OUTPUT_TEXTURE_COORD6_EXT
    enum: GL_OUTPUT_TEXTURE_COORD7_EXT
    enum: GL_OUTPUT_TEXTURE_COORD8_EXT
    enum: GL_OUTPUT_TEXTURE_COORD9_EXT
    enum: GL_OUTPUT_TEXTURE_COORD10_EXT
    enum: GL_OUTPUT_TEXTURE_COORD11_EXT
    enum: GL_OUTPUT_TEXTURE_COORD12_EXT
    enum: GL_OUTPUT_TEXTURE_COORD13_EXT
    enum: GL_OUTPUT_TEXTURE_COORD14_EXT
    enum: GL_OUTPUT_TEXTURE_COORD15_EXT
    enum: GL_OUTPUT_TEXTURE_COORD16_EXT
    enum: GL_OUTPUT_TEXTURE_COORD17_EXT
    enum: GL_OUTPUT_TEXTURE_COORD18_EXT
    enum: GL_OUTPUT_TEXTURE_COORD19_EXT
    enum: GL_OUTPUT_TEXTURE_COORD20_EXT
    enum: GL_OUTPUT_TEXTURE_COORD21_EXT
    enum: GL_OUTPUT_TEXTURE_COORD22_EXT
    enum: GL_OUTPUT_TEXTURE_COORD23_EXT
    enum: GL_OUTPUT_TEXTURE_COORD24_EXT
    enum: GL_OUTPUT_TEXTURE_COORD25_EXT
    enum: GL_OUTPUT_TEXTURE_COORD26_EXT
    enum: GL_OUTPUT_TEXTURE_COORD27_EXT
    enum: GL_OUTPUT_TEXTURE_COORD28_EXT
    enum: GL_OUTPUT_TEXTURE_COORD29_EXT
    enum: GL_OUTPUT_TEXTURE_COORD30_EXT
    enum: GL_OUTPUT_TEXTURE_COORD31_EXT
    enum: GL_OUTPUT_FOG_EXT
    enum: GL_SCALAR_EXT
    enum: GL_VECTOR_EXT
    enum: GL_MATRIX_EXT
    enum: GL_VARIANT_EXT
    enum: GL_INVARIANT_EXT
    enum: GL_LOCAL_CONSTANT_EXT
    enum: GL_LOCAL_EXT
    enum: GL_MAX_VERTEX_SHADER_INSTRUCTIONS_EXT
    enum: GL_MAX_VERTEX_SHADER_VARIANTS_EXT
    enum: GL_MAX_VERTEX_SHADER_INVARIANTS_EXT
    enum: GL_MAX_VERTEX_SHADER_LOCAL_CONSTANTS_EXT
    enum: GL_MAX_VERTEX_SHADER_LOCALS_EXT
    enum: GL_MAX_OPTIMIZED_VERTEX_SHADER_INSTRUCTIONS_EXT
    enum: GL_MAX_OPTIMIZED_VERTEX_SHADER_VARIANTS_EXT
    enum: GL_MAX_OPTIMIZED_VERTEX_SHADER_LOCAL_CONSTANTS_EXT
    enum: GL_MAX_OPTIMIZED_VERTEX_SHADER_INVARIANTS_EXT
    enum: GL_MAX_OPTIMIZED_VERTEX_SHADER_LOCALS_EXT
    enum: GL_VERTEX_SHADER_INSTRUCTIONS_EXT
    enum: GL_VERTEX_SHADER_VARIANTS_EXT
    enum: GL_VERTEX_SHADER_INVARIANTS_EXT
    enum: GL_VERTEX_SHADER_LOCAL_CONSTANTS_EXT
    enum: GL_VERTEX_SHADER_LOCALS_EXT
    enum: GL_VERTEX_SHADER_OPTIMIZED_EXT
    enum: GL_X_EXT
    enum: GL_Y_EXT
    enum: GL_Z_EXT
    enum: GL_W_EXT
    enum: GL_NEGATIVE_X_EXT
    enum: GL_NEGATIVE_Y_EXT
    enum: GL_NEGATIVE_Z_EXT
    enum: GL_NEGATIVE_W_EXT
    enum: GL_ZERO_EXT
    enum: GL_ONE_EXT
    enum: GL_NEGATIVE_ONE_EXT
    enum: GL_NORMALIZED_RANGE_EXT
    enum: GL_FULL_RANGE_EXT
    enum: GL_CURRENT_VERTEX_EXT
    enum: GL_MVP_MATRIX_EXT
    enum: GL_VARIANT_VALUE_EXT
    enum: GL_VARIANT_DATATYPE_EXT
    enum: GL_VARIANT_ARRAY_STRIDE_EXT
    enum: GL_VARIANT_ARRAY_TYPE_EXT
    enum: GL_VARIANT_ARRAY_EXT
    enum: GL_VARIANT_ARRAY_POINTER_EXT
    enum: GL_INVARIANT_VALUE_EXT
    enum: GL_INVARIANT_DATATYPE_EXT
    enum: GL_LOCAL_CONSTANT_VALUE_EXT
    enum: GL_LOCAL_CONSTANT_DATATYPE_EXT
    enum: GL_MAX_VERTEX_STREAMS_ATI
    enum: GL_VERTEX_STREAM0_ATI
    enum: GL_VERTEX_STREAM1_ATI
    enum: GL_VERTEX_STREAM2_ATI
    enum: GL_VERTEX_STREAM3_ATI
    enum: GL_VERTEX_STREAM4_ATI
    enum: GL_VERTEX_STREAM5_ATI
    enum: GL_VERTEX_STREAM6_ATI
    enum: GL_VERTEX_STREAM7_ATI
    enum: GL_VERTEX_SOURCE_ATI
    enum: GL_ELEMENT_ARRAY_ATI
    enum: GL_ELEMENT_ARRAY_TYPE_ATI
    enum: GL_ELEMENT_ARRAY_POINTER_ATI
    enum: GL_QUAD_MESH_SUN
    enum: GL_TRIANGLE_MESH_SUN
    enum: GL_SLICE_ACCUM_SUN
    enum: GL_MULTISAMPLE_FILTER_HINT_NV
    enum: GL_DEPTH_CLAMP_NV
    enum: GL_PIXEL_COUNTER_BITS_NV
    enum: GL_CURRENT_OCCLUSION_QUERY_ID_NV
    enum: GL_PIXEL_COUNT_NV
    enum: GL_PIXEL_COUNT_AVAILABLE_NV
    enum: GL_POINT_SPRITE_NV
    enum: GL_COORD_REPLACE_NV
    enum: GL_POINT_SPRITE_R_MODE_NV
    enum: GL_OFFSET_PROJECTIVE_TEXTURE_2D_NV
    enum: GL_OFFSET_PROJECTIVE_TEXTURE_2D_SCALE_NV
    enum: GL_OFFSET_PROJECTIVE_TEXTURE_RECTANGLE_NV
    enum: GL_OFFSET_PROJECTIVE_TEXTURE_RECTANGLE_SCALE_NV
    enum: GL_OFFSET_HILO_TEXTURE_2D_NV
    enum: GL_OFFSET_HILO_TEXTURE_RECTANGLE_NV
    enum: GL_OFFSET_HILO_PROJECTIVE_TEXTURE_2D_NV
    enum: GL_OFFSET_HILO_PROJECTIVE_TEXTURE_RECTANGLE_NV
    enum: GL_DEPENDENT_HILO_TEXTURE_2D_NV
    enum: GL_DEPENDENT_RGB_TEXTURE_3D_NV
    enum: GL_DEPENDENT_RGB_TEXTURE_CUBE_MAP_NV
    enum: GL_DOT_PRODUCT_PASS_THROUGH_NV
    enum: GL_DOT_PRODUCT_TEXTURE_1D_NV
    enum: GL_DOT_PRODUCT_AFFINE_DEPTH_REPLACE_NV
    enum: GL_HILO8_NV
    enum: GL_SIGNED_HILO8_NV
    enum: GL_FORCE_BLUE_TO_ONE_NV
    enum: GL_STENCIL_TEST_TWO_SIDE_EXT
    enum: GL_ACTIVE_STENCIL_FACE_EXT
    enum: GL_TEXT_FRAGMENT_SHADER_ATI
    enum: GL_UNPACK_CLIENT_STORAGE_APPLE
    enum: GL_ELEMENT_ARRAY_APPLE
    enum: GL_ELEMENT_ARRAY_TYPE_APPLE
    enum: GL_ELEMENT_ARRAY_POINTER_APPLE
    enum: GL_DRAW_PIXELS_APPLE
    enum: GL_FENCE_APPLE
    enum: GL_VERTEX_ARRAY_BINDING_APPLE
    enum: GL_VERTEX_ARRAY_RANGE_APPLE
    enum: GL_VERTEX_ARRAY_RANGE_LENGTH_APPLE
    enum: GL_VERTEX_ARRAY_STORAGE_HINT_APPLE
    enum: GL_VERTEX_ARRAY_RANGE_POINTER_APPLE
    enum: GL_STORAGE_CLIENT_APPLE
    enum: GL_STORAGE_CACHED_APPLE
    enum: GL_STORAGE_SHARED_APPLE
    enum: GL_YCBCR_422_APPLE
    enum: GL_UNSIGNED_SHORT_8_8_APPLE
    enum: GL_UNSIGNED_SHORT_8_8_REV_APPLE
    enum: GL_RGB_S3TC
    enum: GL_RGB4_S3TC
    enum: GL_RGBA_S3TC
    enum: GL_RGBA4_S3TC
    enum: GL_RGBA_DXT5_S3TC
    enum: GL_RGBA4_DXT5_S3TC
    enum: GL_MAX_DRAW_BUFFERS_ATI
    enum: GL_DRAW_BUFFER0_ATI
    enum: GL_DRAW_BUFFER1_ATI
    enum: GL_DRAW_BUFFER2_ATI
    enum: GL_DRAW_BUFFER3_ATI
    enum: GL_DRAW_BUFFER4_ATI
    enum: GL_DRAW_BUFFER5_ATI
    enum: GL_DRAW_BUFFER6_ATI
    enum: GL_DRAW_BUFFER7_ATI
    enum: GL_DRAW_BUFFER8_ATI
    enum: GL_DRAW_BUFFER9_ATI
    enum: GL_DRAW_BUFFER10_ATI
    enum: GL_DRAW_BUFFER11_ATI
    enum: GL_DRAW_BUFFER12_ATI
    enum: GL_DRAW_BUFFER13_ATI
    enum: GL_DRAW_BUFFER14_ATI
    enum: GL_DRAW_BUFFER15_ATI
    enum: GL_RGBA_FLOAT_MODE_ATI
    enum: GL_COLOR_CLEAR_UNCLAMPED_VALUE_ATI
    enum: GL_MODULATE_ADD_ATI
    enum: GL_MODULATE_SIGNED_ADD_ATI
    enum: GL_MODULATE_SUBTRACT_ATI
    enum: GL_RGBA_FLOAT32_ATI
    enum: GL_RGB_FLOAT32_ATI
    enum: GL_ALPHA_FLOAT32_ATI
    enum: GL_INTENSITY_FLOAT32_ATI
    enum: GL_LUMINANCE_FLOAT32_ATI
    enum: GL_LUMINANCE_ALPHA_FLOAT32_ATI
    enum: GL_RGBA_FLOAT16_ATI
    enum: GL_RGB_FLOAT16_ATI
    enum: GL_ALPHA_FLOAT16_ATI
    enum: GL_INTENSITY_FLOAT16_ATI
    enum: GL_LUMINANCE_FLOAT16_ATI
    enum: GL_LUMINANCE_ALPHA_FLOAT16_ATI
    enum: GL_FLOAT_R_NV
    enum: GL_FLOAT_RG_NV
    enum: GL_FLOAT_RGB_NV
    enum: GL_FLOAT_RGBA_NV
    enum: GL_FLOAT_R16_NV
    enum: GL_FLOAT_R32_NV
    enum: GL_FLOAT_RG16_NV
    enum: GL_FLOAT_RG32_NV
    enum: GL_FLOAT_RGB16_NV
    enum: GL_FLOAT_RGB32_NV
    enum: GL_FLOAT_RGBA16_NV
    enum: GL_FLOAT_RGBA32_NV
    enum: GL_TEXTURE_FLOAT_COMPONENTS_NV
    enum: GL_FLOAT_CLEAR_COLOR_VALUE_NV
    enum: GL_FLOAT_RGBA_MODE_NV
    enum: GL_MAX_FRAGMENT_PROGRAM_LOCAL_PARAMETERS_NV
    enum: GL_FRAGMENT_PROGRAM_NV
    enum: GL_MAX_TEXTURE_COORDS_NV
    enum: GL_MAX_TEXTURE_IMAGE_UNITS_NV
    enum: GL_FRAGMENT_PROGRAM_BINDING_NV
    enum: GL_PROGRAM_ERROR_STRING_NV
    enum: GL_HALF_FLOAT_NV
    enum: GL_WRITE_PIXEL_DATA_RANGE_NV
    enum: GL_READ_PIXEL_DATA_RANGE_NV
    enum: GL_WRITE_PIXEL_DATA_RANGE_LENGTH_NV
    enum: GL_READ_PIXEL_DATA_RANGE_LENGTH_NV
    enum: GL_WRITE_PIXEL_DATA_RANGE_POINTER_NV
    enum: GL_READ_PIXEL_DATA_RANGE_POINTER_NV
    enum: GL_PRIMITIVE_RESTART_NV
    enum: GL_PRIMITIVE_RESTART_INDEX_NV
    enum: GL_TEXTURE_UNSIGNED_REMAP_MODE_NV
    enum: GL_STENCIL_BACK_FUNC_ATI
    enum: GL_STENCIL_BACK_FAIL_ATI
    enum: GL_STENCIL_BACK_PASS_DEPTH_FAIL_ATI
    enum: GL_STENCIL_BACK_PASS_DEPTH_PASS_ATI
    enum: GL_IMPLEMENTATION_COLOR_READ_TYPE_OES
    enum: GL_IMPLEMENTATION_COLOR_READ_FORMAT_OES
    enum: GL_DEPTH_BOUNDS_TEST_EXT
    enum: GL_DEPTH_BOUNDS_EXT
    enum: GL_MIRROR_CLAMP_EXT
    enum: GL_MIRROR_CLAMP_TO_EDGE_EXT
    enum: GL_MIRROR_CLAMP_TO_BORDER_EXT
    enum: GL_BLEND_EQUATION_RGB_EXT
    enum: GL_BLEND_EQUATION_ALPHA_EXT
    enum: GL_PACK_INVERT_MESA
    enum: GL_UNSIGNED_SHORT_8_8_MESA
    enum: GL_UNSIGNED_SHORT_8_8_REV_MESA
    enum: GL_YCBCR_MESA
    enum: GL_PIXEL_PACK_BUFFER_EXT
    enum: GL_PIXEL_UNPACK_BUFFER_EXT
    enum: GL_PIXEL_PACK_BUFFER_BINDING_EXT
    enum: GL_PIXEL_UNPACK_BUFFER_BINDING_EXT
    enum: GL_MAX_PROGRAM_EXEC_INSTRUCTIONS_NV
    enum: GL_MAX_PROGRAM_CALL_DEPTH_NV
    enum: GL_MAX_PROGRAM_IF_DEPTH_NV
    enum: GL_MAX_PROGRAM_LOOP_DEPTH_NV
    enum: GL_MAX_PROGRAM_LOOP_COUNT_NV
    enum: GL_INVALID_FRAMEBUFFER_OPERATION_EXT
    enum: GL_MAX_RENDERBUFFER_SIZE_EXT
    enum: GL_FRAMEBUFFER_BINDING_EXT
    enum: GL_RENDERBUFFER_BINDING_EXT
    enum: GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE_EXT
    enum: GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME_EXT
    enum: GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL_EXT
    enum: GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE_EXT
    enum: GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_3D_ZOFFSET_EXT
    enum: GL_FRAMEBUFFER_COMPLETE_EXT
    enum: GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT
    enum: GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT
    enum: GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT
    enum: GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT
    enum: GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT
    enum: GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT
    enum: GL_FRAMEBUFFER_UNSUPPORTED_EXT
    enum: GL_MAX_COLOR_ATTACHMENTS_EXT
    enum: GL_COLOR_ATTACHMENT0_EXT
    enum: GL_COLOR_ATTACHMENT1_EXT
    enum: GL_COLOR_ATTACHMENT2_EXT
    enum: GL_COLOR_ATTACHMENT3_EXT
    enum: GL_COLOR_ATTACHMENT4_EXT
    enum: GL_COLOR_ATTACHMENT5_EXT
    enum: GL_COLOR_ATTACHMENT6_EXT
    enum: GL_COLOR_ATTACHMENT7_EXT
    enum: GL_COLOR_ATTACHMENT8_EXT
    enum: GL_COLOR_ATTACHMENT9_EXT
    enum: GL_COLOR_ATTACHMENT10_EXT
    enum: GL_COLOR_ATTACHMENT11_EXT
    enum: GL_COLOR_ATTACHMENT12_EXT
    enum: GL_COLOR_ATTACHMENT13_EXT
    enum: GL_COLOR_ATTACHMENT14_EXT
    enum: GL_COLOR_ATTACHMENT15_EXT
    enum: GL_DEPTH_ATTACHMENT_EXT
    enum: GL_STENCIL_ATTACHMENT_EXT
    enum: GL_FRAMEBUFFER_EXT
    enum: GL_RENDERBUFFER_EXT
    enum: GL_RENDERBUFFER_WIDTH_EXT
    enum: GL_RENDERBUFFER_HEIGHT_EXT
    enum: GL_RENDERBUFFER_INTERNAL_FORMAT_EXT
    enum: GL_STENCIL_INDEX1_EXT
    enum: GL_STENCIL_INDEX4_EXT
    enum: GL_STENCIL_INDEX8_EXT
    enum: GL_STENCIL_INDEX16_EXT
    enum: GL_RENDERBUFFER_RED_SIZE_EXT
    enum: GL_RENDERBUFFER_GREEN_SIZE_EXT
    enum: GL_RENDERBUFFER_BLUE_SIZE_EXT
    enum: GL_RENDERBUFFER_ALPHA_SIZE_EXT
    enum: GL_RENDERBUFFER_DEPTH_SIZE_EXT
    enum: GL_RENDERBUFFER_STENCIL_SIZE_EXT
    enum: GL_DEPTH_STENCIL_EXT
    enum: GL_UNSIGNED_INT_24_8_EXT
    enum: GL_DEPTH24_STENCIL8_EXT
    enum: GL_TEXTURE_STENCIL_SIZE_EXT
    enum: GL_STENCIL_TAG_BITS_EXT
    enum: GL_STENCIL_CLEAR_TAG_VALUE_EXT
    enum: GL_SRGB_EXT
    enum: GL_SRGB8_EXT
    enum: GL_SRGB_ALPHA_EXT
    enum: GL_SRGB8_ALPHA8_EXT
    enum: GL_SLUMINANCE_ALPHA_EXT
    enum: GL_SLUMINANCE8_ALPHA8_EXT
    enum: GL_SLUMINANCE_EXT
    enum: GL_SLUMINANCE8_EXT
    enum: GL_COMPRESSED_SRGB_EXT
    enum: GL_COMPRESSED_SRGB_ALPHA_EXT
    enum: GL_COMPRESSED_SLUMINANCE_EXT
    enum: GL_COMPRESSED_SLUMINANCE_ALPHA_EXT
    enum: GL_COMPRESSED_SRGB_S3TC_DXT1_EXT
    enum: GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT1_EXT
    enum: GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT3_EXT
    enum: GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT5_EXT
    enum: GL_READ_FRAMEBUFFER_EXT
    enum: GL_DRAW_FRAMEBUFFER_EXT
    enum: GL_DRAW_FRAMEBUFFER_BINDING_EXT
    enum: GL_READ_FRAMEBUFFER_BINDING_EXT
    enum: GL_RENDERBUFFER_SAMPLES_EXT
    enum: GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_EXT
    enum: GL_MAX_SAMPLES_EXT
    enum: GL_TEXTURE_1D_STACK_MESAX
    enum: GL_TEXTURE_2D_STACK_MESAX
    enum: GL_PROXY_TEXTURE_1D_STACK_MESAX
    enum: GL_PROXY_TEXTURE_2D_STACK_MESAX
    enum: GL_TEXTURE_1D_STACK_BINDING_MESAX
    enum: GL_TEXTURE_2D_STACK_BINDING_MESAX
    enum: GL_TIME_ELAPSED_EXT
    enum: GL_BUFFER_SERIALIZED_MODIFY_APPLE
    enum: GL_BUFFER_FLUSHING_UNMAP_APPLE
    enum: GL_MIN_PROGRAM_TEXEL_OFFSET_NV
    enum: GL_MAX_PROGRAM_TEXEL_OFFSET_NV
    enum: GL_PROGRAM_ATTRIB_COMPONENTS_NV
    enum: GL_PROGRAM_RESULT_COMPONENTS_NV
    enum: GL_MAX_PROGRAM_ATTRIB_COMPONENTS_NV
    enum: GL_MAX_PROGRAM_RESULT_COMPONENTS_NV
    enum: GL_MAX_PROGRAM_GENERIC_ATTRIBS_NV
    enum: GL_MAX_PROGRAM_GENERIC_RESULTS_NV
    enum: GL_LINES_ADJACENCY_EXT
    enum: GL_LINE_STRIP_ADJACENCY_EXT
    enum: GL_TRIANGLES_ADJACENCY_EXT
    enum: GL_TRIANGLE_STRIP_ADJACENCY_EXT
    enum: GL_GEOMETRY_PROGRAM_NV
    enum: GL_MAX_PROGRAM_OUTPUT_VERTICES_NV
    enum: GL_MAX_PROGRAM_TOTAL_OUTPUT_COMPONENTS_NV
    enum: GL_GEOMETRY_VERTICES_OUT_EXT
    enum: GL_GEOMETRY_INPUT_TYPE_EXT
    enum: GL_GEOMETRY_OUTPUT_TYPE_EXT
    enum: GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS_EXT
    enum: GL_FRAMEBUFFER_ATTACHMENT_LAYERED_EXT
    enum: GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS_EXT
    enum: GL_FRAMEBUFFER_INCOMPLETE_LAYER_COUNT_EXT
    enum: GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER_EXT
    enum: GL_PROGRAM_POINT_SIZE_EXT
    enum: GL_GEOMETRY_SHADER_EXT
    enum: GL_MAX_GEOMETRY_VARYING_COMPONENTS_EXT
    enum: GL_MAX_VERTEX_VARYING_COMPONENTS_EXT
    enum: GL_MAX_VARYING_COMPONENTS_EXT
    enum: GL_MAX_GEOMETRY_UNIFORM_COMPONENTS_EXT
    enum: GL_MAX_GEOMETRY_OUTPUT_VERTICES_EXT
    enum: GL_MAX_GEOMETRY_TOTAL_OUTPUT_COMPONENTS_EXT
    enum: GL_VERTEX_ATTRIB_ARRAY_INTEGER_NV
    enum: GL_SAMPLER_1D_ARRAY_EXT
    enum: GL_SAMPLER_2D_ARRAY_EXT
    enum: GL_SAMPLER_BUFFER_EXT
    enum: GL_SAMPLER_1D_ARRAY_SHADOW_EXT
    enum: GL_SAMPLER_2D_ARRAY_SHADOW_EXT
    enum: GL_SAMPLER_CUBE_SHADOW_EXT
    enum: GL_UNSIGNED_INT_VEC2_EXT
    enum: GL_UNSIGNED_INT_VEC3_EXT
    enum: GL_UNSIGNED_INT_VEC4_EXT
    enum: GL_INT_SAMPLER_1D_EXT
    enum: GL_INT_SAMPLER_2D_EXT
    enum: GL_INT_SAMPLER_3D_EXT
    enum: GL_INT_SAMPLER_CUBE_EXT
    enum: GL_INT_SAMPLER_2D_RECT_EXT
    enum: GL_INT_SAMPLER_1D_ARRAY_EXT
    enum: GL_INT_SAMPLER_2D_ARRAY_EXT
    enum: GL_INT_SAMPLER_BUFFER_EXT
    enum: GL_UNSIGNED_INT_SAMPLER_1D_EXT
    enum: GL_UNSIGNED_INT_SAMPLER_2D_EXT
    enum: GL_UNSIGNED_INT_SAMPLER_3D_EXT
    enum: GL_UNSIGNED_INT_SAMPLER_CUBE_EXT
    enum: GL_UNSIGNED_INT_SAMPLER_2D_RECT_EXT
    enum: GL_UNSIGNED_INT_SAMPLER_1D_ARRAY_EXT
    enum: GL_UNSIGNED_INT_SAMPLER_2D_ARRAY_EXT
    enum: GL_UNSIGNED_INT_SAMPLER_BUFFER_EXT
    enum: GL_R11F_G11F_B10F_EXT
    enum: GL_UNSIGNED_INT_10F_11F_11F_REV_EXT
    enum: GL_RGBA_SIGNED_COMPONENTS_EXT
    enum: GL_TEXTURE_1D_ARRAY_EXT
    enum: GL_PROXY_TEXTURE_1D_ARRAY_EXT
    enum: GL_TEXTURE_2D_ARRAY_EXT
    enum: GL_PROXY_TEXTURE_2D_ARRAY_EXT
    enum: GL_TEXTURE_BINDING_1D_ARRAY_EXT
    enum: GL_TEXTURE_BINDING_2D_ARRAY_EXT
    enum: GL_MAX_ARRAY_TEXTURE_LAYERS_EXT
    enum: GL_COMPARE_REF_DEPTH_TO_TEXTURE_EXT
    enum: GL_TEXTURE_BUFFER_EXT
    enum: GL_MAX_TEXTURE_BUFFER_SIZE_EXT
    enum: GL_TEXTURE_BINDING_BUFFER_EXT
    enum: GL_TEXTURE_BUFFER_DATA_STORE_BINDING_EXT
    enum: GL_TEXTURE_BUFFER_FORMAT_EXT
    enum: GL_COMPRESSED_LUMINANCE_LATC1_EXT
    enum: GL_COMPRESSED_SIGNED_LUMINANCE_LATC1_EXT
    enum: GL_COMPRESSED_LUMINANCE_ALPHA_LATC2_EXT
    enum: GL_COMPRESSED_SIGNED_LUMINANCE_ALPHA_LATC2_EXT
    enum: GL_COMPRESSED_RED_RGTC1_EXT
    enum: GL_COMPRESSED_SIGNED_RED_RGTC1_EXT
    enum: GL_COMPRESSED_RED_GREEN_RGTC2_EXT
    enum: GL_COMPRESSED_SIGNED_RED_GREEN_RGTC2_EXT
    enum: GL_RGB9_E5_EXT
    enum: GL_UNSIGNED_INT_5_9_9_9_REV_EXT
    enum: GL_TEXTURE_SHARED_SIZE_EXT
    enum: GL_DEPTH_COMPONENT32F_NV
    enum: GL_DEPTH32F_STENCIL8_NV
    enum: GL_FLOAT_32_UNSIGNED_INT_24_8_REV_NV
    enum: GL_DEPTH_BUFFER_FLOAT_MODE_NV
    enum: GL_RENDERBUFFER_COVERAGE_SAMPLES_NV
    enum: GL_RENDERBUFFER_COLOR_SAMPLES_NV
    enum: GL_MAX_MULTISAMPLE_COVERAGE_MODES_NV
    enum: GL_MULTISAMPLE_COVERAGE_MODES_NV
    enum: GL_FRAMEBUFFER_SRGB_EXT
    enum: GL_FRAMEBUFFER_SRGB_CAPABLE_EXT
    enum: GL_MAX_PROGRAM_PARAMETER_BUFFER_BINDINGS_NV
    enum: GL_MAX_PROGRAM_PARAMETER_BUFFER_SIZE_NV
    enum: GL_VERTEX_PROGRAM_PARAMETER_BUFFER_NV
    enum: GL_GEOMETRY_PROGRAM_PARAMETER_BUFFER_NV
    enum: GL_FRAGMENT_PROGRAM_PARAMETER_BUFFER_NV
    enum: GL_BACK_PRIMARY_COLOR_NV
    enum: GL_BACK_SECONDARY_COLOR_NV
    enum: GL_TEXTURE_COORD_NV
    enum: GL_CLIP_DISTANCE_NV
    enum: GL_VERTEX_ID_NV
    enum: GL_PRIMITIVE_ID_NV
    enum: GL_GENERIC_ATTRIB_NV
    enum: GL_TRANSFORM_FEEDBACK_ATTRIBS_NV
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_MODE_NV
    enum: GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS_NV
    enum: GL_ACTIVE_VARYINGS_NV
    enum: GL_ACTIVE_VARYING_MAX_LENGTH_NV
    enum: GL_TRANSFORM_FEEDBACK_VARYINGS_NV
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_START_NV
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_SIZE_NV
    enum: GL_TRANSFORM_FEEDBACK_RECORD_NV
    enum: GL_PRIMITIVES_GENERATED_NV
    enum: GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_NV
    enum: GL_RASTERIZER_DISCARD_NV
    enum: GL_MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS_NV
    enum: GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS_NV
    enum: GL_INTERLEAVED_ATTRIBS_NV
    enum: GL_SEPARATE_ATTRIBS_NV
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_NV
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_BINDING_NV
    enum: GL_LAYER_NV
    enum: GL_NEXT_BUFFER_NV
    enum: GL_SKIP_COMPONENTS4_NV
    enum: GL_SKIP_COMPONENTS3_NV
    enum: GL_SKIP_COMPONENTS2_NV
    enum: GL_SKIP_COMPONENTS1_NV
    enum: GL_MAX_VERTEX_BINDABLE_UNIFORMS_EXT
    enum: GL_MAX_FRAGMENT_BINDABLE_UNIFORMS_EXT
    enum: GL_MAX_GEOMETRY_BINDABLE_UNIFORMS_EXT
    enum: GL_MAX_BINDABLE_UNIFORM_SIZE_EXT
    enum: GL_UNIFORM_BUFFER_EXT
    enum: GL_UNIFORM_BUFFER_BINDING_EXT
    enum: GL_RGBA32UI_EXT
    enum: GL_RGB32UI_EXT
    enum: GL_ALPHA32UI_EXT
    enum: GL_INTENSITY32UI_EXT
    enum: GL_LUMINANCE32UI_EXT
    enum: GL_LUMINANCE_ALPHA32UI_EXT
    enum: GL_RGBA16UI_EXT
    enum: GL_RGB16UI_EXT
    enum: GL_ALPHA16UI_EXT
    enum: GL_INTENSITY16UI_EXT
    enum: GL_LUMINANCE16UI_EXT
    enum: GL_LUMINANCE_ALPHA16UI_EXT
    enum: GL_RGBA8UI_EXT
    enum: GL_RGB8UI_EXT
    enum: GL_ALPHA8UI_EXT
    enum: GL_INTENSITY8UI_EXT
    enum: GL_LUMINANCE8UI_EXT
    enum: GL_LUMINANCE_ALPHA8UI_EXT
    enum: GL_RGBA32I_EXT
    enum: GL_RGB32I_EXT
    enum: GL_ALPHA32I_EXT
    enum: GL_INTENSITY32I_EXT
    enum: GL_LUMINANCE32I_EXT
    enum: GL_LUMINANCE_ALPHA32I_EXT
    enum: GL_RGBA16I_EXT
    enum: GL_RGB16I_EXT
    enum: GL_ALPHA16I_EXT
    enum: GL_INTENSITY16I_EXT
    enum: GL_LUMINANCE16I_EXT
    enum: GL_LUMINANCE_ALPHA16I_EXT
    enum: GL_RGBA8I_EXT
    enum: GL_RGB8I_EXT
    enum: GL_ALPHA8I_EXT
    enum: GL_INTENSITY8I_EXT
    enum: GL_LUMINANCE8I_EXT
    enum: GL_LUMINANCE_ALPHA8I_EXT
    enum: GL_RED_INTEGER_EXT
    enum: GL_GREEN_INTEGER_EXT
    enum: GL_BLUE_INTEGER_EXT
    enum: GL_ALPHA_INTEGER_EXT
    enum: GL_RGB_INTEGER_EXT
    enum: GL_RGBA_INTEGER_EXT
    enum: GL_BGR_INTEGER_EXT
    enum: GL_BGRA_INTEGER_EXT
    enum: GL_LUMINANCE_INTEGER_EXT
    enum: GL_LUMINANCE_ALPHA_INTEGER_EXT
    enum: GL_RGBA_INTEGER_MODE_EXT
    enum: GL_QUERY_WAIT_NV
    enum: GL_QUERY_NO_WAIT_NV
    enum: GL_QUERY_BY_REGION_WAIT_NV
    enum: GL_QUERY_BY_REGION_NO_WAIT_NV
    enum: GL_FRAME_NV
    enum: GL_FIELDS_NV
    enum: GL_CURRENT_TIME_NV
    enum: GL_NUM_FILL_STREAMS_NV
    enum: GL_PRESENT_TIME_NV
    enum: GL_PRESENT_DURATION_NV
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_EXT
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_START_EXT
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_SIZE_EXT
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_BINDING_EXT
    enum: GL_INTERLEAVED_ATTRIBS_EXT
    enum: GL_SEPARATE_ATTRIBS_EXT
    enum: GL_PRIMITIVES_GENERATED_EXT
    enum: GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_EXT
    enum: GL_RASTERIZER_DISCARD_EXT
    enum: GL_MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS_EXT
    enum: GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS_EXT
    enum: GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS_EXT
    enum: GL_TRANSFORM_FEEDBACK_VARYINGS_EXT
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_MODE_EXT
    enum: GL_TRANSFORM_FEEDBACK_VARYING_MAX_LENGTH_EXT
    enum: GL_PROGRAM_MATRIX_EXT
    enum: GL_TRANSPOSE_PROGRAM_MATRIX_EXT
    enum: GL_PROGRAM_MATRIX_STACK_DEPTH_EXT
    enum: GL_TEXTURE_SWIZZLE_R_EXT
    enum: GL_TEXTURE_SWIZZLE_G_EXT
    enum: GL_TEXTURE_SWIZZLE_B_EXT
    enum: GL_TEXTURE_SWIZZLE_A_EXT
    enum: GL_TEXTURE_SWIZZLE_RGBA_EXT
    enum: GL_SAMPLE_POSITION_NV
    enum: GL_SAMPLE_MASK_NV
    enum: GL_SAMPLE_MASK_VALUE_NV
    enum: GL_TEXTURE_BINDING_RENDERBUFFER_NV
    enum: GL_TEXTURE_RENDERBUFFER_DATA_STORE_BINDING_NV
    enum: GL_TEXTURE_RENDERBUFFER_NV
    enum: GL_SAMPLER_RENDERBUFFER_NV
    enum: GL_INT_SAMPLER_RENDERBUFFER_NV
    enum: GL_UNSIGNED_INT_SAMPLER_RENDERBUFFER_NV
    enum: GL_MAX_SAMPLE_MASK_WORDS_NV
    enum: GL_TRANSFORM_FEEDBACK_NV
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_PAUSED_NV
    enum: GL_TRANSFORM_FEEDBACK_BUFFER_ACTIVE_NV
    enum: GL_TRANSFORM_FEEDBACK_BINDING_NV
    enum: GL_VBO_FREE_MEMORY_ATI
    enum: GL_TEXTURE_FREE_MEMORY_ATI
    enum: GL_RENDERBUFFER_FREE_MEMORY_ATI
    enum: GL_COUNTER_TYPE_AMD
    enum: GL_COUNTER_RANGE_AMD
    enum: GL_UNSIGNED_INT64_AMD
    enum: GL_PERCENTAGE_AMD
    enum: GL_PERFMON_RESULT_AVAILABLE_AMD
    enum: GL_PERFMON_RESULT_SIZE_AMD
    enum: GL_PERFMON_RESULT_AMD
    enum: GL_SAMPLER_BUFFER_AMD
    enum: GL_INT_SAMPLER_BUFFER_AMD
    enum: GL_UNSIGNED_INT_SAMPLER_BUFFER_AMD
    enum: GL_TESSELLATION_MODE_AMD
    enum: GL_TESSELLATION_FACTOR_AMD
    enum: GL_DISCRETE_AMD
    enum: GL_CONTINUOUS_AMD
    enum: GL_QUADS_FOLLOW_PROVOKING_VERTEX_CONVENTION_EXT
    enum: GL_FIRST_VERTEX_CONVENTION_EXT
    enum: GL_LAST_VERTEX_CONVENTION_EXT
    enum: GL_PROVOKING_VERTEX_EXT
    enum: GL_ALPHA_SNORM
    enum: GL_LUMINANCE_SNORM
    enum: GL_LUMINANCE_ALPHA_SNORM
    enum: GL_INTENSITY_SNORM
    enum: GL_ALPHA8_SNORM
    enum: GL_LUMINANCE8_SNORM
    enum: GL_LUMINANCE8_ALPHA8_SNORM
    enum: GL_INTENSITY8_SNORM
    enum: GL_ALPHA16_SNORM
    enum: GL_LUMINANCE16_SNORM
    enum: GL_LUMINANCE16_ALPHA16_SNORM
    enum: GL_INTENSITY16_SNORM
    enum: GL_TEXTURE_RANGE_LENGTH_APPLE
    enum: GL_TEXTURE_RANGE_POINTER_APPLE
    enum: GL_TEXTURE_STORAGE_HINT_APPLE
    enum: GL_STORAGE_PRIVATE_APPLE
    enum: GL_HALF_APPLE
    enum: GL_RGBA_FLOAT32_APPLE
    enum: GL_RGB_FLOAT32_APPLE
    enum: GL_ALPHA_FLOAT32_APPLE
    enum: GL_INTENSITY_FLOAT32_APPLE
    enum: GL_LUMINANCE_FLOAT32_APPLE
    enum: GL_LUMINANCE_ALPHA_FLOAT32_APPLE
    enum: GL_RGBA_FLOAT16_APPLE
    enum: GL_RGB_FLOAT16_APPLE
    enum: GL_ALPHA_FLOAT16_APPLE
    enum: GL_INTENSITY_FLOAT16_APPLE
    enum: GL_LUMINANCE_FLOAT16_APPLE
    enum: GL_LUMINANCE_ALPHA_FLOAT16_APPLE
    enum: GL_COLOR_FLOAT_APPLE
    enum: GL_VERTEX_ATTRIB_MAP1_APPLE
    enum: GL_VERTEX_ATTRIB_MAP2_APPLE
    enum: GL_VERTEX_ATTRIB_MAP1_SIZE_APPLE
    enum: GL_VERTEX_ATTRIB_MAP1_COEFF_APPLE
    enum: GL_VERTEX_ATTRIB_MAP1_ORDER_APPLE
    enum: GL_VERTEX_ATTRIB_MAP1_DOMAIN_APPLE
    enum: GL_VERTEX_ATTRIB_MAP2_SIZE_APPLE
    enum: GL_VERTEX_ATTRIB_MAP2_COEFF_APPLE
    enum: GL_VERTEX_ATTRIB_MAP2_ORDER_APPLE
    enum: GL_VERTEX_ATTRIB_MAP2_DOMAIN_APPLE
    enum: GL_AUX_DEPTH_STENCIL_APPLE
    enum: GL_BUFFER_OBJECT_APPLE
    enum: GL_RELEASED_APPLE
    enum: GL_VOLATILE_APPLE
    enum: GL_RETAINED_APPLE
    enum: GL_UNDEFINED_APPLE
    enum: GL_PURGEABLE_APPLE
    enum: GL_PACK_ROW_BYTES_APPLE
    enum: GL_UNPACK_ROW_BYTES_APPLE
    enum: GL_RGB_422_APPLE
    enum: GL_VIDEO_BUFFER_NV
    enum: GL_VIDEO_BUFFER_BINDING_NV
    enum: GL_FIELD_UPPER_NV
    enum: GL_FIELD_LOWER_NV
    enum: GL_NUM_VIDEO_CAPTURE_STREAMS_NV
    enum: GL_NEXT_VIDEO_CAPTURE_BUFFER_STATUS_NV
    enum: GL_VIDEO_CAPTURE_TO_422_SUPPORTED_NV
    enum: GL_LAST_VIDEO_CAPTURE_STATUS_NV
    enum: GL_VIDEO_BUFFER_PITCH_NV
    enum: GL_VIDEO_COLOR_CONVERSION_MATRIX_NV
    enum: GL_VIDEO_COLOR_CONVERSION_MAX_NV
    enum: GL_VIDEO_COLOR_CONVERSION_MIN_NV
    enum: GL_VIDEO_COLOR_CONVERSION_OFFSET_NV
    enum: GL_VIDEO_BUFFER_INTERNAL_FORMAT_NV
    enum: GL_PARTIAL_SUCCESS_NV
    enum: GL_SUCCESS_NV
    enum: GL_FAILURE_NV
    enum: GL_YCBYCR8_422_NV
    enum: GL_YCBAYCR8A_4224_NV
    enum: GL_Z6Y10Z6CB10Z6Y10Z6CR10_422_NV
    enum: GL_Z6Y10Z6CB10Z6A10Z6Y10Z6CR10Z6A10_4224_NV
    enum: GL_Z4Y12Z4CB12Z4Y12Z4CR12_422_NV
    enum: GL_Z4Y12Z4CB12Z4A12Z4Y12Z4CR12Z4A12_4224_NV
    enum: GL_Z4Y12Z4CB12Z4CR12_444_NV
    enum: GL_VIDEO_CAPTURE_FRAME_WIDTH_NV
    enum: GL_VIDEO_CAPTURE_FRAME_HEIGHT_NV
    enum: GL_VIDEO_CAPTURE_FIELD_UPPER_HEIGHT_NV
    enum: GL_VIDEO_CAPTURE_FIELD_LOWER_HEIGHT_NV
    enum: GL_VIDEO_CAPTURE_SURFACE_ORIGIN_NV
    enum: GL_ACTIVE_PROGRAM_EXT
    enum: GL_BUFFER_GPU_ADDRESS_NV
    enum: GL_GPU_ADDRESS_NV
    enum: GL_MAX_SHADER_BUFFER_ADDRESS_NV
    enum: GL_VERTEX_ATTRIB_ARRAY_UNIFIED_NV
    enum: GL_ELEMENT_ARRAY_UNIFIED_NV
    enum: GL_VERTEX_ATTRIB_ARRAY_ADDRESS_NV
    enum: GL_VERTEX_ARRAY_ADDRESS_NV
    enum: GL_NORMAL_ARRAY_ADDRESS_NV
    enum: GL_COLOR_ARRAY_ADDRESS_NV
    enum: GL_INDEX_ARRAY_ADDRESS_NV
    enum: GL_TEXTURE_COORD_ARRAY_ADDRESS_NV
    enum: GL_EDGE_FLAG_ARRAY_ADDRESS_NV
    enum: GL_SECONDARY_COLOR_ARRAY_ADDRESS_NV
    enum: GL_FOG_COORD_ARRAY_ADDRESS_NV
    enum: GL_ELEMENT_ARRAY_ADDRESS_NV
    enum: GL_VERTEX_ATTRIB_ARRAY_LENGTH_NV
    enum: GL_VERTEX_ARRAY_LENGTH_NV
    enum: GL_NORMAL_ARRAY_LENGTH_NV
    enum: GL_COLOR_ARRAY_LENGTH_NV
    enum: GL_INDEX_ARRAY_LENGTH_NV
    enum: GL_TEXTURE_COORD_ARRAY_LENGTH_NV
    enum: GL_EDGE_FLAG_ARRAY_LENGTH_NV
    enum: GL_SECONDARY_COLOR_ARRAY_LENGTH_NV
    enum: GL_FOG_COORD_ARRAY_LENGTH_NV
    enum: GL_ELEMENT_ARRAY_LENGTH_NV
    enum: GL_DRAW_INDIRECT_UNIFIED_NV
    enum: GL_DRAW_INDIRECT_ADDRESS_NV
    enum: GL_DRAW_INDIRECT_LENGTH_NV
    enum: GL_MAX_IMAGE_UNITS_EXT
    enum: GL_MAX_COMBINED_IMAGE_UNITS_AND_FRAGMENT_OUTPUTS_EXT
    enum: GL_IMAGE_BINDING_NAME_EXT
    enum: GL_IMAGE_BINDING_LEVEL_EXT
    enum: GL_IMAGE_BINDING_LAYERED_EXT
    enum: GL_IMAGE_BINDING_LAYER_EXT
    enum: GL_IMAGE_BINDING_ACCESS_EXT
    enum: GL_IMAGE_1D_EXT
    enum: GL_IMAGE_2D_EXT
    enum: GL_IMAGE_3D_EXT
    enum: GL_IMAGE_2D_RECT_EXT
    enum: GL_IMAGE_CUBE_EXT
    enum: GL_IMAGE_BUFFER_EXT
    enum: GL_IMAGE_1D_ARRAY_EXT
    enum: GL_IMAGE_2D_ARRAY_EXT
    enum: GL_IMAGE_CUBE_MAP_ARRAY_EXT
    enum: GL_IMAGE_2D_MULTISAMPLE_EXT
    enum: GL_IMAGE_2D_MULTISAMPLE_ARRAY_EXT
    enum: GL_INT_IMAGE_1D_EXT
    enum: GL_INT_IMAGE_2D_EXT
    enum: GL_INT_IMAGE_3D_EXT
    enum: GL_INT_IMAGE_2D_RECT_EXT
    enum: GL_INT_IMAGE_CUBE_EXT
    enum: GL_INT_IMAGE_BUFFER_EXT
    enum: GL_INT_IMAGE_1D_ARRAY_EXT
    enum: GL_INT_IMAGE_2D_ARRAY_EXT
    enum: GL_INT_IMAGE_CUBE_MAP_ARRAY_EXT
    enum: GL_INT_IMAGE_2D_MULTISAMPLE_EXT
    enum: GL_INT_IMAGE_2D_MULTISAMPLE_ARRAY_EXT
    enum: GL_UNSIGNED_INT_IMAGE_1D_EXT
    enum: GL_UNSIGNED_INT_IMAGE_2D_EXT
    enum: GL_UNSIGNED_INT_IMAGE_3D_EXT
    enum: GL_UNSIGNED_INT_IMAGE_2D_RECT_EXT
    enum: GL_UNSIGNED_INT_IMAGE_CUBE_EXT
    enum: GL_UNSIGNED_INT_IMAGE_BUFFER_EXT
    enum: GL_UNSIGNED_INT_IMAGE_1D_ARRAY_EXT
    enum: GL_UNSIGNED_INT_IMAGE_2D_ARRAY_EXT
    enum: GL_UNSIGNED_INT_IMAGE_CUBE_MAP_ARRAY_EXT
    enum: GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE_EXT
    enum: GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE_ARRAY_EXT
    enum: GL_MAX_IMAGE_SAMPLES_EXT
    enum: GL_IMAGE_BINDING_FORMAT_EXT
    enum: GL_VERTEX_ATTRIB_ARRAY_BARRIER_BIT_EXT
    enum: GL_ELEMENT_ARRAY_BARRIER_BIT_EXT
    enum: GL_UNIFORM_BARRIER_BIT_EXT
    enum: GL_TEXTURE_FETCH_BARRIER_BIT_EXT
    enum: GL_SHADER_IMAGE_ACCESS_BARRIER_BIT_EXT
    enum: GL_COMMAND_BARRIER_BIT_EXT
    enum: GL_PIXEL_BUFFER_BARRIER_BIT_EXT
    enum: GL_TEXTURE_UPDATE_BARRIER_BIT_EXT
    enum: GL_BUFFER_UPDATE_BARRIER_BIT_EXT
    enum: GL_FRAMEBUFFER_BARRIER_BIT_EXT
    enum: GL_TRANSFORM_FEEDBACK_BARRIER_BIT_EXT
    enum: GL_ATOMIC_COUNTER_BARRIER_BIT_EXT
    enum: GL_ALL_BARRIER_BITS_EXT
    enum: GL_DOUBLE_VEC2_EXT
    enum: GL_DOUBLE_VEC3_EXT
    enum: GL_DOUBLE_VEC4_EXT
    enum: GL_DOUBLE_MAT2_EXT
    enum: GL_DOUBLE_MAT3_EXT
    enum: GL_DOUBLE_MAT4_EXT
    enum: GL_DOUBLE_MAT2x3_EXT
    enum: GL_DOUBLE_MAT2x4_EXT
    enum: GL_DOUBLE_MAT3x2_EXT
    enum: GL_DOUBLE_MAT3x4_EXT
    enum: GL_DOUBLE_MAT4x2_EXT
    enum: GL_DOUBLE_MAT4x3_EXT
    enum: GL_MAX_GEOMETRY_PROGRAM_INVOCATIONS_NV
    enum: GL_MIN_FRAGMENT_INTERPOLATION_OFFSET_NV
    enum: GL_MAX_FRAGMENT_INTERPOLATION_OFFSET_NV
    enum: GL_FRAGMENT_PROGRAM_INTERPOLATION_OFFSET_BITS_NV
    enum: GL_MIN_PROGRAM_TEXTURE_GATHER_OFFSET_NV
    enum: GL_MAX_PROGRAM_TEXTURE_GATHER_OFFSET_NV
    enum: GL_MAX_PROGRAM_SUBROUTINE_PARAMETERS_NV
    enum: GL_MAX_PROGRAM_SUBROUTINE_NUM_NV
    enum: GL_INT64_NV
    enum: GL_UNSIGNED_INT64_NV
    enum: GL_INT8_NV
    enum: GL_INT8_VEC2_NV
    enum: GL_INT8_VEC3_NV
    enum: GL_INT8_VEC4_NV
    enum: GL_INT16_NV
    enum: GL_INT16_VEC2_NV
    enum: GL_INT16_VEC3_NV
    enum: GL_INT16_VEC4_NV
    enum: GL_INT64_VEC2_NV
    enum: GL_INT64_VEC3_NV
    enum: GL_INT64_VEC4_NV
    enum: GL_UNSIGNED_INT8_NV
    enum: GL_UNSIGNED_INT8_VEC2_NV
    enum: GL_UNSIGNED_INT8_VEC3_NV
    enum: GL_UNSIGNED_INT8_VEC4_NV
    enum: GL_UNSIGNED_INT16_NV
    enum: GL_UNSIGNED_INT16_VEC2_NV
    enum: GL_UNSIGNED_INT16_VEC3_NV
    enum: GL_UNSIGNED_INT16_VEC4_NV
    enum: GL_UNSIGNED_INT64_VEC2_NV
    enum: GL_UNSIGNED_INT64_VEC3_NV
    enum: GL_UNSIGNED_INT64_VEC4_NV
    enum: GL_FLOAT16_NV
    enum: GL_FLOAT16_VEC2_NV
    enum: GL_FLOAT16_VEC3_NV
    enum: GL_FLOAT16_VEC4_NV
    enum: GL_SHADER_GLOBAL_ACCESS_BARRIER_BIT_NV
    enum: GL_MAX_PROGRAM_PATCH_ATTRIBS_NV
    enum: GL_TESS_CONTROL_PROGRAM_NV
    enum: GL_TESS_EVALUATION_PROGRAM_NV
    enum: GL_TESS_CONTROL_PROGRAM_PARAMETER_BUFFER_NV
    enum: GL_TESS_EVALUATION_PROGRAM_PARAMETER_BUFFER_NV
    enum: GL_COVERAGE_SAMPLES_NV
    enum: GL_COLOR_SAMPLES_NV
    enum: GL_DATA_BUFFER_AMD
    enum: GL_PERFORMANCE_MONITOR_AMD
    enum: GL_QUERY_OBJECT_AMD
    enum: GL_VERTEX_ARRAY_OBJECT_AMD
    enum: GL_SAMPLER_OBJECT_AMD
    enum: GL_MAX_DEBUG_MESSAGE_LENGTH_AMD
    enum: GL_MAX_DEBUG_LOGGED_MESSAGES_AMD
    enum: GL_DEBUG_LOGGED_MESSAGES_AMD
    enum: GL_DEBUG_SEVERITY_HIGH_AMD
    enum: GL_DEBUG_SEVERITY_MEDIUM_AMD
    enum: GL_DEBUG_SEVERITY_LOW_AMD
    enum: GL_DEBUG_CATEGORY_API_ERROR_AMD
    enum: GL_DEBUG_CATEGORY_WINDOW_SYSTEM_AMD
    enum: GL_DEBUG_CATEGORY_DEPRECATION_AMD
    enum: GL_DEBUG_CATEGORY_UNDEFINED_BEHAVIOR_AMD
    enum: GL_DEBUG_CATEGORY_PERFORMANCE_AMD
    enum: GL_DEBUG_CATEGORY_SHADER_COMPILER_AMD
    enum: GL_DEBUG_CATEGORY_APPLICATION_AMD
    enum: GL_DEBUG_CATEGORY_OTHER_AMD
    enum: GL_SURFACE_STATE_NV
    enum: GL_SURFACE_REGISTERED_NV
    enum: GL_SURFACE_MAPPED_NV
    enum: GL_WRITE_DISCARD_NV
    enum: GL_DEPTH_CLAMP_NEAR_AMD
    enum: GL_DEPTH_CLAMP_FAR_AMD
    enum: GL_TEXTURE_SRGB_DECODE_EXT
    enum: GL_DECODE_EXT
    enum: GL_SKIP_DECODE_EXT
    enum: GL_TEXTURE_COVERAGE_SAMPLES_NV
    enum: GL_TEXTURE_COLOR_SAMPLES_NV
    enum: GL_FACTOR_MIN_AMD
    enum: GL_FACTOR_MAX_AMD
    enum: GL_SUBSAMPLE_DISTANCE_AMD
    enum: GL_SYNC_X11_FENCE_EXT
    enum: GL_SCALED_RESOLVE_FASTEST_EXT
    enum: GL_SCALED_RESOLVE_NICEST_EXT
    enum: GL_PATH_FORMAT_SVG_NV
    enum: GL_PATH_FORMAT_PS_NV
    enum: GL_STANDARD_FONT_NAME_NV
    enum: GL_SYSTEM_FONT_NAME_NV
    enum: GL_FILE_NAME_NV
    enum: GL_PATH_STROKE_WIDTH_NV
    enum: GL_PATH_END_CAPS_NV
    enum: GL_PATH_INITIAL_END_CAP_NV
    enum: GL_PATH_TERMINAL_END_CAP_NV
    enum: GL_PATH_JOIN_STYLE_NV
    enum: GL_PATH_MITER_LIMIT_NV
    enum: GL_PATH_DASH_CAPS_NV
    enum: GL_PATH_INITIAL_DASH_CAP_NV
    enum: GL_PATH_TERMINAL_DASH_CAP_NV
    enum: GL_PATH_DASH_OFFSET_NV
    enum: GL_PATH_CLIENT_LENGTH_NV
    enum: GL_PATH_FILL_MODE_NV
    enum: GL_PATH_FILL_MASK_NV
    enum: GL_PATH_FILL_COVER_MODE_NV
    enum: GL_PATH_STROKE_COVER_MODE_NV
    enum: GL_PATH_STROKE_MASK_NV
    enum: GL_PATH_SAMPLE_QUALITY_NV
    enum: GL_PATH_STROKE_BOUND_NV
    enum: GL_PATH_STROKE_OVERSAMPLE_COUNT_NV
    enum: GL_COUNT_UP_NV
    enum: GL_COUNT_DOWN_NV
    enum: GL_PATH_OBJECT_BOUNDING_BOX_NV
    enum: GL_CONVEX_HULL_NV
    enum: GL_MULTI_HULLS_NV
    enum: GL_BOUNDING_BOX_NV
    enum: GL_TRANSLATE_X_NV
    enum: GL_TRANSLATE_Y_NV
    enum: GL_TRANSLATE_2D_NV
    enum: GL_TRANSLATE_3D_NV
    enum: GL_AFFINE_2D_NV
    enum: GL_PROJECTIVE_2D_NV
    enum: GL_AFFINE_3D_NV
    enum: GL_PROJECTIVE_3D_NV
    enum: GL_TRANSPOSE_AFFINE_2D_NV
    enum: GL_TRANSPOSE_PROJECTIVE_2D_NV
    enum: GL_TRANSPOSE_AFFINE_3D_NV
    enum: GL_TRANSPOSE_PROJECTIVE_3D_NV
    enum: GL_UTF8_NV
    enum: GL_UTF16_NV
    enum: GL_BOUNDING_BOX_OF_BOUNDING_BOXES_NV
    enum: GL_PATH_COMMAND_COUNT_NV
    enum: GL_PATH_COORD_COUNT_NV
    enum: GL_PATH_DASH_ARRAY_COUNT_NV
    enum: GL_PATH_COMPUTED_LENGTH_NV
    enum: GL_PATH_FILL_BOUNDING_BOX_NV
    enum: GL_PATH_STROKE_BOUNDING_BOX_NV
    enum: GL_SQUARE_NV
    enum: GL_ROUND_NV
    enum: GL_TRIANGULAR_NV
    enum: GL_BEVEL_NV
    enum: GL_MITER_REVERT_NV
    enum: GL_MITER_TRUNCATE_NV
    enum: GL_SKIP_MISSING_GLYPH_NV
    enum: GL_USE_MISSING_GLYPH_NV
    enum: GL_PATH_ERROR_POSITION_NV
    enum: GL_PATH_FOG_GEN_MODE_NV
    enum: GL_ACCUM_ADJACENT_PAIRS_NV
    enum: GL_ADJACENT_PAIRS_NV
    enum: GL_FIRST_TO_REST_NV
    enum: GL_PATH_GEN_MODE_NV
    enum: GL_PATH_GEN_COEFF_NV
    enum: GL_PATH_GEN_COLOR_FORMAT_NV
    enum: GL_PATH_GEN_COMPONENTS_NV
    enum: GL_PATH_STENCIL_FUNC_NV
    enum: GL_PATH_STENCIL_REF_NV
    enum: GL_PATH_STENCIL_VALUE_MASK_NV
    enum: GL_PATH_STENCIL_DEPTH_OFFSET_FACTOR_NV
    enum: GL_PATH_STENCIL_DEPTH_OFFSET_UNITS_NV
    enum: GL_PATH_COVER_DEPTH_FUNC_NV
    enum: GL_PATH_DASH_OFFSET_RESET_NV
    enum: GL_MOVE_TO_RESETS_NV
    enum: GL_MOVE_TO_CONTINUES_NV
    enum: GL_CLOSE_PATH_NV
    enum: GL_MOVE_TO_NV
    enum: GL_RELATIVE_MOVE_TO_NV
    enum: GL_LINE_TO_NV
    enum: GL_RELATIVE_LINE_TO_NV
    enum: GL_HORIZONTAL_LINE_TO_NV
    enum: GL_RELATIVE_HORIZONTAL_LINE_TO_NV
    enum: GL_VERTICAL_LINE_TO_NV
    enum: GL_RELATIVE_VERTICAL_LINE_TO_NV
    enum: GL_QUADRATIC_CURVE_TO_NV
    enum: GL_RELATIVE_QUADRATIC_CURVE_TO_NV
    enum: GL_CUBIC_CURVE_TO_NV
    enum: GL_RELATIVE_CUBIC_CURVE_TO_NV
    enum: GL_SMOOTH_QUADRATIC_CURVE_TO_NV
    enum: GL_RELATIVE_SMOOTH_QUADRATIC_CURVE_TO_NV
    enum: GL_SMOOTH_CUBIC_CURVE_TO_NV
    enum: GL_RELATIVE_SMOOTH_CUBIC_CURVE_TO_NV
    enum: GL_SMALL_CCW_ARC_TO_NV
    enum: GL_RELATIVE_SMALL_CCW_ARC_TO_NV
    enum: GL_SMALL_CW_ARC_TO_NV
    enum: GL_RELATIVE_SMALL_CW_ARC_TO_NV
    enum: GL_LARGE_CCW_ARC_TO_NV
    enum: GL_RELATIVE_LARGE_CCW_ARC_TO_NV
    enum: GL_LARGE_CW_ARC_TO_NV
    enum: GL_RELATIVE_LARGE_CW_ARC_TO_NV
    enum: GL_RESTART_PATH_NV
    enum: GL_DUP_FIRST_CUBIC_CURVE_TO_NV
    enum: GL_DUP_LAST_CUBIC_CURVE_TO_NV
    enum: GL_RECT_NV
    enum: GL_CIRCULAR_CCW_ARC_TO_NV
    enum: GL_CIRCULAR_CW_ARC_TO_NV
    enum: GL_CIRCULAR_TANGENT_ARC_TO_NV
    enum: GL_ARC_TO_NV
    enum: GL_RELATIVE_ARC_TO_NV
    enum: GL_BOLD_BIT_NV
    enum: GL_ITALIC_BIT_NV
    enum: GL_GLYPH_WIDTH_BIT_NV
    enum: GL_GLYPH_HEIGHT_BIT_NV
    enum: GL_GLYPH_HORIZONTAL_BEARING_X_BIT_NV
    enum: GL_GLYPH_HORIZONTAL_BEARING_Y_BIT_NV
    enum: GL_GLYPH_HORIZONTAL_BEARING_ADVANCE_BIT_NV
    enum: GL_GLYPH_VERTICAL_BEARING_X_BIT_NV
    enum: GL_GLYPH_VERTICAL_BEARING_Y_BIT_NV
    enum: GL_GLYPH_VERTICAL_BEARING_ADVANCE_BIT_NV
    enum: GL_GLYPH_HAS_KERNING_NV
    enum: GL_FONT_X_MIN_BOUNDS_NV
    enum: GL_FONT_Y_MIN_BOUNDS_NV
    enum: GL_FONT_X_MAX_BOUNDS_NV
    enum: GL_FONT_Y_MAX_BOUNDS_NV
    enum: GL_FONT_UNITS_PER_EM_NV
    enum: GL_FONT_ASCENDER_NV
    enum: GL_FONT_DESCENDER_NV
    enum: GL_FONT_HEIGHT_NV
    enum: GL_FONT_MAX_ADVANCE_WIDTH_NV
    enum: GL_FONT_MAX_ADVANCE_HEIGHT_NV
    enum: GL_FONT_UNDERLINE_POSITION_NV
    enum: GL_FONT_UNDERLINE_THICKNESS_NV
    enum: GL_FONT_HAS_KERNING_NV
    enum: GL_EXTERNAL_VIRTUAL_MEMORY_BUFFER_AMD
    enum: GL_SET_AMD
    enum: GL_REPLACE_VALUE_AMD
    enum: GL_STENCIL_OP_VALUE_AMD
    enum: GL_STENCIL_BACK_OP_VALUE_AMD
    enum: GL_QUERY_BUFFER_AMD
    enum: GL_QUERY_BUFFER_BINDING_AMD
    enum: GL_QUERY_RESULT_NO_WAIT_AMD
    enum: GL_VIRTUAL_PAGE_SIZE_X_AMD
    enum: GL_VIRTUAL_PAGE_SIZE_Y_AMD
    enum: GL_VIRTUAL_PAGE_SIZE_Z_AMD
    enum: GL_MAX_SPARSE_TEXTURE_SIZE_AMD
    enum: GL_MAX_SPARSE_3D_TEXTURE_SIZE_AMD
    enum: GL_MAX_SPARSE_ARRAY_TEXTURE_LAYERS
    enum: GL_MIN_SPARSE_LEVEL_AMD
    enum: GL_MIN_LOD_WARNING_AMD
    enum: GL_TEXTURE_STORAGE_SPARSE_BIT_AMD
    enum: GL_VERSION_1_2
    enum: GL_VERSION_1_3
    enum: GL_VERSION_1_4
    enum: GL_VERSION_1_5
    enum: GL_VERSION_2_0
    enum: GL_VERSION_2_1
    enum: GL_VERSION_3_0
    enum: GL_VERSION_3_1
    enum: GL_VERSION_3_2
    enum: GL_VERSION_3_3
    enum: GL_VERSION_4_0
    enum: GL_VERSION_4_1
    enum: GL_VERSION_4_2
    enum: GL_VERSION_4_3
    enum: GL_ARB_multitexture
    enum: GL_ARB_transpose_matrix
    enum: GL_ARB_multisample
    enum: GL_ARB_texture_env_add
    enum: GL_ARB_texture_cube_map
    enum: GL_ARB_texture_compression
    enum: GL_ARB_texture_border_clamp
    enum: GL_ARB_point_parameters
    enum: GL_ARB_vertex_blend
    enum: GL_ARB_matrix_palette
    enum: GL_ARB_texture_env_combine
    enum: GL_ARB_texture_env_crossbar
    enum: GL_ARB_texture_env_dot3
    enum: GL_ARB_texture_mirrored_repeat
    enum: GL_ARB_depth_texture
    enum: GL_ARB_shadow
    enum: GL_ARB_shadow_ambient
    enum: GL_ARB_window_pos
    enum: GL_ARB_vertex_program
    enum: GL_ARB_fragment_program
    enum: GL_ARB_vertex_buffer_object
    enum: GL_ARB_occlusion_query
    enum: GL_ARB_shader_objects
    enum: GL_ARB_vertex_shader
    enum: GL_ARB_fragment_shader
    enum: GL_ARB_shading_language_100
    enum: GL_ARB_texture_non_power_of_two
    enum: GL_ARB_point_sprite
    enum: GL_ARB_fragment_program_shadow
    enum: GL_ARB_draw_buffers
    enum: GL_ARB_texture_rectangle
    enum: GL_ARB_color_buffer_float
    enum: GL_ARB_half_float_pixel
    enum: GL_ARB_texture_float
    enum: GL_ARB_pixel_buffer_object
    enum: GL_ARB_depth_buffer_float
    enum: GL_ARB_draw_instanced
    enum: GL_ARB_framebuffer_object
    enum: GL_ARB_framebuffer_sRGB
    enum: GL_ARB_geometry_shader4
    enum: GL_ARB_half_float_vertex
    enum: GL_ARB_instanced_arrays
    enum: GL_ARB_map_buffer_range
    enum: GL_ARB_texture_buffer_object
    enum: GL_ARB_texture_compression_rgtc
    enum: GL_ARB_texture_rg
    enum: GL_ARB_vertex_array_object
    enum: GL_ARB_uniform_buffer_object
    enum: GL_ARB_compatibility
    enum: GL_ARB_copy_buffer
    enum: GL_ARB_shader_texture_lod
    enum: GL_ARB_depth_clamp
    enum: GL_ARB_draw_elements_base_vertex
    enum: GL_ARB_fragment_coord_conventions
    enum: GL_ARB_provoking_vertex
    enum: GL_ARB_seamless_cube_map
    enum: GL_ARB_sync
    enum: GL_ARB_texture_multisample
    enum: GL_ARB_vertex_array_bgra
    enum: GL_ARB_draw_buffers_blend
    enum: GL_ARB_sample_shading
    enum: GL_ARB_texture_cube_map_array
    enum: GL_ARB_texture_gather
    enum: GL_ARB_texture_query_lod
    enum: GL_ARB_shading_language_include
    enum: GL_ARB_texture_compression_bptc
    enum: GL_ARB_blend_func_extended
    enum: GL_ARB_explicit_attrib_location
    enum: GL_ARB_occlusion_query2
    enum: GL_ARB_sampler_objects
    enum: GL_ARB_shader_bit_encoding
    enum: GL_ARB_texture_rgb10_a2ui
    enum: GL_ARB_texture_swizzle
    enum: GL_ARB_timer_query
    enum: GL_ARB_vertex_type_2_10_10_10_rev
    enum: GL_ARB_draw_indirect
    enum: GL_ARB_gpu_shader5
    enum: GL_ARB_gpu_shader_fp64
    enum: GL_ARB_shader_subroutine
    enum: GL_ARB_tessellation_shader
    enum: GL_ARB_texture_buffer_object_rgb32
    enum: GL_ARB_transform_feedback2
    enum: GL_ARB_transform_feedback3
    enum: GL_ARB_ES2_compatibility
    enum: GL_ARB_get_program_binary
    enum: GL_ARB_separate_shader_objects
    enum: GL_ARB_vertex_attrib_64bit
    enum: GL_ARB_viewport_array
    enum: GL_ARB_cl_event
    enum: GL_ARB_debug_output
    enum: GL_ARB_robustness
    enum: GL_ARB_shader_stencil_export
    enum: GL_ARB_base_instance
    enum: GL_ARB_shading_language_420pack
    enum: GL_ARB_transform_feedback_instanced
    enum: GL_ARB_compressed_texture_pixel_storage
    enum: GL_ARB_conservative_depth
    enum: GL_ARB_internalformat_query
    enum: GL_ARB_map_buffer_alignment
    enum: GL_ARB_shader_atomic_counters
    enum: GL_ARB_shader_image_load_store
    enum: GL_ARB_shading_language_packing
    enum: GL_ARB_texture_storage
    enum: GL_KHR_texture_compression_astc_ldr
    enum: GL_KHR_debug
    enum: GL_ARB_arrays_of_arrays
    enum: GL_ARB_clear_buffer_object
    enum: GL_ARB_compute_shader
    enum: GL_ARB_copy_image
    enum: GL_ARB_texture_view
    enum: GL_ARB_vertex_attrib_binding
    enum: GL_ARB_robustness_isolation
    enum: GL_ARB_ES3_compatibility
    enum: GL_ARB_explicit_uniform_location
    enum: GL_ARB_fragment_layer_viewport
    enum: GL_ARB_framebuffer_no_attachments
    enum: GL_ARB_internalformat_query2
    enum: GL_ARB_invalidate_subdata
    enum: GL_ARB_multi_draw_indirect
    enum: GL_ARB_program_interface_query
    enum: GL_ARB_robust_buffer_access_behavior
    enum: GL_ARB_shader_image_size
    enum: GL_ARB_shader_storage_buffer_object
    enum: GL_ARB_stencil_texturing
    enum: GL_ARB_texture_buffer_range
    enum: GL_ARB_texture_query_levels
    enum: GL_ARB_texture_storage_multisample
    enum: GL_EXT_abgr
    enum: GL_EXT_blend_color
    enum: GL_EXT_polygon_offset
    enum: GL_EXT_texture
    enum: GL_EXT_texture3D
    enum: GL_SGIS_texture_filter4
    enum: GL_EXT_subtexture
    enum: GL_EXT_copy_texture
    enum: GL_EXT_histogram
    enum: GL_EXT_convolution
    enum: GL_SGI_color_matrix
    enum: GL_SGI_color_table
    enum: GL_SGIX_pixel_texture
    enum: GL_SGIS_pixel_texture
    enum: GL_SGIS_texture4D
    enum: GL_SGI_texture_color_table
    enum: GL_EXT_cmyka
    enum: GL_EXT_texture_object
    enum: GL_SGIS_detail_texture
    enum: GL_SGIS_sharpen_texture
    enum: GL_EXT_packed_pixels
    enum: GL_SGIS_texture_lod
    enum: GL_SGIS_multisample
    enum: GL_EXT_rescale_normal
    enum: GL_EXT_vertex_array
    enum: GL_EXT_misc_attribute
    enum: GL_SGIS_generate_mipmap
    enum: GL_SGIX_clipmap
    enum: GL_SGIX_shadow
    enum: GL_SGIS_texture_edge_clamp
    enum: GL_SGIS_texture_border_clamp
    enum: GL_EXT_blend_minmax
    enum: GL_EXT_blend_subtract
    enum: GL_EXT_blend_logic_op
    enum: GL_SGIX_interlace
    enum: GL_SGIX_pixel_tiles
    enum: GL_SGIX_texture_select
    enum: GL_SGIX_sprite
    enum: GL_SGIX_texture_multi_buffer
    enum: GL_EXT_point_parameters
    enum: GL_SGIS_point_parameters
    enum: GL_SGIX_instruments
    enum: GL_SGIX_texture_scale_bias
    enum: GL_SGIX_framezoom
    enum: GL_SGIX_tag_sample_buffer
    enum: GL_SGIX_polynomial_ffd
    enum: GL_SGIX_reference_plane
    enum: GL_SGIX_flush_raster
    enum: GL_SGIX_depth_texture
    enum: GL_SGIS_fog_function
    enum: GL_SGIX_fog_offset
    enum: GL_HP_image_transform
    enum: GL_HP_convolution_border_modes
    enum: GL_SGIX_texture_add_env
    enum: GL_EXT_color_subtable
    enum: GL_PGI_vertex_hints
    enum: GL_PGI_misc_hints
    enum: GL_EXT_paletted_texture
    enum: GL_EXT_clip_volume_hint
    enum: GL_SGIX_list_priority
    enum: GL_SGIX_ir_instrument1
    enum: GL_SGIX_calligraphic_fragment
    enum: GL_SGIX_texture_lod_bias
    enum: GL_SGIX_shadow_ambient
    enum: GL_EXT_index_texture
    enum: GL_EXT_index_material
    enum: GL_EXT_index_func
    enum: GL_EXT_index_array_formats
    enum: GL_EXT_compiled_vertex_array
    enum: GL_EXT_cull_vertex
    enum: GL_SGIX_ycrcb
    enum: GL_SGIX_fragment_lighting
    enum: GL_IBM_rasterpos_clip
    enum: GL_HP_texture_lighting
    enum: GL_EXT_draw_range_elements
    enum: GL_WIN_phong_shading
    enum: GL_WIN_specular_fog
    enum: GL_EXT_light_texture
    enum: GL_SGIX_blend_alpha_minmax
    enum: GL_EXT_bgra
    enum: GL_SGIX_async
    enum: GL_SGIX_async_pixel
    enum: GL_SGIX_async_histogram
    enum: GL_INTEL_parallel_arrays
    enum: GL_HP_occlusion_test
    enum: GL_EXT_pixel_transform
    enum: GL_EXT_pixel_transform_color_table
    enum: GL_EXT_shared_texture_palette
    enum: GL_EXT_separate_specular_color
    enum: GL_EXT_secondary_color
    enum: GL_EXT_texture_perturb_normal
    enum: GL_EXT_multi_draw_arrays
    enum: GL_EXT_fog_coord
    enum: GL_REND_screen_coordinates
    enum: GL_EXT_coordinate_frame
    enum: GL_EXT_texture_env_combine
    enum: GL_APPLE_specular_vector
    enum: GL_APPLE_transform_hint
    enum: GL_SGIX_fog_scale
    enum: GL_SUNX_constant_data
    enum: GL_SUN_global_alpha
    enum: GL_SUN_triangle_list
    enum: GL_SUN_vertex
    enum: GL_EXT_blend_func_separate
    enum: GL_INGR_blend_func_separate
    enum: GL_INGR_color_clamp
    enum: GL_INGR_interlace_read
    enum: GL_EXT_stencil_wrap
    enum: GL_EXT_422_pixels
    enum: GL_NV_texgen_reflection
    enum: GL_SUN_convolution_border_modes
    enum: GL_EXT_texture_env_add
    enum: GL_EXT_texture_lod_bias
    enum: GL_EXT_texture_filter_anisotropic
    enum: GL_EXT_vertex_weighting
    enum: GL_NV_light_max_exponent
    enum: GL_NV_vertex_array_range
    enum: GL_NV_register_combiners
    enum: GL_NV_fog_distance
    enum: GL_NV_texgen_emboss
    enum: GL_NV_blend_square
    enum: GL_NV_texture_env_combine4
    enum: GL_MESA_resize_buffers
    enum: GL_MESA_window_pos
    enum: GL_IBM_cull_vertex
    enum: GL_IBM_multimode_draw_arrays
    enum: GL_IBM_vertex_array_lists
    enum: GL_SGIX_subsample
    enum: GL_SGIX_ycrcba
    enum: GL_SGIX_ycrcb_subsample
    enum: GL_SGIX_depth_pass_instrument
    enum: GL_3DFX_texture_compression_FXT1
    enum: GL_3DFX_multisample
    enum: GL_3DFX_tbuffer
    enum: GL_EXT_multisample
    enum: GL_SGIX_vertex_preclip
    enum: GL_SGIX_convolution_accuracy
    enum: GL_SGIX_resample
    enum: GL_SGIS_point_line_texgen
    enum: GL_SGIS_texture_color_mask
    enum: GL_SGIX_igloo_interface
    enum: GL_EXT_texture_env_dot3
    enum: GL_ATI_texture_mirror_once
    enum: GL_NV_fence
    enum: GL_NV_evaluators
    enum: GL_NV_packed_depth_stencil
    enum: GL_NV_register_combiners2
    enum: GL_NV_texture_compression_vtc
    enum: GL_NV_texture_rectangle
    enum: GL_NV_texture_shader
    enum: GL_NV_texture_shader2
    enum: GL_NV_vertex_array_range2
    enum: GL_NV_vertex_program
    enum: GL_SGIX_texture_coordinate_clamp
    enum: GL_SGIX_scalebias_hint
    enum: GL_OML_interlace
    enum: GL_OML_subsample
    enum: GL_OML_resample
    enum: GL_NV_copy_depth_to_color
    enum: GL_ATI_envmap_bumpmap
    enum: GL_ATI_fragment_shader
    enum: GL_ATI_pn_triangles
    enum: GL_ATI_vertex_array_object
    enum: GL_EXT_vertex_shader
    enum: GL_ATI_vertex_streams
    enum: GL_ATI_element_array
    enum: GL_SUN_mesh_array
    enum: GL_SUN_slice_accum
    enum: GL_NV_multisample_filter_hint
    enum: GL_NV_depth_clamp
    enum: GL_NV_occlusion_query
    enum: GL_NV_point_sprite
    enum: GL_NV_texture_shader3
    enum: GL_NV_vertex_program1_1
    enum: GL_EXT_shadow_funcs
    enum: GL_EXT_stencil_two_side
    enum: GL_ATI_text_fragment_shader
    enum: GL_APPLE_client_storage
    enum: GL_APPLE_element_array
    enum: GL_APPLE_fence
    enum: GL_APPLE_vertex_array_object
    enum: GL_APPLE_vertex_array_range
    enum: GL_APPLE_ycbcr_422
    enum: GL_S3_s3tc
    enum: GL_ATI_draw_buffers
    enum: GL_ATI_pixel_format_float
    enum: GL_ATI_texture_env_combine3
    enum: GL_ATI_texture_float
    enum: GL_NV_float_buffer
    enum: GL_NV_fragment_program
    enum: GL_NV_half_float
    enum: GL_NV_pixel_data_range
    enum: GL_NV_primitive_restart
    enum: GL_NV_texture_expand_normal
    enum: GL_NV_vertex_program2
    enum: GL_ATI_map_object_buffer
    enum: GL_ATI_separate_stencil
    enum: GL_ATI_vertex_attrib_array_object
    enum: GL_OES_read_format
    enum: GL_EXT_depth_bounds_test
    enum: GL_EXT_texture_mirror_clamp
    enum: GL_EXT_blend_equation_separate
    enum: GL_MESA_pack_invert
    enum: GL_MESA_ycbcr_texture
    enum: GL_EXT_pixel_buffer_object
    enum: GL_NV_fragment_program_option
    enum: GL_NV_fragment_program2
    enum: GL_NV_vertex_program2_option
    enum: GL_NV_vertex_program3
    enum: GL_EXT_framebuffer_object
    enum: GL_GREMEDY_string_marker
    enum: GL_EXT_packed_depth_stencil
    enum: GL_EXT_stencil_clear_tag
    enum: GL_EXT_texture_sRGB
    enum: GL_EXT_framebuffer_blit
    enum: GL_EXT_framebuffer_multisample
    enum: GL_MESAX_texture_stack
    enum: GL_EXT_timer_query
    enum: GL_EXT_gpu_program_parameters
    enum: GL_APPLE_flush_buffer_range
    enum: GL_NV_gpu_program4
    enum: GL_NV_geometry_program4
    enum: GL_EXT_geometry_shader4
    enum: GL_NV_vertex_program4
    enum: GL_EXT_gpu_shader4
    enum: GL_EXT_draw_instanced
    enum: GL_EXT_packed_float
    enum: GL_EXT_texture_array
    enum: GL_EXT_texture_buffer_object
    enum: GL_EXT_texture_compression_latc
    enum: GL_EXT_texture_compression_rgtc
    enum: GL_EXT_texture_shared_exponent
    enum: GL_NV_depth_buffer_float
    enum: GL_NV_fragment_program4
    enum: GL_NV_framebuffer_multisample_coverage
    enum: GL_EXT_framebuffer_sRGB
    enum: GL_NV_geometry_shader4
    enum: GL_NV_parameter_buffer_object
    enum: GL_EXT_draw_buffers2
    enum: GL_NV_transform_feedback
    enum: GL_EXT_bindable_uniform
    enum: GL_EXT_texture_integer
    enum: GL_GREMEDY_frame_terminator
    enum: GL_NV_conditional_render
    enum: GL_NV_present_video
    enum: GL_EXT_transform_feedback
    enum: GL_EXT_direct_state_access
    enum: GL_EXT_vertex_array_bgra
    enum: GL_EXT_texture_swizzle
    enum: GL_NV_explicit_multisample
    enum: GL_NV_transform_feedback2
    enum: GL_ATI_meminfo
    enum: GL_AMD_performance_monitor
    enum: GL_AMD_texture_texture4
    enum: GL_AMD_vertex_shader_tesselator
    enum: GL_EXT_provoking_vertex
    enum: GL_EXT_texture_snorm
    enum: GL_AMD_draw_buffers_blend
    enum: GL_APPLE_texture_range
    enum: GL_APPLE_float_pixels
    enum: GL_APPLE_vertex_program_evaluators
    enum: GL_APPLE_aux_depth_stencil
    enum: GL_APPLE_object_purgeable
    enum: GL_APPLE_row_bytes
    enum: GL_APPLE_rgb_422
    enum: GL_NV_video_capture
    enum: GL_NV_copy_image
    enum: GL_EXT_separate_shader_objects
    enum: GL_NV_parameter_buffer_object2
    enum: GL_NV_shader_buffer_load
    enum: GL_NV_vertex_buffer_unified_memory
    enum: GL_NV_texture_barrier
    enum: GL_AMD_shader_stencil_export
    enum: GL_AMD_seamless_cubemap_per_texture
    enum: GL_AMD_conservative_depth
    enum: GL_EXT_shader_image_load_store
    enum: GL_EXT_vertex_attrib_64bit
    enum: GL_NV_gpu_program5
    enum: GL_NV_gpu_shader5
    enum: GL_NV_shader_buffer_store
    enum: GL_NV_tessellation_program5
    enum: GL_NV_vertex_attrib_integer_64bit
    enum: GL_NV_multisample_coverage
    enum: GL_AMD_name_gen_delete
    enum: GL_AMD_debug_output
    enum: GL_NV_vdpau_interop
    enum: GL_AMD_transform_feedback3_lines_triangles
    enum: GL_AMD_depth_clamp_separate
    enum: GL_EXT_texture_sRGB_decode
    enum: GL_NV_texture_multisample
    enum: GL_AMD_blend_minmax_factor
    enum: GL_AMD_sample_positions
    enum: GL_EXT_x11_sync_object
    enum: GL_AMD_multi_draw_indirect
    enum: GL_EXT_framebuffer_multisample_blit_scaled
    enum: GL_NV_path_rendering
    enum: GL_AMD_pinned_memory
    enum: GL_AMD_stencil_operation_extended
    enum: GL_AMD_vertex_shader_viewport_index
    enum: GL_AMD_vertex_shader_layer
    enum: GL_NV_bindless_texture
    enum: GL_NV_shader_atomic_float
    enum: GL_AMD_query_buffer_object
    enum: GL_AMD_sparse_texture


# from /usr/include/GL/gl.h:

ctypedef unsigned int	GLenum
ctypedef unsigned char	GLboolean
ctypedef unsigned int	GLbitfield
ctypedef void		GLvoid
ctypedef signed char	GLbyte
ctypedef short		GLshort
ctypedef int		GLint
ctypedef unsigned char	GLubyte
ctypedef unsigned short	GLushort
ctypedef unsigned int	GLuint
ctypedef int		GLsizei
ctypedef float		GLfloat
ctypedef float		GLclampf
ctypedef double		GLdouble
ctypedef double		GLclampd


# from /usr/include/GL/glext.h:

ctypedef char GLchar
ctypedef ptrdiff_t GLintptr
ctypedef ptrdiff_t GLsizeiptr


cdef extern from *:
    ctypedef GLubyte* const_GLubyte_ptr "const GLubyte*"
    ctypedef GLboolean* const_GLboolean_ptr "const GLboolean*"
    ctypedef GLvoid* const_GLvoid_ptr "const GLvoid*"
    ctypedef GLchar* const_GLchar_ptr "const GLchar*"
    ctypedef GLfloat* const_GLfloat_ptr "const GLfloat*"
    ctypedef GLint* const_GLint_ptr "const GLint*"
    ctypedef GLshort* const_GLshort_ptr "const GLshort*"
    ctypedef GLbyte* const_GLbyte_ptr "const GLbyte*"
    ctypedef GLuint* const_GLuint_ptr "const GLuint*"
    ctypedef GLushort* const_GLushort_ptr "const GLushort*"
    ctypedef GLclampf* const_GLclampf_ptr "const GLclampf*"
    ctypedef GLsizei* const_GLsizei_ptr "const GLsizei*"
    ctypedef GLenum* const_GLenum_ptr "const GLenum*"
    ctypedef void* const_void_ptr "const void*"


# from /usr/include/GL/gl.h with:

cdef extern from 'GL/gl.h':
    cdef void  glClearIndex( GLfloat c )
    cdef void  glClearColor( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha )
    cdef void  glClear( GLbitfield mask )
    cdef void  glIndexMask( GLuint mask )
    cdef void  glColorMask( GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha )
    cdef void  glAlphaFunc( GLenum func, GLclampf ref )
    cdef void  glBlendFunc( GLenum sfactor, GLenum dfactor )
    cdef void  glLogicOp( GLenum opcode )
    cdef void  glCullFace( GLenum mode )
    cdef void  glFrontFace( GLenum mode )
    cdef void  glPointSize( GLfloat size )
    cdef void  glLineWidth( GLfloat width )
    cdef void  glLineStipple( GLint factor, GLushort pattern )
    cdef void  glPolygonMode( GLenum face, GLenum mode )
    cdef void  glPolygonOffset( GLfloat factor, GLfloat units )
    cdef void  glPolygonStipple( GLubyte *mask )
    cdef void  glGetPolygonStipple( GLubyte *mask )
    cdef void  glEdgeFlag( GLboolean flag )
    cdef void  glEdgeFlagv( GLboolean *flag )
    cdef void  glScissor( GLint x, GLint y, GLsizei width, GLsizei height)
    cdef void  glClipPlane( GLenum plane, GLdouble *equation )
    cdef void  glGetClipPlane( GLenum plane, GLdouble *equation )
    cdef void  glDrawBuffer( GLenum mode )
    cdef void  glReadBuffer( GLenum mode )
    cdef void  glEnable( GLenum cap )
    cdef void  glDisable( GLenum cap )
    cdef GLboolean  glIsEnabled( GLenum cap )
    cdef void  glEnableClientState( GLenum cap )  #/* 1.1 */
    cdef void  glDisableClientState( GLenum cap )  #/* 1.1 */
    cdef void  glGetBooleanv( GLenum pname, GLboolean *params )
    cdef void  glGetDoublev( GLenum pname, GLdouble *params )
    cdef void  glGetFloatv( GLenum pname, GLfloat *params )
    cdef void  glGetIntegerv( GLenum pname, GLint *params )
    cdef void  glPushAttrib( GLbitfield mask )
    cdef void  glPopAttrib()
    cdef void  glPushClientAttrib( GLbitfield mask )  #/* 1.1 */
    cdef void  glPopClientAttrib()  #/* 1.1 */
    cdef GLint  glRenderMode( GLenum mode )
    cdef GLenum  glGetError()
    cdef GLubyte *  glGetString( GLenum name )
    cdef void  glFinish()
    cdef void  glFlush()
    cdef void  glHint( GLenum target, GLenum mode )
    cdef void  glClearDepth( GLclampd depth )
    cdef void  glDepthFunc( GLenum func )
    cdef void  glDepthMask( GLboolean flag )
    cdef void  glDepthRange( GLclampd near_val, GLclampd far_val )
    cdef void  glClearAccum( GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha )
    cdef void  glAccum( GLenum op, GLfloat value )
    cdef void  glMatrixMode( GLenum mode )
    cdef void  glOrtho( GLdouble left, GLdouble right,
                                 GLdouble bottom, GLdouble top,
                                 GLdouble near_val, GLdouble far_val )
    cdef void  glFrustum( GLdouble left, GLdouble right,
                                   GLdouble bottom, GLdouble top,
                                   GLdouble near_val, GLdouble far_val )
    cdef void  glViewport( GLint x, GLint y,
                                    GLsizei width, GLsizei height )
    cdef void  glPushMatrix()
    cdef void  glPopMatrix()
    cdef void  glLoadIdentity()
    cdef void  glLoadMatrixd( GLdouble *m )
    cdef void  glLoadMatrixf( GLfloat *m )
    cdef void  glMultMatrixd( GLdouble *m )
    cdef void  glMultMatrixf( GLfloat *m )
    cdef void  glRotated( GLdouble angle,
                                   GLdouble x, GLdouble y, GLdouble z )
    cdef void  glRotatef( GLfloat angle,
                                   GLfloat x, GLfloat y, GLfloat z )
    cdef void  glScaled( GLdouble x, GLdouble y, GLdouble z )
    cdef void  glScalef( GLfloat x, GLfloat y, GLfloat z )
    cdef void  glTranslated( GLdouble x, GLdouble y, GLdouble z )
    cdef void  glTranslatef( GLfloat x, GLfloat y, GLfloat z )
    cdef GLboolean  glIsList( GLuint list )
    cdef void  glDeleteLists( GLuint list, GLsizei range )
    cdef GLuint  glGenLists( GLsizei range )
    cdef void  glNewList( GLuint list, GLenum mode )
    cdef void  glEndList()
    cdef void  glCallList( GLuint list )
    cdef void  glCallLists( GLsizei n, GLenum type,
                                     GLvoid *lists )
    cdef void  glListBase( GLuint base )
    cdef void  glBegin( GLenum mode )
    cdef void  glEnd()
    cdef void  glVertex2d( GLdouble x, GLdouble y )
    cdef void  glVertex2f( GLfloat x, GLfloat y )
    cdef void  glVertex2i( GLint x, GLint y )
    cdef void  glVertex2s( GLshort x, GLshort y )
    cdef void  glVertex3d( GLdouble x, GLdouble y, GLdouble z )
    cdef void  glVertex3f( GLfloat x, GLfloat y, GLfloat z )
    cdef void  glVertex3i( GLint x, GLint y, GLint z )
    cdef void  glVertex3s( GLshort x, GLshort y, GLshort z )
    cdef void  glVertex4d( GLdouble x, GLdouble y, GLdouble z, GLdouble w )
    cdef void  glVertex4f( GLfloat x, GLfloat y, GLfloat z, GLfloat w )
    cdef void  glVertex4i( GLint x, GLint y, GLint z, GLint w )
    cdef void  glVertex4s( GLshort x, GLshort y, GLshort z, GLshort w )
    cdef void  glVertex2dv( GLdouble *v )
    cdef void  glVertex2fv( GLfloat *v )
    cdef void  glVertex2iv( GLint *v )
    cdef void  glVertex2sv( GLshort *v )
    cdef void  glVertex3dv( GLdouble *v )
    cdef void  glVertex3fv( GLfloat *v )
    cdef void  glVertex3iv( GLint *v )
    cdef void  glVertex3sv( GLshort *v )
    cdef void  glVertex4dv( GLdouble *v )
    cdef void  glVertex4fv( GLfloat *v )
    cdef void  glVertex4iv( GLint *v )
    cdef void  glVertex4sv( GLshort *v )
    cdef void  glNormal3b( GLbyte nx, GLbyte ny, GLbyte nz )
    cdef void  glNormal3d( GLdouble nx, GLdouble ny, GLdouble nz )
    cdef void  glNormal3f( GLfloat nx, GLfloat ny, GLfloat nz )
    cdef void  glNormal3i( GLint nx, GLint ny, GLint nz )
    cdef void  glNormal3s( GLshort nx, GLshort ny, GLshort nz )
    cdef void  glNormal3bv( GLbyte *v )
    cdef void  glNormal3dv( GLdouble *v )
    cdef void  glNormal3fv( GLfloat *v )
    cdef void  glNormal3iv( GLint *v )
    cdef void  glNormal3sv( GLshort *v )
    cdef void  glIndexd( GLdouble c )
    cdef void  glIndexf( GLfloat c )
    cdef void  glIndexi( GLint c )
    cdef void  glIndexs( GLshort c )
    cdef void  glIndexub( GLubyte c )  #/* 1.1 */
    cdef void  glIndexdv( GLdouble *c )
    cdef void  glIndexfv( GLfloat *c )
    cdef void  glIndexiv( GLint *c )
    cdef void  glIndexsv( GLshort *c )
    cdef void  glIndexubv( GLubyte *c )  #/* 1.1 */
    cdef void  glColor3b( GLbyte red, GLbyte green, GLbyte blue )
    cdef void  glColor3d( GLdouble red, GLdouble green, GLdouble blue )
    cdef void  glColor3f( GLfloat red, GLfloat green, GLfloat blue )
    cdef void  glColor3i( GLint red, GLint green, GLint blue )
    cdef void  glColor3s( GLshort red, GLshort green, GLshort blue )
    cdef void  glColor3ub( GLubyte red, GLubyte green, GLubyte blue )
    cdef void  glColor3ui( GLuint red, GLuint green, GLuint blue )
    cdef void  glColor3us( GLushort red, GLushort green, GLushort blue )
    cdef void  glColor4b( GLbyte red, GLbyte green,
                                   GLbyte blue, GLbyte alpha )
    cdef void  glColor4d( GLdouble red, GLdouble green,
                                   GLdouble blue, GLdouble alpha )
    cdef void  glColor4f( GLfloat red, GLfloat green,
                                   GLfloat blue, GLfloat alpha )
    cdef void  glColor4i( GLint red, GLint green,
                                   GLint blue, GLint alpha )
    cdef void  glColor4s( GLshort red, GLshort green,
                                   GLshort blue, GLshort alpha )
    cdef void  glColor4ub( GLubyte red, GLubyte green,
                                    GLubyte blue, GLubyte alpha )
    cdef void  glColor4ui( GLuint red, GLuint green,
                                    GLuint blue, GLuint alpha )
    cdef void  glColor4us( GLushort red, GLushort green,
                                    GLushort blue, GLushort alpha )
    cdef void  glColor3bv( GLbyte *v )
    cdef void  glColor3dv( GLdouble *v )
    cdef void  glColor3fv( GLfloat *v )
    cdef void  glColor3iv( GLint *v )
    cdef void  glColor3sv( GLshort *v )
    cdef void  glColor3ubv( GLubyte *v )
    cdef void  glColor3uiv( GLuint *v )
    cdef void  glColor3usv( GLushort *v )
    cdef void  glColor4bv( GLbyte *v )
    cdef void  glColor4dv( GLdouble *v )
    cdef void  glColor4fv( GLfloat *v )
    cdef void  glColor4iv( GLint *v )
    cdef void  glColor4sv( GLshort *v )
    cdef void  glColor4ubv( GLubyte *v )
    cdef void  glColor4uiv( GLuint *v )
    cdef void  glColor4usv( GLushort *v )
    cdef void  glTexCoord1d( GLdouble s )
    cdef void  glTexCoord1f( GLfloat s )
    cdef void  glTexCoord1i( GLint s )
    cdef void  glTexCoord1s( GLshort s )
    cdef void  glTexCoord2d( GLdouble s, GLdouble t )
    cdef void  glTexCoord2f( GLfloat s, GLfloat t )
    cdef void  glTexCoord2i( GLint s, GLint t )
    cdef void  glTexCoord2s( GLshort s, GLshort t )
    cdef void  glTexCoord3d( GLdouble s, GLdouble t, GLdouble r )
    cdef void  glTexCoord3f( GLfloat s, GLfloat t, GLfloat r )
    cdef void  glTexCoord3i( GLint s, GLint t, GLint r )
    cdef void  glTexCoord3s( GLshort s, GLshort t, GLshort r )
    cdef void  glTexCoord4d( GLdouble s, GLdouble t, GLdouble r, GLdouble q )
    cdef void  glTexCoord4f( GLfloat s, GLfloat t, GLfloat r, GLfloat q )
    cdef void  glTexCoord4i( GLint s, GLint t, GLint r, GLint q )
    cdef void  glTexCoord4s( GLshort s, GLshort t, GLshort r, GLshort q )
    cdef void  glTexCoord1dv( GLdouble *v )
    cdef void  glTexCoord1fv( GLfloat *v )
    cdef void  glTexCoord1iv( GLint *v )
    cdef void  glTexCoord1sv( GLshort *v )
    cdef void  glTexCoord2dv( GLdouble *v )
    cdef void  glTexCoord2fv( GLfloat *v )
    cdef void  glTexCoord2iv( GLint *v )
    cdef void  glTexCoord2sv( GLshort *v )
    cdef void  glTexCoord3dv( GLdouble *v )
    cdef void  glTexCoord3fv( GLfloat *v )
    cdef void  glTexCoord3iv( GLint *v )
    cdef void  glTexCoord3sv( GLshort *v )
    cdef void  glTexCoord4dv( GLdouble *v )
    cdef void  glTexCoord4fv( GLfloat *v )
    cdef void  glTexCoord4iv( GLint *v )
    cdef void  glTexCoord4sv( GLshort *v )
    cdef void  glRasterPos2d( GLdouble x, GLdouble y )
    cdef void  glRasterPos2f( GLfloat x, GLfloat y )
    cdef void  glRasterPos2i( GLint x, GLint y )
    cdef void  glRasterPos2s( GLshort x, GLshort y )
    cdef void  glRasterPos3d( GLdouble x, GLdouble y, GLdouble z )
    cdef void  glRasterPos3f( GLfloat x, GLfloat y, GLfloat z )
    cdef void  glRasterPos3i( GLint x, GLint y, GLint z )
    cdef void  glRasterPos3s( GLshort x, GLshort y, GLshort z )
    cdef void  glRasterPos4d( GLdouble x, GLdouble y, GLdouble z, GLdouble w )
    cdef void  glRasterPos4f( GLfloat x, GLfloat y, GLfloat z, GLfloat w )
    cdef void  glRasterPos4i( GLint x, GLint y, GLint z, GLint w )
    cdef void  glRasterPos4s( GLshort x, GLshort y, GLshort z, GLshort w )
    cdef void  glRasterPos2dv( GLdouble *v )
    cdef void  glRasterPos2fv( GLfloat *v )
    cdef void  glRasterPos2iv( GLint *v )
    cdef void  glRasterPos2sv( GLshort *v )
    cdef void  glRasterPos3dv( GLdouble *v )
    cdef void  glRasterPos3fv( GLfloat *v )
    cdef void  glRasterPos3iv( GLint *v )
    cdef void  glRasterPos3sv( GLshort *v )
    cdef void  glRasterPos4dv( GLdouble *v )
    cdef void  glRasterPos4fv( GLfloat *v )
    cdef void  glRasterPos4iv( GLint *v )
    cdef void  glRasterPos4sv( GLshort *v )
    cdef void  glRectd( GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2 )
    cdef void  glRectf( GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2 )
    cdef void  glRecti( GLint x1, GLint y1, GLint x2, GLint y2 )
    cdef void  glRects( GLshort x1, GLshort y1, GLshort x2, GLshort y2 )
    cdef void  glRectdv( GLdouble *v1, GLdouble *v2 )
    cdef void  glRectfv( GLfloat *v1, GLfloat *v2 )
    cdef void  glRectiv( GLint *v1, GLint *v2 )
    cdef void  glRectsv( GLshort *v1, GLshort *v2 )
    cdef void  glVertexPointer( GLint size, GLenum type,
                                       GLsizei stride, GLvoid *ptr )
    cdef void  glNormalPointer( GLenum type, GLsizei stride,
                                       GLvoid *ptr )
    cdef void  glColorPointer( GLint size, GLenum type,
                                      GLsizei stride, GLvoid *ptr )
    cdef void  glIndexPointer( GLenum type, GLsizei stride,
                                      GLvoid *ptr )
    cdef void  glTexCoordPointer( GLint size, GLenum type,
                                         GLsizei stride, GLvoid *ptr )
    cdef void  glEdgeFlagPointer( GLsizei stride, GLvoid *ptr )
    cdef void  glGetPointerv( GLenum pname, GLvoid **params )
    cdef void  glArrayElement( GLint i )
    cdef void  glDrawArrays( GLenum mode, GLint first, GLsizei count )
    cdef void  glDrawElements( GLenum mode, GLsizei count,
                                      GLenum type, GLvoid *indices )
    cdef void  glInterleavedArrays( GLenum format, GLsizei stride,
                                           GLvoid *pointer )
    cdef void  glShadeModel( GLenum mode )
    cdef void  glLightf( GLenum light, GLenum pname, GLfloat param )
    cdef void  glLighti( GLenum light, GLenum pname, GLint param )
    cdef void  glLightfv( GLenum light, GLenum pname,
                                 GLfloat *params )
    cdef void  glLightiv( GLenum light, GLenum pname,
                                 GLint *params )
    cdef void  glGetLightfv( GLenum light, GLenum pname,
                                    GLfloat *params )
    cdef void  glGetLightiv( GLenum light, GLenum pname,
                                    GLint *params )
    cdef void  glLightModelf( GLenum pname, GLfloat param )
    cdef void  glLightModeli( GLenum pname, GLint param )
    cdef void  glLightModelfv( GLenum pname, GLfloat *params )
    cdef void  glLightModeliv( GLenum pname, GLint *params )
    cdef void  glMaterialf( GLenum face, GLenum pname, GLfloat param )
    cdef void  glMateriali( GLenum face, GLenum pname, GLint param )
    cdef void  glMaterialfv( GLenum face, GLenum pname, GLfloat *params )
    cdef void  glMaterialiv( GLenum face, GLenum pname, GLint *params )
    cdef void  glGetMaterialfv( GLenum face, GLenum pname, GLfloat *params )
    cdef void  glGetMaterialiv( GLenum face, GLenum pname, GLint *params )
    cdef void  glColorMaterial( GLenum face, GLenum mode )
    cdef void  glPixelZoom( GLfloat xfactor, GLfloat yfactor )
    cdef void  glPixelStoref( GLenum pname, GLfloat param )
    cdef void  glPixelStorei( GLenum pname, GLint param )
    cdef void  glPixelTransferf( GLenum pname, GLfloat param )
    cdef void  glPixelTransferi( GLenum pname, GLint param )
    cdef void  glPixelMapfv( GLenum map, GLsizei mapsize,
                                    GLfloat *values )
    cdef void  glPixelMapuiv( GLenum map, GLsizei mapsize,
                                     GLuint *values )
    cdef void  glPixelMapusv( GLenum map, GLsizei mapsize,
                                     GLushort *values )
    cdef void  glGetPixelMapfv( GLenum map, GLfloat *values )
    cdef void  glGetPixelMapuiv( GLenum map, GLuint *values )
    cdef void  glGetPixelMapusv( GLenum map, GLushort *values )
    cdef void  glBitmap( GLsizei width, GLsizei height,
                                GLfloat xorig, GLfloat yorig,
                                GLfloat xmove, GLfloat ymove,
                                GLubyte *bitmap )
    cdef void  glReadPixels( GLint x, GLint y,
                                    GLsizei width, GLsizei height,
                                    GLenum format, GLenum type,
                                    GLvoid *pixels )
    cdef void  glDrawPixels( GLsizei width, GLsizei height,
                                    GLenum format, GLenum type,
                                    GLvoid *pixels )
    cdef void  glCopyPixels( GLint x, GLint y,
                                    GLsizei width, GLsizei height,
                                    GLenum type )
    cdef void  glStencilFunc( GLenum func, GLint ref, GLuint mask )
    cdef void  glStencilMask( GLuint mask )
    cdef void  glStencilOp( GLenum fail, GLenum zfail, GLenum zpass )
    cdef void  glClearStencil( GLint s )
    cdef void  glTexGend( GLenum coord, GLenum pname, GLdouble param )
    cdef void  glTexGenf( GLenum coord, GLenum pname, GLfloat param )
    cdef void  glTexGeni( GLenum coord, GLenum pname, GLint param )
    cdef void  glTexGendv( GLenum coord, GLenum pname, GLdouble *params )
    cdef void  glTexGenfv( GLenum coord, GLenum pname, GLfloat *params )
    cdef void  glTexGeniv( GLenum coord, GLenum pname, GLint *params )
    cdef void  glGetTexGendv( GLenum coord, GLenum pname, GLdouble *params )
    cdef void  glGetTexGenfv( GLenum coord, GLenum pname, GLfloat *params )
    cdef void  glGetTexGeniv( GLenum coord, GLenum pname, GLint *params )
    cdef void  glTexEnvf( GLenum target, GLenum pname, GLfloat param )
    cdef void  glTexEnvi( GLenum target, GLenum pname, GLint param )
    cdef void  glTexEnvfv( GLenum target, GLenum pname, GLfloat *params )
    cdef void  glTexEnviv( GLenum target, GLenum pname, GLint *params )
    cdef void  glGetTexEnvfv( GLenum target, GLenum pname, GLfloat *params )
    cdef void  glGetTexEnviv( GLenum target, GLenum pname, GLint *params )
    cdef void  glTexParameterf( GLenum target, GLenum pname, GLfloat param )
    cdef void  glTexParameteri( GLenum target, GLenum pname, GLint param )
    cdef void  glTexParameterfv( GLenum target, GLenum pname,
                                          GLfloat *params )
    cdef void  glTexParameteriv( GLenum target, GLenum pname,
                                          GLint *params )
    cdef void  glGetTexParameterfv( GLenum target,
                                           GLenum pname, GLfloat *params)
    cdef void  glGetTexParameteriv( GLenum target,
                                           GLenum pname, GLint *params )
    cdef void  glGetTexLevelParameterfv( GLenum target, GLint level,
                                                GLenum pname, GLfloat *params )
    cdef void  glGetTexLevelParameteriv( GLenum target, GLint level,
                                                GLenum pname, GLint *params )
    cdef void  glTexImage1D( GLenum target, GLint level,
                                    GLint internalFormat,
                                    GLsizei width, GLint border,
                                    GLenum format, GLenum type,
                                    GLvoid *pixels )
    cdef void  glTexImage2D( GLenum target, GLint level,
                                    GLint internalFormat,
                                    GLsizei width, GLsizei height,
                                    GLint border, GLenum format, GLenum type,
                                    GLvoid *pixels )
    cdef void  glGetTexImage( GLenum target, GLint level,
                                     GLenum format, GLenum type,
                                     GLvoid *pixels )
    cdef void  glGenTextures( GLsizei n, GLuint *textures )
    cdef void  glDeleteTextures( GLsizei n, GLuint *textures)
    cdef void  glBindTexture( GLenum target, GLuint texture )
    cdef void  glPrioritizeTextures( GLsizei n,
                                            GLuint *textures,
                                            GLclampf *priorities )
    cdef GLboolean  glAreTexturesResident( GLsizei n,
                                                  GLuint *textures,
                                                  GLboolean *residences )
    cdef GLboolean  glIsTexture( GLuint texture )
    cdef void  glTexSubImage1D( GLenum target, GLint level,
                                       GLint xoffset,
                                       GLsizei width, GLenum format,
                                       GLenum type, GLvoid *pixels )
    cdef void  glTexSubImage2D( GLenum target, GLint level,
                                       GLint xoffset, GLint yoffset,
                                       GLsizei width, GLsizei height,
                                       GLenum format, GLenum type,
                                       GLvoid *pixels )
    cdef void  glCopyTexImage1D( GLenum target, GLint level,
                                        GLenum internalformat,
                                        GLint x, GLint y,
                                        GLsizei width, GLint border )
    cdef void  glCopyTexImage2D( GLenum target, GLint level,
                                        GLenum internalformat,
                                        GLint x, GLint y,
                                        GLsizei width, GLsizei height,
                                        GLint border )
    cdef void  glCopyTexSubImage1D( GLenum target, GLint level,
                                           GLint xoffset, GLint x, GLint y,
                                           GLsizei width )
    cdef void  glCopyTexSubImage2D( GLenum target, GLint level,
                                           GLint xoffset, GLint yoffset,
                                           GLint x, GLint y,
                                           GLsizei width, GLsizei height )
    cdef void  glMap1d( GLenum target, GLdouble u1, GLdouble u2,
                               GLint stride,
                               GLint order, GLdouble *points )
    cdef void  glMap1f( GLenum target, GLfloat u1, GLfloat u2,
                               GLint stride,
                               GLint order, GLfloat *points )
    cdef void  glMap2d( GLenum target,
		     GLdouble u1, GLdouble u2, GLint ustride, GLint uorder,
		     GLdouble v1, GLdouble v2, GLint vstride, GLint vorder,
		     GLdouble *points )
    cdef void  glMap2f( GLenum target,
		     GLfloat u1, GLfloat u2, GLint ustride, GLint uorder,
		     GLfloat v1, GLfloat v2, GLint vstride, GLint vorder,
		     GLfloat *points )
    cdef void  glGetMapdv( GLenum target, GLenum query, GLdouble *v )
    cdef void  glGetMapfv( GLenum target, GLenum query, GLfloat *v )
    cdef void  glGetMapiv( GLenum target, GLenum query, GLint *v )
    cdef void  glEvalCoord1d( GLdouble u )
    cdef void  glEvalCoord1f( GLfloat u )
    cdef void  glEvalCoord1dv( GLdouble *u )
    cdef void  glEvalCoord1fv( GLfloat *u )
    cdef void  glEvalCoord2d( GLdouble u, GLdouble v )
    cdef void  glEvalCoord2f( GLfloat u, GLfloat v )
    cdef void  glEvalCoord2dv( GLdouble *u )
    cdef void  glEvalCoord2fv( GLfloat *u )
    cdef void  glMapGrid1d( GLint un, GLdouble u1, GLdouble u2 )
    cdef void  glMapGrid1f( GLint un, GLfloat u1, GLfloat u2 )
    cdef void  glMapGrid2d( GLint un, GLdouble u1, GLdouble u2,
                                   GLint vn, GLdouble v1, GLdouble v2 )
    cdef void  glMapGrid2f( GLint un, GLfloat u1, GLfloat u2,
                                   GLint vn, GLfloat v1, GLfloat v2 )
    cdef void  glEvalPoint1( GLint i )
    cdef void  glEvalPoint2( GLint i, GLint j )
    cdef void  glEvalMesh1( GLenum mode, GLint i1, GLint i2 )
    cdef void  glEvalMesh2( GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2 )
    cdef void  glFogf( GLenum pname, GLfloat param )
    cdef void  glFogi( GLenum pname, GLint param )
    cdef void  glFogfv( GLenum pname, GLfloat *params )
    cdef void  glFogiv( GLenum pname, GLint *params )
    cdef void  glFeedbackBuffer( GLsizei size, GLenum type, GLfloat *buffer )
    cdef void  glPassThrough( GLfloat token )
    cdef void  glSelectBuffer( GLsizei size, GLuint *buffer )
    cdef void  glInitNames()
    cdef void  glLoadName( GLuint name )
    cdef void  glPushName( GLuint name )
    cdef void  glPopName()
    cdef void  glDrawRangeElements( GLenum mode, GLuint start,
	GLuint end, GLsizei count, GLenum type, GLvoid *indices )
    cdef void  glTexImage3D( GLenum target, GLint level,
                                      GLint internalFormat,
                                      GLsizei width, GLsizei height,
                                      GLsizei depth, GLint border,
                                      GLenum format, GLenum type,
                                      GLvoid *pixels )
    cdef void  glTexSubImage3D( GLenum target, GLint level,
                                         GLint xoffset, GLint yoffset,
                                         GLint zoffset, GLsizei width,
                                         GLsizei height, GLsizei depth,
                                         GLenum format,
                                         GLenum type, GLvoid *pixels)
    cdef void  glCopyTexSubImage3D( GLenum target, GLint level,
                                             GLint xoffset, GLint yoffset,
                                             GLint zoffset, GLint x,
                                             GLint y, GLsizei width,
                                             GLsizei height )
    cdef void  glColorTable( GLenum target, GLenum internalformat,
                                    GLsizei width, GLenum format,
                                    GLenum type, GLvoid *table )
    cdef void  glColorSubTable( GLenum target,
                                       GLsizei start, GLsizei count,
                                       GLenum format, GLenum type,
                                       GLvoid *data )
    cdef void  glColorTableParameteriv(GLenum target, GLenum pname,
                                              GLint *params)
    cdef void  glColorTableParameterfv(GLenum target, GLenum pname,
                                              GLfloat *params)
    cdef void  glCopyColorSubTable( GLenum target, GLsizei start,
                                           GLint x, GLint y, GLsizei width )
    cdef void  glCopyColorTable( GLenum target, GLenum internalformat,
                                        GLint x, GLint y, GLsizei width )
    cdef void  glGetColorTable( GLenum target, GLenum format,
                                       GLenum type, GLvoid *table )
    cdef void  glGetColorTableParameterfv( GLenum target, GLenum pname,
                                                  GLfloat *params )
    cdef void  glGetColorTableParameteriv( GLenum target, GLenum pname,
                                                  GLint *params )
    cdef void  glBlendEquation( GLenum mode )
    cdef void  glBlendColor( GLclampf red, GLclampf green,
                                    GLclampf blue, GLclampf alpha )
    cdef void  glHistogram( GLenum target, GLsizei width,
				   GLenum internalformat, GLboolean sink )
    cdef void  glResetHistogram( GLenum target )
    cdef void  glGetHistogram( GLenum target, GLboolean reset,
				      GLenum format, GLenum type,
				      GLvoid *values )
    cdef void  glGetHistogramParameterfv( GLenum target, GLenum pname,
						 GLfloat *params )
    cdef void  glGetHistogramParameteriv( GLenum target, GLenum pname,
						 GLint *params )
    cdef void  glMinmax( GLenum target, GLenum internalformat,
				GLboolean sink )
    cdef void  glResetMinmax( GLenum target )
    cdef void  glGetMinmax( GLenum target, GLboolean reset,
                                   GLenum format, GLenum types,
                                   GLvoid *values )
    cdef void  glGetMinmaxParameterfv( GLenum target, GLenum pname,
					      GLfloat *params )
    cdef void  glGetMinmaxParameteriv( GLenum target, GLenum pname,
					      GLint *params )
    cdef void  glConvolutionFilter1D( GLenum target,
	GLenum internalformat, GLsizei width, GLenum format, GLenum type,
	GLvoid *image )
    cdef void  glConvolutionFilter2D( GLenum target,
	GLenum internalformat, GLsizei width, GLsizei height, GLenum format,
	GLenum type, GLvoid *image )
    cdef void  glConvolutionParameterf( GLenum target, GLenum pname,
	GLfloat params )
    cdef void  glConvolutionParameterfv( GLenum target, GLenum pname,
	GLfloat *params )
    cdef void  glConvolutionParameteri( GLenum target, GLenum pname,
	GLint params )
    cdef void  glConvolutionParameteriv( GLenum target, GLenum pname,
	GLint *params )
    cdef void  glCopyConvolutionFilter1D( GLenum target,
	GLenum internalformat, GLint x, GLint y, GLsizei width )
    cdef void  glCopyConvolutionFilter2D( GLenum target,
	GLenum internalformat, GLint x, GLint y, GLsizei width,
	GLsizei height)
    cdef void  glGetConvolutionFilter( GLenum target, GLenum format,
	GLenum type, GLvoid *image )
    cdef void  glGetConvolutionParameterfv( GLenum target, GLenum pname,
	GLfloat *params )
    cdef void  glGetConvolutionParameteriv( GLenum target, GLenum pname,
	GLint *params )
    cdef void  glSeparableFilter2D( GLenum target,
	GLenum internalformat, GLsizei width, GLsizei height, GLenum format,
	GLenum type, GLvoid *row, GLvoid *column )
    cdef void  glGetSeparableFilter( GLenum target, GLenum format,
	GLenum type, GLvoid *row, GLvoid *column, GLvoid *span )
    cdef void  glActiveTexture( GLenum texture )
    cdef void  glClientActiveTexture( GLenum texture )
    cdef void  glCompressedTexImage1D( GLenum target, GLint level, GLenum internalformat, GLsizei width, GLint border, GLsizei imageSize, GLvoid *data )
    cdef void  glCompressedTexImage2D( GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLint border, GLsizei imageSize, GLvoid *data )
    cdef void  glCompressedTexImage3D( GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLint border, GLsizei imageSize, GLvoid *data )
    cdef void  glCompressedTexSubImage1D( GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLsizei imageSize, GLvoid *data )
    cdef void  glCompressedTexSubImage2D( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLsizei imageSize, GLvoid *data )
    cdef void  glCompressedTexSubImage3D( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLsizei width, GLsizei height, GLsizei depth, GLenum format, GLsizei imageSize, GLvoid *data )
    cdef void  glGetCompressedTexImage( GLenum target, GLint lod, GLvoid *img )
    cdef void  glMultiTexCoord1d( GLenum target, GLdouble s )
    cdef void  glMultiTexCoord1dv( GLenum target, GLdouble *v )
    cdef void  glMultiTexCoord1f( GLenum target, GLfloat s )
    cdef void  glMultiTexCoord1fv( GLenum target, GLfloat *v )
    cdef void  glMultiTexCoord1i( GLenum target, GLint s )
    cdef void  glMultiTexCoord1iv( GLenum target, GLint *v )
    cdef void  glMultiTexCoord1s( GLenum target, GLshort s )
    cdef void  glMultiTexCoord1sv( GLenum target, GLshort *v )
    cdef void  glMultiTexCoord2d( GLenum target, GLdouble s, GLdouble t )
    cdef void  glMultiTexCoord2dv( GLenum target, GLdouble *v )
    cdef void  glMultiTexCoord2f( GLenum target, GLfloat s, GLfloat t )
    cdef void  glMultiTexCoord2fv( GLenum target, GLfloat *v )
    cdef void  glMultiTexCoord2i( GLenum target, GLint s, GLint t )
    cdef void  glMultiTexCoord2iv( GLenum target, GLint *v )
    cdef void  glMultiTexCoord2s( GLenum target, GLshort s, GLshort t )
    cdef void  glMultiTexCoord2sv( GLenum target, GLshort *v )
    cdef void  glMultiTexCoord3d( GLenum target, GLdouble s, GLdouble t, GLdouble r )
    cdef void  glMultiTexCoord3dv( GLenum target, GLdouble *v )
    cdef void  glMultiTexCoord3f( GLenum target, GLfloat s, GLfloat t, GLfloat r )
    cdef void  glMultiTexCoord3fv( GLenum target, GLfloat *v )
    cdef void  glMultiTexCoord3i( GLenum target, GLint s, GLint t, GLint r )
    cdef void  glMultiTexCoord3iv( GLenum target, GLint *v )
    cdef void  glMultiTexCoord3s( GLenum target, GLshort s, GLshort t, GLshort r )
    cdef void  glMultiTexCoord3sv( GLenum target, GLshort *v )
    cdef void  glMultiTexCoord4d( GLenum target, GLdouble s, GLdouble t, GLdouble r, GLdouble q )
    cdef void  glMultiTexCoord4dv( GLenum target, GLdouble *v )
    cdef void  glMultiTexCoord4f( GLenum target, GLfloat s, GLfloat t, GLfloat r, GLfloat q )
    cdef void  glMultiTexCoord4fv( GLenum target, GLfloat *v )
    cdef void  glMultiTexCoord4i( GLenum target, GLint s, GLint t, GLint r, GLint q )
    cdef void  glMultiTexCoord4iv( GLenum target, GLint *v )
    cdef void  glMultiTexCoord4s( GLenum target, GLshort s, GLshort t, GLshort r, GLshort q )
    cdef void  glMultiTexCoord4sv( GLenum target, GLshort *v )
    cdef void  glLoadTransposeMatrixd( GLdouble m[16] )
    cdef void  glLoadTransposeMatrixf( GLfloat m[16] )
    cdef void  glMultTransposeMatrixd( GLdouble m[16] )
    cdef void  glMultTransposeMatrixf( GLfloat m[16] )
    cdef void  glSampleCoverage( GLclampf value, GLboolean invert )


# from /usr/include/GL/glext.h with:

cdef extern from 'GL/glext.h':
    cdef void  glBlendColor (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha)
    cdef void  glBlendEquation (GLenum mode)
    cdef void  glDrawRangeElements (GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, GLvoid *indices)
    cdef void  glTexImage3D (GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLsizei depth, GLint border, GLenum format, GLenum type, GLvoid *pixels)
    cdef void  glTexSubImage3D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLsizei width, GLsizei height, GLsizei depth, GLenum format, GLenum type, GLvoid *pixels)
    cdef void  glCopyTexSubImage3D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLint x, GLint y, GLsizei width, GLsizei height)
    cdef void  glColorTable (GLenum target, GLenum internalformat, GLsizei width, GLenum format, GLenum type, GLvoid *table)
    cdef void  glColorTableParameterfv (GLenum target, GLenum pname, GLfloat *params)
    cdef void  glColorTableParameteriv (GLenum target, GLenum pname, GLint *params)
    cdef void  glCopyColorTable (GLenum target, GLenum internalformat, GLint x, GLint y, GLsizei width)
    cdef void  glGetColorTable (GLenum target, GLenum format, GLenum type, GLvoid *table)
    cdef void  glGetColorTableParameterfv (GLenum target, GLenum pname, GLfloat *params)
    cdef void  glGetColorTableParameteriv (GLenum target, GLenum pname, GLint *params)
    cdef void  glColorSubTable (GLenum target, GLsizei start, GLsizei count, GLenum format, GLenum type, GLvoid *data)
    cdef void  glCopyColorSubTable (GLenum target, GLsizei start, GLint x, GLint y, GLsizei width)
    cdef void  glConvolutionFilter1D (GLenum target, GLenum internalformat, GLsizei width, GLenum format, GLenum type, GLvoid *image)
    cdef void  glConvolutionFilter2D (GLenum target, GLenum internalformat, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *image)
    cdef void  glConvolutionParameterf (GLenum target, GLenum pname, GLfloat params)
    cdef void  glConvolutionParameterfv (GLenum target, GLenum pname, GLfloat *params)
    cdef void  glConvolutionParameteri (GLenum target, GLenum pname, GLint params)
    cdef void  glConvolutionParameteriv (GLenum target, GLenum pname, GLint *params)
    cdef void  glCopyConvolutionFilter1D (GLenum target, GLenum internalformat, GLint x, GLint y, GLsizei width)
    cdef void  glCopyConvolutionFilter2D (GLenum target, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height)
    cdef void  glGetConvolutionFilter (GLenum target, GLenum format, GLenum type, GLvoid *image)
    cdef void  glGetConvolutionParameterfv (GLenum target, GLenum pname, GLfloat *params)
    cdef void  glGetConvolutionParameteriv (GLenum target, GLenum pname, GLint *params)
    cdef void  glGetSeparableFilter (GLenum target, GLenum format, GLenum type, GLvoid *row, GLvoid *column, GLvoid *span)
    cdef void  glSeparableFilter2D (GLenum target, GLenum internalformat, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *row, GLvoid *column)
    cdef void  glGetHistogram (GLenum target, GLboolean reset, GLenum format, GLenum type, GLvoid *values)
    cdef void  glGetHistogramParameterfv (GLenum target, GLenum pname, GLfloat *params)
    cdef void  glGetHistogramParameteriv (GLenum target, GLenum pname, GLint *params)
    cdef void  glGetMinmax (GLenum target, GLboolean reset, GLenum format, GLenum type, GLvoid *values)
    cdef void  glGetMinmaxParameterfv (GLenum target, GLenum pname, GLfloat *params)
    cdef void  glGetMinmaxParameteriv (GLenum target, GLenum pname, GLint *params)
    cdef void  glHistogram (GLenum target, GLsizei width, GLenum internalformat, GLboolean sink)
    cdef void  glMinmax (GLenum target, GLenum internalformat, GLboolean sink)
    cdef void  glResetHistogram (GLenum target)
    cdef void  glResetMinmax (GLenum target)
    cdef void  glActiveTexture (GLenum texture)
    cdef void  glSampleCoverage (GLfloat value, GLboolean invert)
    cdef void  glCompressedTexImage3D (GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLint border, GLsizei imageSize, GLvoid *data)
    cdef void  glCompressedTexImage2D (GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLint border, GLsizei imageSize, GLvoid *data)
    cdef void  glCompressedTexImage1D (GLenum target, GLint level, GLenum internalformat, GLsizei width, GLint border, GLsizei imageSize, GLvoid *data)
    cdef void  glCompressedTexSubImage3D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLsizei width, GLsizei height, GLsizei depth, GLenum format, GLsizei imageSize, GLvoid *data)
    cdef void  glCompressedTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLsizei imageSize, GLvoid *data)
    cdef void  glCompressedTexSubImage1D (GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLsizei imageSize, GLvoid *data)
    cdef void  glGetCompressedTexImage (GLenum target, GLint level, GLvoid *img)
    cdef void  glClientActiveTexture (GLenum texture)
    cdef void  glMultiTexCoord1d (GLenum target, GLdouble s)
    cdef void  glMultiTexCoord1dv (GLenum target, GLdouble *v)
    cdef void  glMultiTexCoord1f (GLenum target, GLfloat s)
    cdef void  glMultiTexCoord1fv (GLenum target, GLfloat *v)
    cdef void  glMultiTexCoord1i (GLenum target, GLint s)
    cdef void  glMultiTexCoord1iv (GLenum target, GLint *v)
    cdef void  glMultiTexCoord1s (GLenum target, GLshort s)
    cdef void  glMultiTexCoord1sv (GLenum target, GLshort *v)
    cdef void  glMultiTexCoord2d (GLenum target, GLdouble s, GLdouble t)
    cdef void  glMultiTexCoord2dv (GLenum target, GLdouble *v)
    cdef void  glMultiTexCoord2f (GLenum target, GLfloat s, GLfloat t)
    cdef void  glMultiTexCoord2fv (GLenum target, GLfloat *v)
    cdef void  glMultiTexCoord2i (GLenum target, GLint s, GLint t)
    cdef void  glMultiTexCoord2iv (GLenum target, GLint *v)
    cdef void  glMultiTexCoord2s (GLenum target, GLshort s, GLshort t)
    cdef void  glMultiTexCoord2sv (GLenum target, GLshort *v)
    cdef void  glMultiTexCoord3d (GLenum target, GLdouble s, GLdouble t, GLdouble r)
    cdef void  glMultiTexCoord3dv (GLenum target, GLdouble *v)
    cdef void  glMultiTexCoord3f (GLenum target, GLfloat s, GLfloat t, GLfloat r)
    cdef void  glMultiTexCoord3fv (GLenum target, GLfloat *v)
    cdef void  glMultiTexCoord3i (GLenum target, GLint s, GLint t, GLint r)
    cdef void  glMultiTexCoord3iv (GLenum target, GLint *v)
    cdef void  glMultiTexCoord3s (GLenum target, GLshort s, GLshort t, GLshort r)
    cdef void  glMultiTexCoord3sv (GLenum target, GLshort *v)
    cdef void  glMultiTexCoord4d (GLenum target, GLdouble s, GLdouble t, GLdouble r, GLdouble q)
    cdef void  glMultiTexCoord4dv (GLenum target, GLdouble *v)
    cdef void  glMultiTexCoord4f (GLenum target, GLfloat s, GLfloat t, GLfloat r, GLfloat q)
    cdef void  glMultiTexCoord4fv (GLenum target, GLfloat *v)
    cdef void  glMultiTexCoord4i (GLenum target, GLint s, GLint t, GLint r, GLint q)
    cdef void  glMultiTexCoord4iv (GLenum target, GLint *v)
    cdef void  glMultiTexCoord4s (GLenum target, GLshort s, GLshort t, GLshort r, GLshort q)
    cdef void  glMultiTexCoord4sv (GLenum target, GLshort *v)
    cdef void  glLoadTransposeMatrixf (GLfloat *m)
    cdef void  glLoadTransposeMatrixd (GLdouble *m)
    cdef void  glMultTransposeMatrixf (GLfloat *m)
    cdef void  glMultTransposeMatrixd (GLdouble *m)
    cdef void  glBlendFuncSeparate (GLenum sfactorRGB, GLenum dfactorRGB, GLenum sfactorAlpha, GLenum dfactorAlpha)
    cdef void  glMultiDrawArrays (GLenum mode, GLint *first, GLsizei *count, GLsizei drawcount)
    cdef void  glMultiDrawElements (GLenum mode, GLsizei *count, GLenum type, GLvoid* *indices, GLsizei drawcount)
    cdef void  glPointParameterf (GLenum pname, GLfloat param)
    cdef void  glPointParameterfv (GLenum pname, GLfloat *params)
    cdef void  glPointParameteri (GLenum pname, GLint param)
    cdef void  glPointParameteriv (GLenum pname, GLint *params)
    cdef void  glFogCoordf (GLfloat coord)
    cdef void  glFogCoordfv (GLfloat *coord)
    cdef void  glFogCoordd (GLdouble coord)
    cdef void  glFogCoorddv (GLdouble *coord)
    cdef void  glFogCoordPointer (GLenum type, GLsizei stride, GLvoid *pointer)
    cdef void  glSecondaryColor3b (GLbyte red, GLbyte green, GLbyte blue)
    cdef void  glSecondaryColor3bv (GLbyte *v)
    cdef void  glSecondaryColor3d (GLdouble red, GLdouble green, GLdouble blue)
    cdef void  glSecondaryColor3dv (GLdouble *v)
    cdef void  glSecondaryColor3f (GLfloat red, GLfloat green, GLfloat blue)
    cdef void  glSecondaryColor3fv (GLfloat *v)
    cdef void  glSecondaryColor3i (GLint red, GLint green, GLint blue)
    cdef void  glSecondaryColor3iv (GLint *v)
    cdef void  glSecondaryColor3s (GLshort red, GLshort green, GLshort blue)
    cdef void  glSecondaryColor3sv (GLshort *v)
    cdef void  glSecondaryColor3ub (GLubyte red, GLubyte green, GLubyte blue)
    cdef void  glSecondaryColor3ubv (GLubyte *v)
    cdef void  glSecondaryColor3ui (GLuint red, GLuint green, GLuint blue)
    cdef void  glSecondaryColor3uiv (GLuint *v)
    cdef void  glSecondaryColor3us (GLushort red, GLushort green, GLushort blue)
    cdef void  glSecondaryColor3usv (GLushort *v)
    cdef void  glSecondaryColorPointer (GLint size, GLenum type, GLsizei stride, GLvoid *pointer)
    cdef void  glWindowPos2d (GLdouble x, GLdouble y)
    cdef void  glWindowPos2dv (GLdouble *v)
    cdef void  glWindowPos2f (GLfloat x, GLfloat y)
    cdef void  glWindowPos2fv (GLfloat *v)
    cdef void  glWindowPos2i (GLint x, GLint y)
    cdef void  glWindowPos2iv (GLint *v)
    cdef void  glWindowPos2s (GLshort x, GLshort y)
    cdef void  glWindowPos2sv (GLshort *v)
    cdef void  glWindowPos3d (GLdouble x, GLdouble y, GLdouble z)
    cdef void  glWindowPos3dv (GLdouble *v)
    cdef void  glWindowPos3f (GLfloat x, GLfloat y, GLfloat z)
    cdef void  glWindowPos3fv (GLfloat *v)
    cdef void  glWindowPos3i (GLint x, GLint y, GLint z)
    cdef void  glWindowPos3iv (GLint *v)
    cdef void  glWindowPos3s (GLshort x, GLshort y, GLshort z)
    cdef void  glWindowPos3sv (GLshort *v)
    cdef void  glGenQueries (GLsizei n, GLuint *ids)
    cdef void  glDeleteQueries (GLsizei n, GLuint *ids)
    cdef GLboolean  glIsQuery (GLuint id)
    cdef void  glBeginQuery (GLenum target, GLuint id)
    cdef void  glEndQuery (GLenum target)
    cdef void  glGetQueryiv (GLenum target, GLenum pname, GLint *params)
    cdef void  glGetQueryObjectiv (GLuint id, GLenum pname, GLint *params)
    cdef void  glGetQueryObjectuiv (GLuint id, GLenum pname, GLuint *params)
    cdef void  glBindBuffer (GLenum target, GLuint buffer)
    cdef void  glDeleteBuffers (GLsizei n, GLuint *buffers)
    cdef void  glGenBuffers (GLsizei n, GLuint *buffers)
    cdef GLboolean  glIsBuffer (GLuint buffer)
    cdef void  glBufferData (GLenum target, GLsizeiptr size, GLvoid *data, GLenum usage)
    cdef void  glBufferSubData (GLenum target, GLintptr offset, GLsizeiptr size, GLvoid *data)
    cdef void  glGetBufferSubData (GLenum target, GLintptr offset, GLsizeiptr size, GLvoid *data)
    cdef GLvoid*  glMapBuffer (GLenum target, GLenum access)
    cdef GLboolean  glUnmapBuffer (GLenum target)
    cdef void  glGetBufferParameteriv (GLenum target, GLenum pname, GLint *params)
    cdef void  glGetBufferPointerv (GLenum target, GLenum pname, GLvoid* *params)
    cdef void  glBlendEquationSeparate (GLenum modeRGB, GLenum modeAlpha)
    cdef void  glDrawBuffers (GLsizei n, GLenum *bufs)
    cdef void  glStencilOpSeparate (GLenum face, GLenum sfail, GLenum dpfail, GLenum dppass)
    cdef void  glStencilFuncSeparate (GLenum face, GLenum func, GLint ref, GLuint mask)
    cdef void  glStencilMaskSeparate (GLenum face, GLuint mask)
    cdef void  glAttachShader (GLuint program, GLuint shader)
    cdef void  glBindAttribLocation (GLuint program, GLuint index, GLchar *name)
    cdef void  glCompileShader (GLuint shader)
    cdef GLuint  glCreateProgram ()
    cdef GLuint  glCreateShader (GLenum type)
    cdef void  glDeleteProgram (GLuint program)
    cdef void  glDeleteShader (GLuint shader)
    cdef void  glDetachShader (GLuint program, GLuint shader)
    cdef void  glDisableVertexAttribArray (GLuint index)
    cdef void  glEnableVertexAttribArray (GLuint index)
    cdef void  glGetActiveAttrib (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name)
    cdef void  glGetActiveUniform (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name)
    cdef void  glGetAttachedShaders (GLuint program, GLsizei maxCount, GLsizei *count, GLuint *obj)
    cdef GLint  glGetAttribLocation (GLuint program, GLchar *name)
    cdef void  glGetProgramiv (GLuint program, GLenum pname, GLint *params)
    cdef void  glGetProgramInfoLog (GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog)
    cdef void  glGetShaderiv (GLuint shader, GLenum pname, GLint *params)
    cdef void  glGetShaderInfoLog (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *infoLog)
    cdef void  glGetShaderSource (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *source)
    cdef GLint  glGetUniformLocation (GLuint program, GLchar *name)
    cdef void  glGetUniformfv (GLuint program, GLint location, GLfloat *params)
    cdef void  glGetUniformiv (GLuint program, GLint location, GLint *params)
    cdef void  glGetVertexAttribdv (GLuint index, GLenum pname, GLdouble *params)
    cdef void  glGetVertexAttribfv (GLuint index, GLenum pname, GLfloat *params)
    cdef void  glGetVertexAttribiv (GLuint index, GLenum pname, GLint *params)
    cdef void  glGetVertexAttribPointerv (GLuint index, GLenum pname, GLvoid* *pointer)
    cdef GLboolean  glIsProgram (GLuint program)
    cdef GLboolean  glIsShader (GLuint shader)
    cdef void  glLinkProgram (GLuint program)
    cdef void  glShaderSource (GLuint shader, GLsizei count, GLchar* *string, GLint *length)
    cdef void  glUseProgram (GLuint program)
    cdef void  glUniform1f (GLint location, GLfloat v0)
    cdef void  glUniform2f (GLint location, GLfloat v0, GLfloat v1)
    cdef void  glUniform3f (GLint location, GLfloat v0, GLfloat v1, GLfloat v2)
    cdef void  glUniform4f (GLint location, GLfloat v0, GLfloat v1, GLfloat v2, GLfloat v3)
    cdef void  glUniform1i (GLint location, GLint v0)
    cdef void  glUniform2i (GLint location, GLint v0, GLint v1)
    cdef void  glUniform3i (GLint location, GLint v0, GLint v1, GLint v2)
    cdef void  glUniform4i (GLint location, GLint v0, GLint v1, GLint v2, GLint v3)
    cdef void  glUniform1fv (GLint location, GLsizei count, GLfloat *value)
    cdef void  glUniform2fv (GLint location, GLsizei count, GLfloat *value)
    cdef void  glUniform3fv (GLint location, GLsizei count, GLfloat *value)
    cdef void  glUniform4fv (GLint location, GLsizei count, GLfloat *value)
    cdef void  glUniform1iv (GLint location, GLsizei count, GLint *value)
    cdef void  glUniform2iv (GLint location, GLsizei count, GLint *value)
    cdef void  glUniform3iv (GLint location, GLsizei count, GLint *value)
    cdef void  glUniform4iv (GLint location, GLsizei count, GLint *value)
    cdef void  glUniformMatrix2fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glUniformMatrix3fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glUniformMatrix4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glValidateProgram (GLuint program)
    cdef void  glVertexAttrib1d (GLuint index, GLdouble x)
    cdef void  glVertexAttrib1dv (GLuint index, GLdouble *v)
    cdef void  glVertexAttrib1f (GLuint index, GLfloat x)
    cdef void  glVertexAttrib1fv (GLuint index, GLfloat *v)
    cdef void  glVertexAttrib1s (GLuint index, GLshort x)
    cdef void  glVertexAttrib1sv (GLuint index, GLshort *v)
    cdef void  glVertexAttrib2d (GLuint index, GLdouble x, GLdouble y)
    cdef void  glVertexAttrib2dv (GLuint index, GLdouble *v)
    cdef void  glVertexAttrib2f (GLuint index, GLfloat x, GLfloat y)
    cdef void  glVertexAttrib2fv (GLuint index, GLfloat *v)
    cdef void  glVertexAttrib2s (GLuint index, GLshort x, GLshort y)
    cdef void  glVertexAttrib2sv (GLuint index, GLshort *v)
    cdef void  glVertexAttrib3d (GLuint index, GLdouble x, GLdouble y, GLdouble z)
    cdef void  glVertexAttrib3dv (GLuint index, GLdouble *v)
    cdef void  glVertexAttrib3f (GLuint index, GLfloat x, GLfloat y, GLfloat z)
    cdef void  glVertexAttrib3fv (GLuint index, GLfloat *v)
    cdef void  glVertexAttrib3s (GLuint index, GLshort x, GLshort y, GLshort z)
    cdef void  glVertexAttrib3sv (GLuint index, GLshort *v)
    cdef void  glVertexAttrib4Nbv (GLuint index, GLbyte *v)
    cdef void  glVertexAttrib4Niv (GLuint index, GLint *v)
    cdef void  glVertexAttrib4Nsv (GLuint index, GLshort *v)
    cdef void  glVertexAttrib4Nub (GLuint index, GLubyte x, GLubyte y, GLubyte z, GLubyte w)
    cdef void  glVertexAttrib4Nubv (GLuint index, GLubyte *v)
    cdef void  glVertexAttrib4Nuiv (GLuint index, GLuint *v)
    cdef void  glVertexAttrib4Nusv (GLuint index, GLushort *v)
    cdef void  glVertexAttrib4bv (GLuint index, GLbyte *v)
    cdef void  glVertexAttrib4d (GLuint index, GLdouble x, GLdouble y, GLdouble z, GLdouble w)
    cdef void  glVertexAttrib4dv (GLuint index, GLdouble *v)
    cdef void  glVertexAttrib4f (GLuint index, GLfloat x, GLfloat y, GLfloat z, GLfloat w)
    cdef void  glVertexAttrib4fv (GLuint index, GLfloat *v)
    cdef void  glVertexAttrib4iv (GLuint index, GLint *v)
    cdef void  glVertexAttrib4s (GLuint index, GLshort x, GLshort y, GLshort z, GLshort w)
    cdef void  glVertexAttrib4sv (GLuint index, GLshort *v)
    cdef void  glVertexAttrib4ubv (GLuint index, GLubyte *v)
    cdef void  glVertexAttrib4uiv (GLuint index, GLuint *v)
    cdef void  glVertexAttrib4usv (GLuint index, GLushort *v)
    cdef void  glVertexAttribPointer (GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, GLvoid *pointer)
    cdef void  glUniformMatrix2x3fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glUniformMatrix3x2fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glUniformMatrix2x4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glUniformMatrix4x2fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glUniformMatrix3x4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glUniformMatrix4x3fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glColorMaski (GLuint index, GLboolean r, GLboolean g, GLboolean b, GLboolean a)
    cdef void  glGetBooleani_v (GLenum target, GLuint index, GLboolean *data)
    cdef void  glGetIntegeri_v (GLenum target, GLuint index, GLint *data)
    cdef void  glEnablei (GLenum target, GLuint index)
    cdef void  glDisablei (GLenum target, GLuint index)
    cdef GLboolean  glIsEnabledi (GLenum target, GLuint index)
    cdef void  glBeginTransformFeedback (GLenum primitiveMode)
    cdef void  glEndTransformFeedback ()
    cdef void  glBindBufferRange (GLenum target, GLuint index, GLuint buffer, GLintptr offset, GLsizeiptr size)
    cdef void  glBindBufferBase (GLenum target, GLuint index, GLuint buffer)
    cdef void  glTransformFeedbackVaryings (GLuint program, GLsizei count, GLchar* *varyings, GLenum bufferMode)
    cdef void  glGetTransformFeedbackVarying (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLsizei *size, GLenum *type, GLchar *name)
    cdef void  glClampColor (GLenum target, GLenum clamp)
    cdef void  glBeginConditionalRender (GLuint id, GLenum mode)
    cdef void  glEndConditionalRender ()
    cdef void  glVertexAttribIPointer (GLuint index, GLint size, GLenum type, GLsizei stride, GLvoid *pointer)
    cdef void  glGetVertexAttribIiv (GLuint index, GLenum pname, GLint *params)
    cdef void  glGetVertexAttribIuiv (GLuint index, GLenum pname, GLuint *params)
    cdef void  glVertexAttribI1i (GLuint index, GLint x)
    cdef void  glVertexAttribI2i (GLuint index, GLint x, GLint y)
    cdef void  glVertexAttribI3i (GLuint index, GLint x, GLint y, GLint z)
    cdef void  glVertexAttribI4i (GLuint index, GLint x, GLint y, GLint z, GLint w)
    cdef void  glVertexAttribI1ui (GLuint index, GLuint x)
    cdef void  glVertexAttribI2ui (GLuint index, GLuint x, GLuint y)
    cdef void  glVertexAttribI3ui (GLuint index, GLuint x, GLuint y, GLuint z)
    cdef void  glVertexAttribI4ui (GLuint index, GLuint x, GLuint y, GLuint z, GLuint w)
    cdef void  glVertexAttribI1iv (GLuint index, GLint *v)
    cdef void  glVertexAttribI2iv (GLuint index, GLint *v)
    cdef void  glVertexAttribI3iv (GLuint index, GLint *v)
    cdef void  glVertexAttribI4iv (GLuint index, GLint *v)
    cdef void  glVertexAttribI1uiv (GLuint index, GLuint *v)
    cdef void  glVertexAttribI2uiv (GLuint index, GLuint *v)
    cdef void  glVertexAttribI3uiv (GLuint index, GLuint *v)
    cdef void  glVertexAttribI4uiv (GLuint index, GLuint *v)
    cdef void  glVertexAttribI4bv (GLuint index, GLbyte *v)
    cdef void  glVertexAttribI4sv (GLuint index, GLshort *v)
    cdef void  glVertexAttribI4ubv (GLuint index, GLubyte *v)
    cdef void  glVertexAttribI4usv (GLuint index, GLushort *v)
    cdef void  glGetUniformuiv (GLuint program, GLint location, GLuint *params)
    cdef void  glBindFragDataLocation (GLuint program, GLuint color, GLchar *name)
    cdef GLint  glGetFragDataLocation (GLuint program, GLchar *name)
    cdef void  glUniform1ui (GLint location, GLuint v0)
    cdef void  glUniform2ui (GLint location, GLuint v0, GLuint v1)
    cdef void  glUniform3ui (GLint location, GLuint v0, GLuint v1, GLuint v2)
    cdef void  glUniform4ui (GLint location, GLuint v0, GLuint v1, GLuint v2, GLuint v3)
    cdef void  glUniform1uiv (GLint location, GLsizei count, GLuint *value)
    cdef void  glUniform2uiv (GLint location, GLsizei count, GLuint *value)
    cdef void  glUniform3uiv (GLint location, GLsizei count, GLuint *value)
    cdef void  glUniform4uiv (GLint location, GLsizei count, GLuint *value)
    cdef void  glTexParameterIiv (GLenum target, GLenum pname, GLint *params)
    cdef void  glTexParameterIuiv (GLenum target, GLenum pname, GLuint *params)
    cdef void  glGetTexParameterIiv (GLenum target, GLenum pname, GLint *params)
    cdef void  glGetTexParameterIuiv (GLenum target, GLenum pname, GLuint *params)
    cdef void  glClearBufferiv (GLenum buffer, GLint drawbuffer, GLint *value)
    cdef void  glClearBufferuiv (GLenum buffer, GLint drawbuffer, GLuint *value)
    cdef void  glClearBufferfv (GLenum buffer, GLint drawbuffer, GLfloat *value)
    cdef void  glClearBufferfi (GLenum buffer, GLint drawbuffer, GLfloat depth, GLint stencil)
    cdef GLubyte *  glGetStringi (GLenum name, GLuint index)
    cdef void  glDrawArraysInstanced (GLenum mode, GLint first, GLsizei count, GLsizei instancecount)
    cdef void  glDrawElementsInstanced (GLenum mode, GLsizei count, GLenum type, GLvoid *indices, GLsizei instancecount)
    cdef void  glTexBuffer (GLenum target, GLenum internalformat, GLuint buffer)
    cdef void  glPrimitiveRestartIndex (GLuint index)
    cdef void  glFramebufferTexture (GLenum target, GLenum attachment, GLuint texture, GLint level)
    cdef void  glVertexAttribDivisor (GLuint index, GLuint divisor)
    cdef void  glMinSampleShading (GLfloat value)
    cdef void  glBlendEquationi (GLuint buf, GLenum mode)
    cdef void  glBlendEquationSeparatei (GLuint buf, GLenum modeRGB, GLenum modeAlpha)
    cdef void  glBlendFunci (GLuint buf, GLenum src, GLenum dst)
    cdef void  glBlendFuncSeparatei (GLuint buf, GLenum srcRGB, GLenum dstRGB, GLenum srcAlpha, GLenum dstAlpha)
    cdef GLboolean  glIsRenderbuffer (GLuint renderbuffer)
    cdef void  glBindRenderbuffer (GLenum target, GLuint renderbuffer)
    cdef void  glDeleteRenderbuffers (GLsizei n, GLuint *renderbuffers)
    cdef void  glGenRenderbuffers (GLsizei n, GLuint *renderbuffers)
    cdef void  glRenderbufferStorage (GLenum target, GLenum internalformat, GLsizei width, GLsizei height)
    cdef void  glGetRenderbufferParameteriv (GLenum target, GLenum pname, GLint *params)
    cdef GLboolean  glIsFramebuffer (GLuint framebuffer)
    cdef void  glBindFramebuffer (GLenum target, GLuint framebuffer)
    cdef void  glDeleteFramebuffers (GLsizei n, GLuint *framebuffers)
    cdef void  glGenFramebuffers (GLsizei n, GLuint *framebuffers)
    cdef GLenum  glCheckFramebufferStatus (GLenum target)
    cdef void  glFramebufferTexture1D (GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level)
    cdef void  glFramebufferTexture2D (GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level)
    cdef void  glFramebufferTexture3D (GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level, GLint zoffset)
    cdef void  glFramebufferRenderbuffer (GLenum target, GLenum attachment, GLenum renderbuffertarget, GLuint renderbuffer)
    cdef void  glGetFramebufferAttachmentParameteriv (GLenum target, GLenum attachment, GLenum pname, GLint *params)
    cdef void  glGenerateMipmap (GLenum target)
    cdef void  glBlitFramebuffer (GLint srcX0, GLint srcY0, GLint srcX1, GLint srcY1, GLint dstX0, GLint dstY0, GLint dstX1, GLint dstY1, GLbitfield mask, GLenum filter)
    cdef void  glRenderbufferStorageMultisample (GLenum target, GLsizei samples, GLenum internalformat, GLsizei width, GLsizei height)
    cdef void  glFramebufferTextureLayer (GLenum target, GLenum attachment, GLuint texture, GLint level, GLint layer)
    cdef GLvoid*  glMapBufferRange (GLenum target, GLintptr offset, GLsizeiptr length, GLbitfield access)
    cdef void  glFlushMappedBufferRange (GLenum target, GLintptr offset, GLsizeiptr length)
    cdef void  glBindVertexArray (GLuint array)
    cdef void  glDeleteVertexArrays (GLsizei n, GLuint *arrays)
    cdef void  glGenVertexArrays (GLsizei n, GLuint *arrays)
    cdef GLboolean  glIsVertexArray (GLuint array)
    cdef void  glGetUniformIndices (GLuint program, GLsizei uniformCount, GLchar* *uniformNames, GLuint *uniformIndices)
    cdef void  glGetActiveUniformsiv (GLuint program, GLsizei uniformCount, GLuint *uniformIndices, GLenum pname, GLint *params)
    cdef void  glGetActiveUniformName (GLuint program, GLuint uniformIndex, GLsizei bufSize, GLsizei *length, GLchar *uniformName)
    cdef GLuint  glGetUniformBlockIndex (GLuint program, GLchar *uniformBlockName)
    cdef void  glGetActiveUniformBlockiv (GLuint program, GLuint uniformBlockIndex, GLenum pname, GLint *params)
    cdef void  glGetActiveUniformBlockName (GLuint program, GLuint uniformBlockIndex, GLsizei bufSize, GLsizei *length, GLchar *uniformBlockName)
    cdef void  glUniformBlockBinding (GLuint program, GLuint uniformBlockIndex, GLuint uniformBlockBinding)
    cdef void  glCopyBufferSubData (GLenum readTarget, GLenum writeTarget, GLintptr readOffset, GLintptr writeOffset, GLsizeiptr size)
    cdef void  glDrawElementsBaseVertex (GLenum mode, GLsizei count, GLenum type, GLvoid *indices, GLint basevertex)
    cdef void  glDrawRangeElementsBaseVertex (GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, GLvoid *indices, GLint basevertex)
    cdef void  glDrawElementsInstancedBaseVertex (GLenum mode, GLsizei count, GLenum type, GLvoid *indices, GLsizei instancecount, GLint basevertex)
    cdef void  glMultiDrawElementsBaseVertex (GLenum mode, GLsizei *count, GLenum type, GLvoid* *indices, GLsizei drawcount, GLint *basevertex)
    cdef void  glProvokingVertex (GLenum mode)
    cdef void  glTexImage2DMultisample (GLenum target, GLsizei samples, GLint internalformat, GLsizei width, GLsizei height, GLboolean fixedsamplelocations)
    cdef void  glTexImage3DMultisample (GLenum target, GLsizei samples, GLint internalformat, GLsizei width, GLsizei height, GLsizei depth, GLboolean fixedsamplelocations)
    cdef void  glGetMultisamplefv (GLenum pname, GLuint index, GLfloat *val)
    cdef void  glSampleMaski (GLuint index, GLbitfield mask)
    cdef void  glBindFragDataLocationIndexed (GLuint program, GLuint colorNumber, GLuint index, GLchar *name)
    cdef GLint  glGetFragDataIndex (GLuint program, GLchar *name)
    cdef void  glGenSamplers (GLsizei count, GLuint *samplers)
    cdef void  glDeleteSamplers (GLsizei count, GLuint *samplers)
    cdef GLboolean  glIsSampler (GLuint sampler)
    cdef void  glBindSampler (GLuint unit, GLuint sampler)
    cdef void  glSamplerParameteri (GLuint sampler, GLenum pname, GLint param)
    cdef void  glSamplerParameteriv (GLuint sampler, GLenum pname, GLint *param)
    cdef void  glSamplerParameterf (GLuint sampler, GLenum pname, GLfloat param)
    cdef void  glSamplerParameterfv (GLuint sampler, GLenum pname, GLfloat *param)
    cdef void  glSamplerParameterIiv (GLuint sampler, GLenum pname, GLint *param)
    cdef void  glSamplerParameterIuiv (GLuint sampler, GLenum pname, GLuint *param)
    cdef void  glGetSamplerParameteriv (GLuint sampler, GLenum pname, GLint *params)
    cdef void  glGetSamplerParameterIiv (GLuint sampler, GLenum pname, GLint *params)
    cdef void  glGetSamplerParameterfv (GLuint sampler, GLenum pname, GLfloat *params)
    cdef void  glGetSamplerParameterIuiv (GLuint sampler, GLenum pname, GLuint *params)
    cdef void  glQueryCounter (GLuint id, GLenum target)
    cdef void  glVertexP2ui (GLenum type, GLuint value)
    cdef void  glVertexP2uiv (GLenum type, GLuint *value)
    cdef void  glVertexP3ui (GLenum type, GLuint value)
    cdef void  glVertexP3uiv (GLenum type, GLuint *value)
    cdef void  glVertexP4ui (GLenum type, GLuint value)
    cdef void  glVertexP4uiv (GLenum type, GLuint *value)
    cdef void  glTexCoordP1ui (GLenum type, GLuint coords)
    cdef void  glTexCoordP1uiv (GLenum type, GLuint *coords)
    cdef void  glTexCoordP2ui (GLenum type, GLuint coords)
    cdef void  glTexCoordP2uiv (GLenum type, GLuint *coords)
    cdef void  glTexCoordP3ui (GLenum type, GLuint coords)
    cdef void  glTexCoordP3uiv (GLenum type, GLuint *coords)
    cdef void  glTexCoordP4ui (GLenum type, GLuint coords)
    cdef void  glTexCoordP4uiv (GLenum type, GLuint *coords)
    cdef void  glMultiTexCoordP1ui (GLenum texture, GLenum type, GLuint coords)
    cdef void  glMultiTexCoordP1uiv (GLenum texture, GLenum type, GLuint *coords)
    cdef void  glMultiTexCoordP2ui (GLenum texture, GLenum type, GLuint coords)
    cdef void  glMultiTexCoordP2uiv (GLenum texture, GLenum type, GLuint *coords)
    cdef void  glMultiTexCoordP3ui (GLenum texture, GLenum type, GLuint coords)
    cdef void  glMultiTexCoordP3uiv (GLenum texture, GLenum type, GLuint *coords)
    cdef void  glMultiTexCoordP4ui (GLenum texture, GLenum type, GLuint coords)
    cdef void  glMultiTexCoordP4uiv (GLenum texture, GLenum type, GLuint *coords)
    cdef void  glNormalP3ui (GLenum type, GLuint coords)
    cdef void  glNormalP3uiv (GLenum type, GLuint *coords)
    cdef void  glColorP3ui (GLenum type, GLuint color)
    cdef void  glColorP3uiv (GLenum type, GLuint *color)
    cdef void  glColorP4ui (GLenum type, GLuint color)
    cdef void  glColorP4uiv (GLenum type, GLuint *color)
    cdef void  glSecondaryColorP3ui (GLenum type, GLuint color)
    cdef void  glSecondaryColorP3uiv (GLenum type, GLuint *color)
    cdef void  glVertexAttribP1ui (GLuint index, GLenum type, GLboolean normalized, GLuint value)
    cdef void  glVertexAttribP1uiv (GLuint index, GLenum type, GLboolean normalized, GLuint *value)
    cdef void  glVertexAttribP2ui (GLuint index, GLenum type, GLboolean normalized, GLuint value)
    cdef void  glVertexAttribP2uiv (GLuint index, GLenum type, GLboolean normalized, GLuint *value)
    cdef void  glVertexAttribP3ui (GLuint index, GLenum type, GLboolean normalized, GLuint value)
    cdef void  glVertexAttribP3uiv (GLuint index, GLenum type, GLboolean normalized, GLuint *value)
    cdef void  glVertexAttribP4ui (GLuint index, GLenum type, GLboolean normalized, GLuint value)
    cdef void  glVertexAttribP4uiv (GLuint index, GLenum type, GLboolean normalized, GLuint *value)
    cdef void  glDrawArraysIndirect (GLenum mode, GLvoid *indirect)
    cdef void  glDrawElementsIndirect (GLenum mode, GLenum type, GLvoid *indirect)
    cdef void  glUniform1d (GLint location, GLdouble x)
    cdef void  glUniform2d (GLint location, GLdouble x, GLdouble y)
    cdef void  glUniform3d (GLint location, GLdouble x, GLdouble y, GLdouble z)
    cdef void  glUniform4d (GLint location, GLdouble x, GLdouble y, GLdouble z, GLdouble w)
    cdef void  glUniform1dv (GLint location, GLsizei count, GLdouble *value)
    cdef void  glUniform2dv (GLint location, GLsizei count, GLdouble *value)
    cdef void  glUniform3dv (GLint location, GLsizei count, GLdouble *value)
    cdef void  glUniform4dv (GLint location, GLsizei count, GLdouble *value)
    cdef void  glUniformMatrix2dv (GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glUniformMatrix3dv (GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glUniformMatrix4dv (GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glUniformMatrix2x3dv (GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glUniformMatrix2x4dv (GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glUniformMatrix3x2dv (GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glUniformMatrix3x4dv (GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glUniformMatrix4x2dv (GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glUniformMatrix4x3dv (GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glGetUniformdv (GLuint program, GLint location, GLdouble *params)
    cdef GLint  glGetSubroutineUniformLocation (GLuint program, GLenum shadertype, GLchar *name)
    cdef GLuint  glGetSubroutineIndex (GLuint program, GLenum shadertype, GLchar *name)
    cdef void  glGetActiveSubroutineUniformiv (GLuint program, GLenum shadertype, GLuint index, GLenum pname, GLint *values)
    cdef void  glGetActiveSubroutineUniformName (GLuint program, GLenum shadertype, GLuint index, GLsizei bufsize, GLsizei *length, GLchar *name)
    cdef void  glGetActiveSubroutineName (GLuint program, GLenum shadertype, GLuint index, GLsizei bufsize, GLsizei *length, GLchar *name)
    cdef void  glUniformSubroutinesuiv (GLenum shadertype, GLsizei count, GLuint *indices)
    cdef void  glGetUniformSubroutineuiv (GLenum shadertype, GLint location, GLuint *params)
    cdef void  glGetProgramStageiv (GLuint program, GLenum shadertype, GLenum pname, GLint *values)
    cdef void  glPatchParameteri (GLenum pname, GLint value)
    cdef void  glPatchParameterfv (GLenum pname, GLfloat *values)
    cdef void  glBindTransformFeedback (GLenum target, GLuint id)
    cdef void  glDeleteTransformFeedbacks (GLsizei n, GLuint *ids)
    cdef void  glGenTransformFeedbacks (GLsizei n, GLuint *ids)
    cdef GLboolean  glIsTransformFeedback (GLuint id)
    cdef void  glPauseTransformFeedback ()
    cdef void  glResumeTransformFeedback ()
    cdef void  glDrawTransformFeedback (GLenum mode, GLuint id)
    cdef void  glDrawTransformFeedbackStream (GLenum mode, GLuint id, GLuint stream)
    cdef void  glBeginQueryIndexed (GLenum target, GLuint index, GLuint id)
    cdef void  glEndQueryIndexed (GLenum target, GLuint index)
    cdef void  glGetQueryIndexediv (GLenum target, GLuint index, GLenum pname, GLint *params)
    cdef void  glReleaseShaderCompiler ()
    cdef void  glShaderBinary (GLsizei count, GLuint *shaders, GLenum binaryformat, GLvoid *binary, GLsizei length)
    cdef void  glGetShaderPrecisionFormat (GLenum shadertype, GLenum precisiontype, GLint *range, GLint *precision)
    cdef void  glDepthRangef (GLfloat n, GLfloat f)
    cdef void  glClearDepthf (GLfloat d)
    cdef void  glGetProgramBinary (GLuint program, GLsizei bufSize, GLsizei *length, GLenum *binaryFormat, GLvoid *binary)
    cdef void  glProgramBinary (GLuint program, GLenum binaryFormat, GLvoid *binary, GLsizei length)
    cdef void  glProgramParameteri (GLuint program, GLenum pname, GLint value)
    cdef void  glUseProgramStages (GLuint pipeline, GLbitfield stages, GLuint program)
    cdef void  glActiveShaderProgram (GLuint pipeline, GLuint program)
    cdef GLuint  glCreateShaderProgramv (GLenum type, GLsizei count, GLchar* *strings)
    cdef void  glBindProgramPipeline (GLuint pipeline)
    cdef void  glDeleteProgramPipelines (GLsizei n, GLuint *pipelines)
    cdef void  glGenProgramPipelines (GLsizei n, GLuint *pipelines)
    cdef GLboolean  glIsProgramPipeline (GLuint pipeline)
    cdef void  glGetProgramPipelineiv (GLuint pipeline, GLenum pname, GLint *params)
    cdef void  glProgramUniform1i (GLuint program, GLint location, GLint v0)
    cdef void  glProgramUniform1iv (GLuint program, GLint location, GLsizei count, GLint *value)
    cdef void  glProgramUniform1f (GLuint program, GLint location, GLfloat v0)
    cdef void  glProgramUniform1fv (GLuint program, GLint location, GLsizei count, GLfloat *value)
    cdef void  glProgramUniform1d (GLuint program, GLint location, GLdouble v0)
    cdef void  glProgramUniform1dv (GLuint program, GLint location, GLsizei count, GLdouble *value)
    cdef void  glProgramUniform1ui (GLuint program, GLint location, GLuint v0)
    cdef void  glProgramUniform1uiv (GLuint program, GLint location, GLsizei count, GLuint *value)
    cdef void  glProgramUniform2i (GLuint program, GLint location, GLint v0, GLint v1)
    cdef void  glProgramUniform2iv (GLuint program, GLint location, GLsizei count, GLint *value)
    cdef void  glProgramUniform2f (GLuint program, GLint location, GLfloat v0, GLfloat v1)
    cdef void  glProgramUniform2fv (GLuint program, GLint location, GLsizei count, GLfloat *value)
    cdef void  glProgramUniform2d (GLuint program, GLint location, GLdouble v0, GLdouble v1)
    cdef void  glProgramUniform2dv (GLuint program, GLint location, GLsizei count, GLdouble *value)
    cdef void  glProgramUniform2ui (GLuint program, GLint location, GLuint v0, GLuint v1)
    cdef void  glProgramUniform2uiv (GLuint program, GLint location, GLsizei count, GLuint *value)
    cdef void  glProgramUniform3i (GLuint program, GLint location, GLint v0, GLint v1, GLint v2)
    cdef void  glProgramUniform3iv (GLuint program, GLint location, GLsizei count, GLint *value)
    cdef void  glProgramUniform3f (GLuint program, GLint location, GLfloat v0, GLfloat v1, GLfloat v2)
    cdef void  glProgramUniform3fv (GLuint program, GLint location, GLsizei count, GLfloat *value)
    cdef void  glProgramUniform3d (GLuint program, GLint location, GLdouble v0, GLdouble v1, GLdouble v2)
    cdef void  glProgramUniform3dv (GLuint program, GLint location, GLsizei count, GLdouble *value)
    cdef void  glProgramUniform3ui (GLuint program, GLint location, GLuint v0, GLuint v1, GLuint v2)
    cdef void  glProgramUniform3uiv (GLuint program, GLint location, GLsizei count, GLuint *value)
    cdef void  glProgramUniform4i (GLuint program, GLint location, GLint v0, GLint v1, GLint v2, GLint v3)
    cdef void  glProgramUniform4iv (GLuint program, GLint location, GLsizei count, GLint *value)
    cdef void  glProgramUniform4f (GLuint program, GLint location, GLfloat v0, GLfloat v1, GLfloat v2, GLfloat v3)
    cdef void  glProgramUniform4fv (GLuint program, GLint location, GLsizei count, GLfloat *value)
    cdef void  glProgramUniform4d (GLuint program, GLint location, GLdouble v0, GLdouble v1, GLdouble v2, GLdouble v3)
    cdef void  glProgramUniform4dv (GLuint program, GLint location, GLsizei count, GLdouble *value)
    cdef void  glProgramUniform4ui (GLuint program, GLint location, GLuint v0, GLuint v1, GLuint v2, GLuint v3)
    cdef void  glProgramUniform4uiv (GLuint program, GLint location, GLsizei count, GLuint *value)
    cdef void  glProgramUniformMatrix2fv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glProgramUniformMatrix3fv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glProgramUniformMatrix4fv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glProgramUniformMatrix2dv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glProgramUniformMatrix3dv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glProgramUniformMatrix4dv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glProgramUniformMatrix2x3fv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glProgramUniformMatrix3x2fv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glProgramUniformMatrix2x4fv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glProgramUniformMatrix4x2fv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glProgramUniformMatrix3x4fv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glProgramUniformMatrix4x3fv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glProgramUniformMatrix2x3dv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glProgramUniformMatrix3x2dv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glProgramUniformMatrix2x4dv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glProgramUniformMatrix4x2dv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glProgramUniformMatrix3x4dv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glProgramUniformMatrix4x3dv (GLuint program, GLint location, GLsizei count, GLboolean transpose, GLdouble *value)
    cdef void  glValidateProgramPipeline (GLuint pipeline)
    cdef void  glGetProgramPipelineInfoLog (GLuint pipeline, GLsizei bufSize, GLsizei *length, GLchar *infoLog)
    cdef void  glVertexAttribL1d (GLuint index, GLdouble x)
    cdef void  glVertexAttribL2d (GLuint index, GLdouble x, GLdouble y)
    cdef void  glVertexAttribL3d (GLuint index, GLdouble x, GLdouble y, GLdouble z)
    cdef void  glVertexAttribL4d (GLuint index, GLdouble x, GLdouble y, GLdouble z, GLdouble w)
    cdef void  glVertexAttribL1dv (GLuint index, GLdouble *v)
    cdef void  glVertexAttribL2dv (GLuint index, GLdouble *v)
    cdef void  glVertexAttribL3dv (GLuint index, GLdouble *v)
    cdef void  glVertexAttribL4dv (GLuint index, GLdouble *v)
    cdef void  glVertexAttribLPointer (GLuint index, GLint size, GLenum type, GLsizei stride, GLvoid *pointer)
    cdef void  glGetVertexAttribLdv (GLuint index, GLenum pname, GLdouble *params)
    cdef void  glViewportArrayv (GLuint first, GLsizei count, GLfloat *v)
    cdef void  glViewportIndexedf (GLuint index, GLfloat x, GLfloat y, GLfloat w, GLfloat h)
    cdef void  glViewportIndexedfv (GLuint index, GLfloat *v)
    cdef void  glScissorArrayv (GLuint first, GLsizei count, GLint *v)
    cdef void  glScissorIndexed (GLuint index, GLint left, GLint bottom, GLsizei width, GLsizei height)
    cdef void  glScissorIndexedv (GLuint index, GLint *v)
    cdef void  glDepthRangeArrayv (GLuint first, GLsizei count, GLdouble *v)
    cdef void  glDepthRangeIndexed (GLuint index, GLdouble n, GLdouble f)
    cdef void  glGetFloati_v (GLenum target, GLuint index, GLfloat *data)
    cdef void  glGetDoublei_v (GLenum target, GLuint index, GLdouble *data)
    cdef void  glDrawArraysInstancedBaseInstance (GLenum mode, GLint first, GLsizei count, GLsizei instancecount, GLuint baseinstance)
    cdef void  glDrawElementsInstancedBaseInstance (GLenum mode, GLsizei count, GLenum type, void *indices, GLsizei instancecount, GLuint baseinstance)
    cdef void  glDrawElementsInstancedBaseVertexBaseInstance (GLenum mode, GLsizei count, GLenum type, void *indices, GLsizei instancecount, GLint basevertex, GLuint baseinstance)
    cdef void  glDrawTransformFeedbackInstanced (GLenum mode, GLuint id, GLsizei instancecount)
    cdef void  glDrawTransformFeedbackStreamInstanced (GLenum mode, GLuint id, GLuint stream, GLsizei instancecount)
    cdef void  glGetInternalformativ (GLenum target, GLenum internalformat, GLenum pname, GLsizei bufSize, GLint *params)
    cdef void  glGetActiveAtomicCounterBufferiv (GLuint program, GLuint bufferIndex, GLenum pname, GLint *params)
    cdef void  glBindImageTexture (GLuint unit, GLuint texture, GLint level, GLboolean layered, GLint layer, GLenum access, GLenum format)
    cdef void  glMemoryBarrier (GLbitfield barriers)
    cdef void  glTexStorage1D (GLenum target, GLsizei levels, GLenum internalformat, GLsizei width)
    cdef void  glTexStorage2D (GLenum target, GLsizei levels, GLenum internalformat, GLsizei width, GLsizei height)
    cdef void  glTexStorage3D (GLenum target, GLsizei levels, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth)
    cdef void  glDebugMessageControl (GLenum source, GLenum type, GLenum severity, GLsizei count, GLuint *ids, GLboolean enabled)
    cdef void  glDebugMessageInsert (GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, GLchar *buf)
    cdef GLuint  glGetDebugMessageLog (GLuint count, GLsizei bufsize, GLenum *sources, GLenum *types, GLuint *ids, GLenum *severities, GLsizei *lengths, GLchar *messageLog)
    cdef void  glPushDebugGroup (GLenum source, GLuint id, GLsizei length, GLchar *message)
    cdef void  glPopDebugGroup ()
    cdef void  glObjectLabel (GLenum identifier, GLuint name, GLsizei length, GLchar *label)
    cdef void  glGetObjectLabel (GLenum identifier, GLuint name, GLsizei bufSize, GLsizei *length, GLchar *label)
    cdef void  glObjectPtrLabel (void *ptr, GLsizei length, GLchar *label)
    cdef void  glGetObjectPtrLabel (void *ptr, GLsizei bufSize, GLsizei *length, GLchar *label)
    cdef void  glClearBufferData (GLenum target, GLenum internalformat, GLenum format, GLenum type, void *data)
    cdef void  glClearBufferSubData (GLenum target, GLenum internalformat, GLintptr offset, GLsizeiptr size, GLenum format, GLenum type, void *data)
    cdef void  glDispatchCompute (GLuint num_groups_x, GLuint num_groups_y, GLuint num_groups_z)
    cdef void  glDispatchComputeIndirect (GLintptr indirect)
    cdef void  glCopyImageSubData (GLuint srcName, GLenum srcTarget, GLint srcLevel, GLint srcX, GLint srcY, GLint srcZ, GLuint dstName, GLenum dstTarget, GLint dstLevel, GLint dstX, GLint dstY, GLint dstZ, GLsizei srcWidth, GLsizei srcHeight, GLsizei srcDepth)
    cdef void  glTextureView (GLuint texture, GLenum target, GLuint origtexture, GLenum internalformat, GLuint minlevel, GLuint numlevels, GLuint minlayer, GLuint numlayers)
    cdef void  glBindVertexBuffer (GLuint bindingindex, GLuint buffer, GLintptr offset, GLsizei stride)
    cdef void  glVertexAttribFormat (GLuint attribindex, GLint size, GLenum type, GLboolean normalized, GLuint relativeoffset)
    cdef void  glVertexAttribIFormat (GLuint attribindex, GLint size, GLenum type, GLuint relativeoffset)
    cdef void  glVertexAttribLFormat (GLuint attribindex, GLint size, GLenum type, GLuint relativeoffset)
    cdef void  glVertexAttribBinding (GLuint attribindex, GLuint bindingindex)
    cdef void  glVertexBindingDivisor (GLuint bindingindex, GLuint divisor)
    cdef void  glFramebufferParameteri (GLenum target, GLenum pname, GLint param)
    cdef void  glGetFramebufferParameteriv (GLenum target, GLenum pname, GLint *params)
    cdef void  glInvalidateTexSubImage (GLuint texture, GLint level, GLint xoffset, GLint yoffset, GLint zoffset, GLsizei width, GLsizei height, GLsizei depth)
    cdef void  glInvalidateTexImage (GLuint texture, GLint level)
    cdef void  glInvalidateBufferSubData (GLuint buffer, GLintptr offset, GLsizeiptr length)
    cdef void  glInvalidateBufferData (GLuint buffer)
    cdef void  glInvalidateFramebuffer (GLenum target, GLsizei numAttachments, GLenum *attachments)
    cdef void  glInvalidateSubFramebuffer (GLenum target, GLsizei numAttachments, GLenum *attachments, GLint x, GLint y, GLsizei width, GLsizei height)
    cdef void  glMultiDrawArraysIndirect (GLenum mode, void *indirect, GLsizei drawcount, GLsizei stride)
    cdef void  glMultiDrawElementsIndirect (GLenum mode, GLenum type, void *indirect, GLsizei drawcount, GLsizei stride)
    cdef void  glGetProgramInterfaceiv (GLuint program, GLenum programInterface, GLenum pname, GLint *params)
    cdef GLuint  glGetProgramResourceIndex (GLuint program, GLenum programInterface, GLchar *name)
    cdef void  glGetProgramResourceName (GLuint program, GLenum programInterface, GLuint index, GLsizei bufSize, GLsizei *length, GLchar *name)
    cdef void  glGetProgramResourceiv (GLuint program, GLenum programInterface, GLuint index, GLsizei propCount, GLenum *props, GLsizei bufSize, GLsizei *length, GLint *params)
    cdef GLint  glGetProgramResourceLocation (GLuint program, GLenum programInterface, GLchar *name)
    cdef GLint  glGetProgramResourceLocationIndex (GLuint program, GLenum programInterface, GLchar *name)
    cdef void  glShaderStorageBlockBinding (GLuint program, GLuint storageBlockIndex, GLuint storageBlockBinding)
    cdef void  glTexBufferRange (GLenum target, GLenum internalformat, GLuint buffer, GLintptr offset, GLsizeiptr size)
    cdef void  glTexStorage2DMultisample (GLenum target, GLsizei samples, GLenum internalformat, GLsizei width, GLsizei height, GLboolean fixedsamplelocations)
    cdef void  glTexStorage3DMultisample (GLenum target, GLsizei samples, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLboolean fixedsamplelocations)
